# Generated by roxygen2: do not edit by hand

S3method(as.character,ps_handle)
S3method(format,ps_handle)
S3method(print,ps_handle)
S3method(print,with_process_cleanup)
export(CleanupReporter)
export(errno)
export(ps)
export(ps_apps)
export(ps_boot_time)
export(ps_children)
export(ps_cmdline)
export(ps_connections)
export(ps_cpu_count)
export(ps_cpu_times)
export(ps_create_time)
export(ps_cwd)
export(ps_descent)
export(ps_disk_io_counters)
export(ps_disk_partitions)
export(ps_disk_usage)
export(ps_environ)
export(ps_environ_raw)
export(ps_exe)
export(ps_find_tree)
export(ps_fs_info)
export(ps_fs_mount_point)
export(ps_fs_stat)
export(ps_get_cpu_affinity)
export(ps_get_nice)
export(ps_gids)
export(ps_handle)
export(ps_interrupt)
export(ps_is_running)
export(ps_is_supported)
export(ps_kill)
export(ps_kill_tree)
export(ps_loadavg)
export(ps_mark_tree)
export(ps_memory_full_info)
export(ps_memory_info)
export(ps_name)
export(ps_num_fds)
export(ps_num_threads)
export(ps_open_files)
export(ps_os_type)
export(ps_parent)
export(ps_pid)
export(ps_pids)
export(ps_ppid)
export(ps_resume)
export(ps_send_signal)
export(ps_set_cpu_affinity)
export(ps_set_nice)
export(ps_shared_lib_users)
export(ps_shared_libs)
export(ps_status)
export(ps_string)
export(ps_suspend)
export(ps_system_cpu_times)
export(ps_system_memory)
export(ps_system_swap)
export(ps_terminal)
export(ps_terminate)
export(ps_tty_size)
export(ps_uids)
export(ps_username)
export(ps_users)
export(ps_wait)
export(ps_windows_nice_values)
export(signals)
export(with_process_cleanup)
importFrom(utils,head)
importFrom(utils,read.delim)
importFrom(utils,read.table)
importFrom(utils,tail)
useDynLib(ps, .registration = TRUE)

# ps

> List, Query, Manipulate System Processes

<!-- badges: start -->
[![lifecycle](https://lifecycle.r-lib.org/articles/figures/lifecycle-stable.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![R-CMD-check](https://github.com/r-lib/ps/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/ps/actions/workflows/R-CMD-check.yaml)
[![CRAN status](https://www.r-pkg.org/badges/version/ps)](https://cran.r-project.org/package=ps)
[![CRAN RStudio mirror downloads](https://cranlogs.r-pkg.org/badges/ps)](https://www.r-pkg.org/pkg/ps)
[![Codecov test coverage](https://codecov.io/gh/r-lib/ps/branch/main/graph/badge.svg)](https://app.codecov.io/gh/r-lib/ps?branch=main)
<!-- badges: end -->

ps implements an API to query and manipulate system processes. Most of its
code is based on the [psutil](https://github.com/giampaolo/psutil) Python
package.


- [Installation](#installation)
- [Supported platforms](#supported-platforms)
- [Listing all processes](#listing-all-processes)
- [Process API](#process-api)
  - [Query functions](#query-functions)
  - [Process manipulation](#process-manipulation)
- [Finished and zombie processes](#finished-and-zombie-processes)
- [Pid reuse](#pid-reuse)
- [Recipes](#recipes)
  - [Find process by name](#find-process-by-name)
  - [Wait for a process to finish](#wait-for-a-process-to-finish)
  - [Wait for several processes to
    finish](#wait-for-several-processes-to-finish)
  - [Kill process tree](#kill-process-tree)
  - [Filtering and sorting processes](#filtering-and-sorting-processes)
- [Code of Conduct](#code-of-conduct)
- [License](#license)

## Installation

You can install the released version of ps from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("ps")
```

If you need the development version, install it with

``` r
pak::pak("r-lib/ps")
```

``` r
library(ps)
library(pillar) # nicer printing of data frames
```

## Supported platforms

ps currently supports Windows (from Vista), macOS and Linux systems. On
unsupported platforms the package can be installed and loaded, but all
of its functions fail with an error of class `"not_implemented"`.

## Listing all processes

`ps_pids()` returns all process ids on the system. This can be useful to
iterate over all processes.

``` r
ps_pids()[1:20]
```

    ##  [1]    0    1 1125 1127 1129 1130 1133 1136 1138 1139 1144 1149 1153 1155 1156 1161 1164 1165 1166
    ## [20] 1167

`ps()` returns a data frame, with data about each process. It contains a
handle to each process, in the `ps_handle` column, you can use these to
perform more queries on the processes.

``` r
ps()
```

    ## # A data frame: 477 × 11
    ##      pid  ppid name  username status    user  system     rss      vms created             ps_handle
    ##  * <int> <int> <chr> <chr>    <chr>    <dbl>   <dbl>   <dbl>    <dbl> <dttm>              <I<list>>
    ##  1 81068     1 mdwo… gaborcs… runni…  0.0377  0.0143  2.61e7  4.20e11 2024-09-01 09:39:46 <ps_handl>
    ##  2 81067     1 mdwo… gaborcs… runni…  0.0423  0.0133  2.19e7  4.20e11 2024-09-01 09:39:46 <ps_handl>
    ##  3 80413     1 mdwo… gaborcs… runni…  0.0717  0.0220  2.84e7  4.20e11 2024-09-01 09:39:27 <ps_handl>
    ##  4 80412     1 mdwo… gaborcs… runni…  0.0749  0.0241  3.00e7  4.20e11 2024-09-01 09:39:27 <ps_handl>
    ##  5 80366     1 mdwo… gaborcs… runni…  0.287   0.0797  3.00e7  4.20e11 2024-09-01 09:37:07 <ps_handl>
    ##  6 80360 69319 Goog… gaborcs… runni…  0.0462  0.0191  9.14e7  1.66e12 2024-09-01 09:37:05 <ps_handl>
    ##  7 80264     1 coun… root     runni… NA      NA      NA      NA       2024-09-01 09:36:10 <ps_handl>
    ##  8 80261     1 mdwo… gaborcs… runni…  0.300   0.0755  3.00e7  4.20e11 2024-09-01 09:35:55 <ps_handl>
    ##  9 78900 69319 Goog… gaborcs… runni…  2.26    0.429   1.76e8  1.66e12 2024-09-01 09:26:07 <ps_handl>
    ## 10 78888 69319 Goog… gaborcs… runni…  5.68    0.595   2.38e8  1.66e12 2024-09-01 09:25:57 <ps_handl>
    ## # ℹ 467 more rows

## Process API

This is a short summary of the API. Please see the documentation of the
various methods for details, in particular regarding handles to finished
processes and pid reuse. See also “Finished and zombie processes” and
“pid reuse” below.

`ps_handle(pid)` creates a process handle for the supplied process id.
If `pid` is omitted, a handle to the calling process is returned:

``` r
p <- ps_handle()
p
```

    ## <ps::ps_handle> PID=56773, NAME=R, AT=2024-08-31 14:11:26.708347

### Query functions

`ps_pid(p)` returns the pid of the process.

``` r
ps_pid(p)
```

    ## [1] 56773

`ps_create_time()` returns the creation time of the process (according
to the OS).

``` r
ps_create_time(p)
```

    ## [1] "2024-08-31 14:11:26 GMT"

The process id and the creation time uniquely identify a process in a
system. ps uses them to make sure that it reports information about, and
manipulates the correct process.

`ps_is_running(p)` returns whether `p` is still running. It handles pid
reuse safely.

``` r
ps_is_running(p)
```

    ## [1] TRUE

`ps_ppid(p)` returns the pid of the parent of `p`.

``` r
ps_ppid(p)
```

    ## [1] 55975

`ps_parent(p)` returns a process handle to the parent process of `p`.

``` r
ps_parent(p)
```

    ## <ps::ps_handle> PID=55975, NAME=zsh, AT=2024-08-29 15:04:35.020175

`ps_name(p)` returns the name of the program `p` is running.

``` r
ps_name(p)
```

    ## [1] "R"

`ps_exe(p)` returns the full path to the executable the `p` is running.

``` r
ps_exe(p)
```

    ## [1] "/Library/Frameworks/R.framework/Versions/4.4-arm64/Resources/bin/exec/R"

`ps_cmdline(p)` returns the command line (executable and arguments) of
`p`.

``` r
ps_cmdline(p)
```

    ## [1] "/Library/Frameworks/R.framework/Versions/4.4-arm64/Resources/bin/exec/R"

`ps_status(p)` returns the status of the process. Possible values are OS
dependent, but typically there is `"running"` and `"stopped"`.

``` r
ps_status(p)
```

    ## [1] "running"

`ps_username(p)` returns the name of the user the process belongs to.

``` r
ps_username(p)
```

    ## [1] "gaborcsardi"

`ps_uids(p)` and `ps_gids(p)` return the real, effective and saved user
ids of the process. They are only implemented on POSIX systems.

``` r
if (ps_os_type()[["POSIX"]]) ps_uids(p)
```

    ##      real effective     saved
    ##       501       501       501

``` r
if (ps_os_type()[["POSIX"]]) ps_gids(p)
```

    ##      real effective     saved
    ##        20        20        20

`ps_cwd(p)` returns the current working directory of the process.

``` r
ps_cwd(p)
```

    ## [1] "/Users/gaborcsardi/works/ps"

`ps_terminal(p)` returns the name of the terminal of the process, if
any. For processes without a terminal, and on Windows it returns
`NA_character_`.

``` r
ps_terminal(p)
```

    ## [1] "/dev/ttys015"

`ps_environ(p)` returns the environment variables of the process.
`ps_environ_raw(p)` does the same, in a different form. Typically they
reflect the environment variables at the start of the process.

``` r
ps_environ(p)[c("TERM", "USER", "SHELL", "R_HOME")]
```

    ## TERM                          xterm-256color
    ## USER                          gaborcsardi
    ## SHELL                         /bin/zsh
    ## R_HOME                        /Library/Frameworks/R.framework/Versions/4.4-arm64/Resources

`ps_num_threads(p)` returns the current number of threads of the
process.

``` r
ps_num_threads(p)
```

    ## [1] 3

`ps_cpu_times(p)` returns the CPU times of the process, similarly to
`proc.time()`.

``` r
ps_cpu_times(p)
```

    ##            user          system   children_user children_system
    ##        33.87137        15.05281              NA              NA

`ps_memory_info(p)` returns memory usage information. See the manual for
details.

``` r
ps_memory_info(p)
```

    ##          rss          vms      pfaults      pageins
    ##    255983616 423297335296      2173700          742

`ps_children(p)` lists all child processes (potentially recursively) of
the current process.

``` r
ps_children(ps_parent(p))
```

    ## [[1]]
    ## <ps::ps_handle> PID=56773, NAME=R, AT=2024-08-31 14:11:26.708347
    ##
    ## [[2]]
    ## <ps::ps_handle> PID=57966, NAME=zsh, AT=2024-08-29 15:04:37.802087

`ps_num_fds(p)` returns the number of open file descriptors (handles on
Windows):

``` r
ps_num_fds(p)
```

    ## [1] 49

``` r
f <- file(tmp <- tempfile(), "w")
ps_num_fds(p)
```

    ## [1] 50

``` r
close(f)
unlink(tmp)
```

`ps_open_files(p)` lists all open files:

``` r
ps_open_files(p)
```

    ## # A data frame: 3 × 2
    ##       fd path
    ##    <int> <chr>
    ##  1     0 /dev/ttys015
    ##  2     1 /dev/ttys015
    ##  3     2 /dev/ttys015

``` r
f <- file(tmp <- tempfile(), "w")
ps_open_files(p)
```

    ## # A data frame: 4 × 2
    ##       fd path
    ##    <int> <chr>
    ##  1     0 /dev/ttys015
    ##  2     1 /dev/ttys015
    ##  3     2 /dev/ttys015
    ##  4    45 /private/var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpFPtZXU/fileddc51cac4863

``` r
close(f)
unlink(tmp)
ps_open_files(p)
```

    ## # A data frame: 3 × 2
    ##       fd path
    ##    <int> <chr>
    ##  1     0 /dev/ttys015
    ##  2     1 /dev/ttys015
    ##  3     2 /dev/ttys015

### Process manipulation

`ps_suspend(p)` suspends (stops) the process. On POSIX it sends a
SIGSTOP signal. On Windows it stops all threads.

`ps_resume(p)` resumes the process. On POSIX it sends a SIGCONT signal.
On Windows it resumes all stopped threads.

`ps_send_signal(p)` sends a signal to the process. It is implemented on
POSIX systems only. It makes an effort to work around pid reuse.

`ps_terminate(p)` send SIGTERM to the process. On POSIX systems only.

`ps_kill(p)` terminates the process. Sends `SIGKILL` on POSIX systems,
uses `TerminateProcess()` on Windows. It make an effort to work around
pid reuse.

`ps_interrupt(p)` interrupts a process. It sends a `SIGINT` signal on
POSIX systems, and it can send a CTRL+C or a CTRL+BREAK event on
Windows.

## Finished and zombie processes

ps handles finished and Zombie processes as much as possible.

The essential `ps_pid()`, `ps_create_time()`, `ps_is_running()`
functions and the `format()` and `print()` methods work for all
processes, including finished and zombie processes. Other functions fail
with an error of class `"no_such_process"` for finished processes.

The `ps_ppid()`, `ps_parent()`, `ps_children()`, `ps_name()`,
`ps_status()`, `ps_username()`, `ps_uids()`, `ps_gids()`,
`ps_terminal()`, `ps_children()` and the signal sending functions work
properly for zombie processes. Other functions fail with
`"zombie_process"` error.

## Pid reuse

ps functions handle pid reuse as well as technically possible.

The query functions never return information about the wrong process,
even if the process has finished and its process id was re-assigned.

On Windows, the process manipulation functions never manipulate the
wrong process.

On POSIX systems, this is technically impossible, it is not possible to
send a signal to a process without creating a race condition. In ps the
time window of the race condition is very small, a few microseconds, and
the process would need to finish, *and* the OS would need to reuse its
pid within this time window to create problems. This is very unlikely to
happen.

## Recipes

In the spirit of [psutil
recipes](http://psutil.readthedocs.io/en/latest/#recipes).

### Find process by name

Using `ps()` and dplyr:

``` r
library(dplyr)
find_procs_by_name <- function(name) {
  ps() %>%
    filter(name == !!name)  %>%
    pull(ps_handle)
}

find_procs_by_name("R")
```

    ## [[1]]
    ## <ps::ps_handle> PID=56773, NAME=R, AT=2024-08-31 14:11:26.708347

Without creating the full table of processes:

``` r
find_procs_by_name <- function(name) {
  procs <- lapply(ps_pids(), function(p) {
    tryCatch({
      h <- ps_handle(p)
      if (ps_name(h) == name) h else NULL },
      no_such_process = function(e) NULL,
      access_denied = function(e) NULL
    )
  })
  procs[!vapply(procs, is.null, logical(1))]
  }

find_procs_by_name("R")
```

    ## [[1]]
    ## <ps::ps_handle> PID=56773, NAME=R, AT=2024-08-31 14:11:26.708347

### Wait for a process to finish

`ps_wait()`, from ps 1.8.0, implements a new way, efficient for waiting
on a list of processes, so this is now very easy:

``` r
px <- processx::process$new("sleep", "2")
p <- px$as_ps_handle()
ps_wait(p, 1000)
```

    ## [1] FALSE

``` r
ps_wait(p)
```

    ## [1] TRUE

### Wait for several processes to finish

Again, this is much simpler with `ps_wait()`, added in ps 1.8.0.

``` r
px1 <- processx::process$new("sleep", "10")
px2 <- processx::process$new("sleep", "10")
px3 <- processx::process$new("sleep", "1")
px4 <- processx::process$new("sleep", "1")

p1 <- px1$as_ps_handle()
p2 <- px2$as_ps_handle()
p3 <- px3$as_ps_handle()
p4 <- px4$as_ps_handle()

ps_wait(list(p1, p2, p3, p4), timeout = 2000)
```

    ## [1] FALSE FALSE  TRUE  TRUE

### Kill process tree

From ps 1.8.0, `ps_kill()` will first send `SIGTERM` signals on Unix,
and `SIGKILL` after a grace period, if needed.

Note, that some R IDEs, including RStudio, run a multithreaded R
process, and other threads may start processes as well.
`reap_children()` will clean up all these as well, potentially causing
the IDE to misbehave or crash.

``` r
kill_proc_tree <- function(pid, include_parent = TRUE, ...) {
  if (pid == Sys.getpid() && include_parent) stop("I refuse to kill myself")
  parent <- ps_handle(pid)
  children <- ps_children(parent, recursive = TRUE)
  if (include_parent) children <- c(children, list(parent))
  ps_kill(children, ...)
}

p1 <- processx::process$new("sleep", "10")
p2 <- processx::process$new("sleep", "10")
p3 <- processx::process$new("sleep", "10")
kill_proc_tree(Sys.getpid(), include_parent = FALSE)
```

    ## [1] "terminated" "terminated" "terminated" "terminated" "terminated"

### Filtering and sorting processes

Process name ending with “sh”:

``` r
ps() %>%
  filter(grepl("sh$", name))
```

    ## # A data frame: 38 × 11
    ##      pid  ppid name    username status    user  system    rss     vms created             ps_handle
    ##    <int> <int> <chr>   <chr>    <chr>    <dbl>   <dbl>  <dbl>   <dbl> <dttm>              <I<list>>
    ##  1 67374     1 Report… gaborcs… runni… 0.00899 0.0148  5.47e6 4.20e11 2024-09-01 03:02:35 <ps_handl>
    ##  2 44801 44603 zsh     gaborcs… runni… 0.00261 0.00945 7.86e5 4.21e11 2024-08-31 09:42:36 <ps_handl>
    ##  3 44603 44602 zsh     gaborcs… runni… 0.155   0.0547  9.99e5 4.21e11 2024-08-31 09:42:35 <ps_handl>
    ##  4 24830 24631 zsh     gaborcs… runni… 0.00736 0.0332  7.86e5 4.21e11 2024-08-30 22:11:24 <ps_handl>
    ##  5 24631 24630 zsh     gaborcs… runni… 0.333   0.134   9.67e5 4.21e11 2024-08-30 22:11:24 <ps_handl>
    ##  6 58680 55972 zsh     gaborcs… runni… 0.168   0.651   9.34e5 4.21e11 2024-08-29 15:04:39 <ps_handl>
    ##  7 58570 55971 zsh     gaborcs… runni… 0.0186  0.0693  2.38e6 4.21e11 2024-08-29 15:04:39 <ps_handl>
    ##  8 58509 55974 zsh     gaborcs… runni… 0.00364 0.0117  7.54e5 4.21e11 2024-08-29 15:04:38 <ps_handl>
    ##  9 58474 55964 zsh     gaborcs… runni… 0.00367 0.00942 7.54e5 4.21e11 2024-08-29 15:04:38 <ps_handl>
    ## 10 58437 55966 zsh     gaborcs… runni… 0.00341 0.00986 7.54e5 4.21e11 2024-08-29 15:04:38 <ps_handl>
    ## # ℹ 28 more rows

Processes owned by user:

``` r
ps() %>%
  filter(username == Sys.info()[["user"]]) %>%
  select(pid, name)
```

    ## # A data frame: 286 × 2
    ##      pid name
    ##    <int> <chr>
    ##  1 81199 Google Chrome Helper (Renderer)
    ##  2 81198 Google Chrome Helper (Renderer)
    ##  3 81197 Google Chrome Helper (Renderer)
    ##  4 81068 mdworker_shared
    ##  5 81067 mdworker_shared
    ##  6 80413 mdworker_shared
    ##  7 80412 mdworker_shared
    ##  8 80366 mdworker_shared
    ##  9 80360 Google Chrome Helper (Renderer)
    ## 10 80261 mdworker_shared
    ## # ℹ 276 more rows

Processes consuming more than 100MB of memory:

``` r
ps() %>%
  filter(rss > 100 * 1024 * 1024)
```

    ## # A data frame: 29 × 11
    ##      pid  ppid name    username status    user  system    rss     vms created             ps_handle
    ##    <int> <int> <chr>   <chr>    <chr>    <dbl>   <dbl>  <dbl>   <dbl> <dttm>              <I<list>>
    ##  1 81199 69319 Google… gaborcs… runni…  0.0412  0.0162 1.07e8 1.66e12 2024-09-01 09:39:54 <ps_handl>
    ##  2 81198 69319 Google… gaborcs… runni…  0.0704  0.0242 1.22e8 1.66e12 2024-09-01 09:39:53 <ps_handl>
    ##  3 81197 69319 Google… gaborcs… runni…  1.23    0.565  3.12e8 1.66e12 2024-09-01 09:39:53 <ps_handl>
    ##  4 78900 69319 Google… gaborcs… runni…  2.27    0.430  1.76e8 1.66e12 2024-09-01 09:26:07 <ps_handl>
    ##  5 78888 69319 Google… gaborcs… runni…  5.68    0.596  2.38e8 1.66e12 2024-09-01 09:25:57 <ps_handl>
    ##  6 76222 69319 Google… gaborcs… runni… 34.6     2.86   2.22e8 1.66e12 2024-09-01 09:17:46 <ps_handl>
    ##  7 74007 69319 Google… gaborcs… runni…  6.59    1.08   1.46e8 1.66e12 2024-09-01 08:37:02 <ps_handl>
    ##  8 73963 69319 Google… gaborcs… runni…  7.16    0.987  1.86e8 1.66e12 2024-09-01 08:34:37 <ps_handl>
    ##  9 73601 69319 Google… gaborcs… runni… 98.1    19.2    1.94e8 1.66e12 2024-09-01 07:54:27 <ps_handl>
    ## 10 69386 69319 Google… gaborcs… runni…  1.38    0.244  1.09e8 1.66e12 2024-09-01 07:06:19 <ps_handl>
    ## # ℹ 19 more rows

Top 3 memory consuming processes:

``` r
ps() %>%
  top_n(3, rss) %>%
  arrange(desc(rss))
```

    ## # A data frame: 3 × 11
    ##     pid  ppid name       username status   user system    rss     vms created             ps_handle
    ##   <int> <int> <chr>      <chr>    <chr>   <dbl>  <dbl>  <dbl>   <dbl> <dttm>              <I<list>>
    ## 1 64110     1 com.apple… gaborcs… runni… 68283. 7992.  7.39e8 4.29e11 2024-07-17 08:14:51 <ps_handl>
    ## 2 56754 56683 ark        gaborcs… runni…   813.  112.  5.66e8 4.22e11 2024-08-29 15:04:36 <ps_handl>
    ## 3 69319     1 Google Ch… gaborcs… runni…   280.   86.9 5.63e8 4.56e11 2024-09-01 07:06:06 <ps_handl>

Top 3 processes which consumed the most CPU time:

``` r
ps() %>%
  mutate(cpu_time = user + system) %>%
  top_n(3, cpu_time) %>%
  arrange(desc(cpu_time)) %>%
  select(pid, name, cpu_time)
```

    ## # A data frame: 3 × 3
    ##     pid name                                    cpu_time
    ##   <int> <chr>                                      <dbl>
    ## 1 64110 com.apple.Virtualization.VirtualMachine   76275.
    ## 2  3525 Dato                                      12825.
    ## 3  2617 CursorUIViewService                        6700.

## Code of Conduct

Please note that the ps project is released with a [Contributor Code of
Conduct](https://ps.r-lib.org/CODE_OF_CONDUCT.html). By contributing to
this project, you agree to abide by its terms.

## License

MIT © RStudio

# ps

> List, Query, Manipulate System Processes

<!-- badges: start -->
[![lifecycle](https://lifecycle.r-lib.org/articles/figures/lifecycle-stable.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![R-CMD-check](https://github.com/r-lib/ps/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/ps/actions/workflows/R-CMD-check.yaml)
[![CRAN status](https://www.r-pkg.org/badges/version/ps)](https://cran.r-project.org/package=ps)
[![CRAN RStudio mirror downloads](https://cranlogs.r-pkg.org/badges/ps)](https://www.r-pkg.org/pkg/ps)
[![Codecov test coverage](https://codecov.io/gh/r-lib/ps/graph/badge.svg)](https://app.codecov.io/gh/r-lib/ps)
<!-- badges: end -->

ps implements an API to query and manipulate system processes. Most of its
code is based on the [psutil](https://github.com/giampaolo/psutil) Python
package.


-   [Installation](#installation)
-   [Supported platforms](#supported-platforms)
-   [Listing all processes](#listing-all-processes)
-   [Process API](#process-api)
    -   [Query functions](#query-functions)
    -   [Process manipulation](#process-manipulation)
-   [Finished and zombie processes](#finished-and-zombie-processes)
-   [Pid reuse](#pid-reuse)
-   [Recipes](#recipes)
    -   [Find process by name](#find-process-by-name)
    -   [Wait for a process to finish](#wait-for-a-process-to-finish)
    -   [Wait for several processes to
        finish](#wait-for-several-processes-to-finish)
    -   [Kill process tree](#kill-process-tree)
    -   [Filtering and sorting
        processes](#filtering-and-sorting-processes)
-   [Code of Conduct](#code-of-conduct)
-   [License](#license)

## Installation

You can install the released version of ps from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("ps")
```

If you need the development version, install it with

``` r
pak::pak("r-lib/ps")
```

``` r
library(ps)
library(pillar) # nicer printing of data frames
```

## Supported platforms

ps currently supports Windows (from Vista), macOS and Linux systems. On
unsupported platforms the package can be installed and loaded, but all
of its functions fail with an error of class `"not_implemented"`.

## Listing all processes

`ps_pids()` returns all process ids on the system. This can be useful to
iterate over all processes.

``` r
ps_pids()[1:20]
```

    ##  [1]   0   1 265 266 275 276 557 559 561 562 564 567 569 570 574 578 580 584 586 587

`ps()` returns a data frame, with data about each process. It contains a
handle to each process, in the `ps_handle` column, you can use these to
perform more queries on the processes.

``` r
ps()
```

    ## # A data frame: 572 × 11
    ##      pid  ppid name      username status   user system    rss     vms created             ps_handle 
    ##  * <int> <int> <chr>     <chr>    <chr>   <dbl>  <dbl>  <dbl>   <dbl> <dttm>              <I<list>> 
    ##  1 45911  4159 Google C… gaborcs… sleep… 0.0473 0.0227 1.05e8 1.91e12 2025-04-28 08:05:25 <ps_handl>
    ##  2 45860  4159 Google C… gaborcs… sleep… 0.0751 0.0347 1.21e8 1.91e12 2025-04-28 08:05:19 <ps_handl>
    ##  3 45700     1 mdworker… gaborcs… sleep… 0.327  0.0861 2.88e7 4.37e11 2025-04-28 08:04:26 <ps_handl>
    ##  4 44844 39673 R         gaborcs… runni… 1.70   0.634  2.55e8 4.22e11 2025-04-28 08:01:34 <ps_handl>
    ##  5 44738  4159 Google C… gaborcs… sleep… 0.0325 0.0157 4.77e7 4.89e11 2025-04-28 08:00:51 <ps_handl>
    ##  6 43786     1 Messages… gaborcs… sleep… 0.162  0.233  6.90e6 4.37e11 2025-04-28 07:59:53 <ps_handl>
    ##  7 43785     1 business… gaborcs… sleep… 0.110  0.112  8.72e6 4.37e11 2025-04-28 07:59:53 <ps_handl>
    ##  8 43628     1 MENotifi… gaborcs… sleep… 0.0291 0.0116 3.08e6 4.21e11 2025-04-28 07:59:35 <ps_handl>
    ##  9 43627     1 MTLAsset… gaborcs… sleep… 0.0498 0.0208 3.29e6 4.37e11 2025-04-28 07:59:35 <ps_handl>
    ## 10 43472     1 replayd   gaborcs… sleep… 2.32   0.560  2.46e7 4.37e11 2025-04-28 07:59:21 <ps_handl>
    ## # ℹ 562 more rows

## Process API

This is a short summary of the API. Please see the documentation of the
various methods for details, in particular regarding handles to finished
processes and pid reuse. See also “Finished and zombie processes” and
“pid reuse” below.

`ps_handle(pid)` creates a process handle for the supplied process id.
If `pid` is omitted, a handle to the calling process is returned:

``` r
p <- ps_handle()
p
```

    ## <ps::ps_handle> PID=44844, NAME=R, AT=2025-04-28 08:01:34.873763

### Query functions

`ps_pid(p)` returns the pid of the process.

``` r
ps_pid(p)
```

    ## [1] 44844

`ps_create_time()` returns the creation time of the process (according
to the OS).

``` r
ps_create_time(p)
```

    ## [1] "2025-04-28 08:01:34 GMT"

The process id and the creation time uniquely identify a process in a
system. ps uses them to make sure that it reports information about, and
manipulates the correct process.

`ps_is_running(p)` returns whether `p` is still running. It handles pid
reuse safely.

``` r
ps_is_running(p)
```

    ## [1] TRUE

`ps_ppid(p)` returns the pid of the parent of `p`.

``` r
ps_ppid(p)
```

    ## [1] 39673

`ps_parent(p)` returns a process handle to the parent process of `p`.

``` r
ps_parent(p)
```

    ## <ps::ps_handle> PID=39673, NAME=zsh, AT=2025-04-28 07:53:00.030062

`ps_name(p)` returns the name of the program `p` is running.

``` r
ps_name(p)
```

    ## [1] "R"

`ps_exe(p)` returns the full path to the executable the `p` is running.

``` r
ps_exe(p)
```

    ## [1] "/Library/Frameworks/R.framework/Versions/4.5-arm64/Resources/bin/exec/R"

`ps_cmdline(p)` returns the command line (executable and arguments) of
`p`.

``` r
ps_cmdline(p)
```

    ## [1] "/Library/Frameworks/R.framework/Versions/4.5-arm64/Resources/bin/exec/R"

`ps_status(p)` returns the status of the process. Possible values are OS
dependent, but typically there is `"running"` and `"stopped"`.

``` r
ps_status(p)
```

    ## [1] "running"

`ps_username(p)` returns the name of the user the process belongs to.

``` r
ps_username(p)
```

    ## [1] "gaborcsardi"

`ps_uids(p)` and `ps_gids(p)` return the real, effective and saved user
ids of the process. They are only implemented on POSIX systems.

``` r
if (ps_os_type()[["POSIX"]]) ps_uids(p)
```

    ##      real effective     saved 
    ##       501       501       501

``` r
if (ps_os_type()[["POSIX"]]) ps_gids(p)
```

    ##      real effective     saved 
    ##        20        20        20

`ps_cwd(p)` returns the current working directory of the process.

``` r
ps_cwd(p)
```

    ## [1] "/Users/gaborcsardi/works/ps"

`ps_terminal(p)` returns the name of the terminal of the process, if
any. For processes without a terminal, and on Windows it returns
`NA_character_`.

``` r
ps_terminal(p)
```

    ## [1] "/dev/ttys019"

`ps_environ(p)` returns the environment variables of the process.
`ps_environ_raw(p)` does the same, in a different form. Typically they
reflect the environment variables at the start of the process.

``` r
ps_environ(p)[c("TERM", "USER", "SHELL", "R_HOME")]
```

    ## TERM                          xterm-256color
    ## USER                          gaborcsardi
    ## SHELL                         /bin/zsh
    ## R_HOME                        /Library/Frameworks/R.framework/Versions/4.5-arm64/Resources

`ps_num_threads(p)` returns the current number of threads of the
process.

``` r
ps_num_threads(p)
```

    ## [1] 3

`ps_cpu_times(p)` returns the CPU times of the process, similarly to
`proc.time()`.

``` r
ps_cpu_times(p)
```

    ##            user          system   children_user children_system 
    ##       1.7681955       0.6514954              NA              NA

`ps_memory_info(p)` returns memory usage information. See the manual for
details.

``` r
ps_memory_info(p)
```

    ##          rss          vms      pfaults      pageins 
    ##    261783552 421933465600        41422           13

`ps_children(p)` lists all child processes (potentially recursively) of
the current process.

``` r
ps_children(ps_parent(p))
```

    ## [[1]]
    ## <ps::ps_handle> PID=39874, NAME=zsh, AT=2025-04-28 07:53:00.889347
    ## 
    ## [[2]]
    ## <ps::ps_handle> PID=44844, NAME=R, AT=2025-04-28 08:01:34.873763

`ps_num_fds(p)` returns the number of open file descriptors (handles on
Windows):

``` r
ps_num_fds(p)
```

    ## [1] 23

``` r
f <- file(tmp <- tempfile(), "w")
ps_num_fds(p)
```

    ## [1] 24

``` r
close(f)
unlink(tmp)
```

`ps_open_files(p)` lists all open files:

``` r
ps_open_files(p)
```

    ## # A data frame: 4 × 2
    ##      fd path                                                                              
    ##   <int> <chr>                                                                             
    ## 1     0 /dev/ttys019                                                                      
    ## 2     1 /dev/ttys019                                                                      
    ## 3     2 /dev/ttys019                                                                      
    ## 4    14 /private/var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpoaXwPG/Rfaf2c3705b93c

``` r
f <- file(tmp <- tempfile(), "w")
ps_open_files(p)
```

    ## # A data frame: 5 × 2
    ##      fd path                                                                                
    ##   <int> <chr>                                                                               
    ## 1     0 /dev/ttys019                                                                        
    ## 2     1 /dev/ttys019                                                                        
    ## 3     2 /dev/ttys019                                                                        
    ## 4    14 /private/var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpoaXwPG/Rfaf2c3705b93c  
    ## 5    23 /private/var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpoaXwPG/fileaf2c58c42e5c

``` r
close(f)
unlink(tmp)
ps_open_files(p)
```

    ## # A data frame: 4 × 2
    ##      fd path                                                                              
    ##   <int> <chr>                                                                             
    ## 1     0 /dev/ttys019                                                                      
    ## 2     1 /dev/ttys019                                                                      
    ## 3     2 /dev/ttys019                                                                      
    ## 4    14 /private/var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpoaXwPG/Rfaf2c3705b93c

### Process manipulation

`ps_suspend(p)` suspends (stops) the process. On POSIX it sends a
SIGSTOP signal. On Windows it stops all threads.

`ps_resume(p)` resumes the process. On POSIX it sends a SIGCONT signal.
On Windows it resumes all stopped threads.

`ps_send_signal(p)` sends a signal to the process. It is implemented on
POSIX systems only. It makes an effort to work around pid reuse.

`ps_terminate(p)` send SIGTERM to the process. On POSIX systems only.

`ps_kill(p)` terminates the process. Sends `SIGKILL` on POSIX systems,
uses `TerminateProcess()` on Windows. It make an effort to work around
pid reuse.

`ps_interrupt(p)` interrupts a process. It sends a `SIGINT` signal on
POSIX systems, and it can send a CTRL+C or a CTRL+BREAK event on
Windows.

## Finished and zombie processes

ps handles finished and Zombie processes as much as possible.

The essential `ps_pid()`, `ps_create_time()`, `ps_is_running()`
functions and the `format()` and `print()` methods work for all
processes, including finished and zombie processes. Other functions fail
with an error of class `"no_such_process"` for finished processes.

The `ps_ppid()`, `ps_parent()`, `ps_children()`, `ps_name()`,
`ps_status()`, `ps_username()`, `ps_uids()`, `ps_gids()`,
`ps_terminal()`, `ps_children()` and the signal sending functions work
properly for zombie processes. Other functions fail with
`"zombie_process"` error.

## Pid reuse

ps functions handle pid reuse as well as technically possible.

The query functions never return information about the wrong process,
even if the process has finished and its process id was re-assigned.

On Windows, the process manipulation functions never manipulate the
wrong process.

On POSIX systems, this is technically impossible, it is not possible to
send a signal to a process without creating a race condition. In ps the
time window of the race condition is very small, a few microseconds, and
the process would need to finish, *and* the OS would need to reuse its
pid within this time window to create problems. This is very unlikely to
happen.

## Recipes

In the spirit of [psutil
recipes](http://psutil.readthedocs.io/en/latest/#recipes).

### Find process by name

Using `ps()` and dplyr:

``` r
library(dplyr)
find_procs_by_name <- function(name) {
  ps() |>
    filter(name == !!name)  |>
    pull(ps_handle)
}

find_procs_by_name("R")
```

    ## [[1]]
    ## <ps::ps_handle> PID=44844, NAME=R, AT=2025-04-28 08:01:34.873763
    ## 
    ## [[2]]
    ## <ps::ps_handle> PID=32722, NAME=R, AT=2025-04-28 07:36:42.295781

Without creating the full table of processes:

``` r
find_procs_by_name <- function(name) {
  procs <- lapply(ps_pids(), function(p) {
    tryCatch({
      h <- ps_handle(p)
      if (ps_name(h) == name) h else NULL },
      no_such_process = function(e) NULL,
      access_denied = function(e) NULL
    )
  })
  procs[!vapply(procs, is.null, logical(1))]
  }

find_procs_by_name("R")
```

    ## [[1]]
    ## <ps::ps_handle> PID=32722, NAME=R, AT=2025-04-28 07:36:42.295781
    ## 
    ## [[2]]
    ## <ps::ps_handle> PID=44844, NAME=R, AT=2025-04-28 08:01:34.873763

### Wait for a process to finish

`ps_wait()`, from ps 1.8.0, implements a new way, efficient for waiting
on a list of processes, so this is now very easy:

``` r
px <- processx::process$new("sleep", "2")
p <- px$as_ps_handle()
ps_wait(p, 1000)
```

    ## [1] FALSE

``` r
ps_wait(p)
```

    ## [1] TRUE

### Wait for several processes to finish

Again, this is much simpler with `ps_wait()`, added in ps 1.8.0.

``` r
px1 <- processx::process$new("sleep", "10")
px2 <- processx::process$new("sleep", "10")
px3 <- processx::process$new("sleep", "1")
px4 <- processx::process$new("sleep", "1")

p1 <- px1$as_ps_handle()
p2 <- px2$as_ps_handle()
p3 <- px3$as_ps_handle()
p4 <- px4$as_ps_handle()

ps_wait(list(p1, p2, p3, p4), timeout = 2000)
```

    ## [1] FALSE FALSE  TRUE  TRUE

### Kill process tree

From ps 1.8.0, `ps_kill()` will first send `SIGTERM` signals on Unix,
and `SIGKILL` after a grace period, if needed.

Note, that some R IDEs, including RStudio, run a multithreaded R
process, and other threads may start processes as well.
`reap_children()` will clean up all these as well, potentially causing
the IDE to misbehave or crash.

``` r
kill_proc_tree <- function(pid, include_parent = TRUE, ...) {
  if (pid == Sys.getpid() && include_parent) stop("I refuse to kill myself")
  parent <- ps_handle(pid)
  children <- ps_children(parent, recursive = TRUE)
  if (include_parent) children <- c(children, list(parent))
  ps_kill(children, ...)
}

p1 <- processx::process$new("sleep", "10")
p2 <- processx::process$new("sleep", "10")
p3 <- processx::process$new("sleep", "10")
kill_proc_tree(Sys.getpid(), include_parent = FALSE)
```

    ## [1] "terminated" "terminated" "terminated" "terminated" "terminated"

### Filtering and sorting processes

Process name ending with “sh”:

``` r
ps() |>
  filter(grepl("sh$", name))
```

    ## # A data frame: 35 × 11
    ##      pid  ppid name        username    status      user  system     rss      vms created            
    ##    <int> <int> <chr>       <chr>       <chr>      <dbl>   <dbl>   <dbl>    <dbl> <dttm>             
    ##  1 41838     1 ReportCrash root        sleepi… NA       NA      NA      NA       2025-04-28 07:57:52
    ##  2 40290 40090 zsh         gaborcsardi sleepi…  0.0361   0.154   3.01e6  4.21e11 2025-04-28 07:53:47
    ##  3 40090 40088 zsh         gaborcsardi sleepi…  0.587    0.406   3.14e7  4.21e11 2025-04-28 07:53:47
    ##  4 39874 39673 zsh         gaborcsardi sleepi…  0.00552  0.0208  2.62e6  4.21e11 2025-04-28 07:53:00
    ##  5 39673 39672 zsh         gaborcsardi sleepi…  0.270    0.0922  2.96e7  4.21e11 2025-04-28 07:53:00
    ##  6 35995 35795 zsh         gaborcsardi sleepi…  0.0117   0.0485  9.67e5  4.21e11 2025-04-28 07:45:03
    ##  7 35795 35794 zsh         gaborcsardi sleepi…  0.236    0.128   1.69e6  4.21e11 2025-04-28 07:45:03
    ##  8 35595 35391 zsh         gaborcsardi sleepi…  0.00781  0.0332  9.67e5  4.21e11 2025-04-28 07:44:16
    ##  9 35391 35390 zsh         gaborcsardi sleepi…  0.262    0.115   1.69e6  4.21e11 2025-04-28 07:44:15
    ## 10 30303 30302 bash        gaborcsardi sleepi…  0.00806  0.0194  8.68e5  4.20e11 2025-04-28 07:32:12
    ## # ℹ 25 more rows
    ## # ℹ 1 more variable: ps_handle <I<list>>

Processes owned by user:

``` r
ps() |>
  filter(username == Sys.info()[["user"]]) |>
  select(pid, name)
```

    ## # A data frame: 347 × 2
    ##      pid name                           
    ##    <int> <chr>                          
    ##  1 45911 Google Chrome Helper (Renderer)
    ##  2 45860 Google Chrome Helper (Renderer)
    ##  3 45700 mdworker_shared                
    ##  4 44844 R                              
    ##  5 44738 Google Chrome Helper           
    ##  6 43786 MessagesBlastDoorService       
    ##  7 43785 businessservicesd              
    ##  8 43628 MENotificationAgent            
    ##  9 43627 MTLAssetUpgraderD              
    ## 10 43472 replayd                        
    ## # ℹ 337 more rows

Processes consuming more than 100MB of memory:

``` r
ps() |>
  filter(rss > 100 * 1024 * 1024)
```

    ## # A data frame: 20 × 11
    ##      pid  ppid name    username status    user  system    rss     vms created             ps_handle 
    ##    <int> <int> <chr>   <chr>    <chr>    <dbl>   <dbl>  <dbl>   <dbl> <dttm>              <I<list>> 
    ##  1 45911  4159 Google… gaborcs… sleep… 4.73e-2 2.27e-2 1.05e8 1.91e12 2025-04-28 08:05:25 <ps_handl>
    ##  2 45860  4159 Google… gaborcs… sleep… 7.51e-2 3.47e-2 1.21e8 1.91e12 2025-04-28 08:05:19 <ps_handl>
    ##  3 44844 39673 R       gaborcs… runni… 2.08e+0 9.20e-1 2.80e8 4.22e11 2025-04-28 08:01:34 <ps_handl>
    ##  4 40521  4159 Google… gaborcs… sleep… 3.14e+0 4.45e-1 2.03e8 1.91e12 2025-04-28 07:54:07 <ps_handl>
    ##  5 40037  4159 Google… gaborcs… sleep… 7.78e+0 1.03e+0 2.66e8 1.91e12 2025-04-28 07:53:23 <ps_handl>
    ##  6 39666  4159 Google… gaborcs… sleep… 1.30e+0 2.37e-1 1.45e8 1.91e12 2025-04-28 07:52:34 <ps_handl>
    ##  7 28176  4159 Google… gaborcs… sleep… 1.16e+1 1.09e+0 2.18e8 1.91e12 2025-04-28 07:23:37 <ps_handl>
    ##  8 18695  4159 Google… gaborcs… sleep… 2.33e+1 2.18e+0 2.23e8 1.91e12 2025-04-28 06:48:50 <ps_handl>
    ##  9 18673  4159 Google… gaborcs… sleep… 2.08e+1 1.73e+0 2.39e8 1.91e12 2025-04-28 06:47:44 <ps_handl>
    ## 10 18257  4159 Google… gaborcs… sleep… 2.11e+1 2.85e+0 2.02e8 1.91e12 2025-04-28 06:31:40 <ps_handl>
    ## 11 17380  4159 Google… gaborcs… sleep… 1.25e+1 1.50e+0 1.76e8 1.91e12 2025-04-28 05:48:24 <ps_handl>
    ## 12 16963  4159 Google… gaborcs… sleep… 4.73e+2 6.83e+1 5.52e8 1.91e12 2025-04-28 05:31:43 <ps_handl>
    ## 13  4214  4159 Google… gaborcs… sleep… 4.88e+0 1.34e+0 1.34e8 1.91e12 2025-04-26 23:27:03 <ps_handl>
    ## 14  4199  4159 Google… gaborcs… sleep… 1.26e+2 1.08e+1 3.45e8 1.91e12 2025-04-26 23:26:59 <ps_handl>
    ## 15  4176  4159 Google… gaborcs… sleep… 2.38e+2 2.31e+2 1.23e8 4.55e11 2025-04-26 23:26:58 <ps_handl>
    ## 16  4175  4159 Google… gaborcs… sleep… 2.05e+3 1.09e+3 1.35e8 4.56e11 2025-04-26 23:26:58 <ps_handl>
    ## 17  4159     1 Google… gaborcs… sleep… 1.21e+3 3.79e+2 6.00e8 4.56e11 2025-04-26 23:26:54 <ps_handl>
    ## 18 22365 22350 qemu-s… gaborcs… sleep… 1.26e+4 1.65e+3 1.58e8 4.26e11 2025-04-24 10:02:15 <ps_handl>
    ## 19  1726     1 Spotli… gaborcs… sleep… 1.64e+2 4.42e+1 1.47e8 4.25e11 2025-04-22 14:34:30 <ps_handl>
    ## 20  1644     1 iTerm2  gaborcs… sleep… 6.32e+3 1.32e+3 3.87e8 4.23e11 2025-04-22 14:34:22 <ps_handl>

Top 3 memory consuming processes:

``` r
ps() |>
  top_n(3, rss) |>
  arrange(desc(rss))
```

    ## # A data frame: 3 × 11
    ##     pid  ppid name        username status  user system    rss     vms created             ps_handle 
    ##   <int> <int> <chr>       <chr>    <chr>  <dbl>  <dbl>  <dbl>   <dbl> <dttm>              <I<list>> 
    ## 1  4159     1 Google Chr… gaborcs… sleep… 1211.  379.  6.00e8 4.56e11 2025-04-26 23:26:54 <ps_handl>
    ## 2 16963  4159 Google Chr… gaborcs… sleep…  473.   68.3 5.52e8 1.91e12 2025-04-28 05:31:43 <ps_handl>
    ## 3  1644     1 iTerm2      gaborcs… sleep… 6316. 1322.  3.87e8 4.23e11 2025-04-22 14:34:22 <ps_handl>

Top 3 processes which consumed the most CPU time:

``` r
ps() |>
  mutate(cpu_time = user + system) |>
  top_n(3, cpu_time) |>
  arrange(desc(cpu_time)) |>
  select(pid, name, cpu_time)
```

    ## # A data frame: 3 × 3
    ##     pid name                       cpu_time
    ##   <int> <chr>                         <dbl>
    ## 1 22365 qemu-system-aarch64          14269.
    ## 2  1644 iTerm2                        7639.
    ## 3  4175 Google Chrome Helper (GPU)    3140.

## Code of Conduct

Please note that the ps project is released with a [Contributor Code of
Conduct](https://ps.r-lib.org/CODE_OF_CONDUCT.html). By contributing to
this project, you agree to abide by its terms.

## License

MIT © RStudio

winver_ver <- function(v = NULL) {
  if (is.null(v)) {
    v <- system("cmd /c ver", intern = TRUE)
  }
  v2 <- grep("\\[.*\\s.*\\]", v, value = TRUE)[1]
  v3 <- sub("^.*\\[[^ ]+\\s+", "", v2)
  v4 <- sub("\\]$", "", v3)
  if (is.na(v4)) {
    stop("Failed to parse windows version")
  }
  v4
}

winver_wmic <- function(v = NULL) {
  cmd <- "wmic os get Version /value"
  if (is.null(v)) {
    v <- system(cmd, intern = TRUE)
  }
  v2 <- grep("=", v, value = TRUE)
  v3 <- strsplit(v2, "=", fixed = TRUE)[[1]][2]
  v4 <- sub("\\s*$", "", sub("^\\s*", "", v3))
  if (is.na(v4)) {
    stop("Failed to parse windows version")
  }
  v4
}

winver <- function() {
  ## First we try with `wmic`
  v <- if (Sys.which("wmic") != "") {
    tryCatch(winver_wmic(), error = function(e) NULL)
  }
  ## Otherwise `ver`
  if (is.null(v)) winver_ver() else v
}

if (is.null(sys.calls())) {
  cat(winver())
}

test_that("winver_ver", {
  cases <- list(
    list(c("", "Microsoft Windows [Version 6.3.9600]"), "6.3.9600"),
    list("Microsoft Windows [version 6.1.7601]", "6.1.7601"),
    list("Microsoft Windows [vers\u00e3o 10.0.18362.207]", "10.0.18362.207"))

  source(system.file("tools", "winver.R", package = "ps"), local = TRUE)
  
  for (x in cases) expect_identical(winver_ver(x[[1]]), x[[2]])
})

test_that("winver_wmic", {
  cases <- list(
    list(c("\r", "\r", "Version=6.3.9600\r", "\r", "\r", "\r"),
         "6.3.9600"),
    list(c("\r", "\r", "version=6.3.9600\r", "\r", "\r", "\r"),
         "6.3.9600"),
    list(c("\r", "\r", "vers\u00e3o=6.3.9600\r", "\r", "\r", "\r"),
         "6.3.9600"))    

  source(system.file("tools", "winver.R", package = "ps"), local = TRUE)
  
  for (x in cases) expect_identical(winver_wmic(x[[1]]), x[[2]])
})

library(testthat)
library(ps)

if (ps::ps_is_supported() && Sys.getenv("R_COVR", "") != "true" &&
    Sys.getenv("NOT_CRAN") != "" ) {
  reporter <- ps::CleanupReporter(testthat::SummaryReporter)$new()
} else {
  reporter <- "summary"
}

if (ps_is_supported() && Sys.getenv("NOT_CRAN") != "") {
  test_check("ps", reporter = reporter)
}

progs <- if (WINDOWS) {
           c("px.exe", "interrupt.exe")
         } else {
           "px"
         }

dest <- file.path(R_PACKAGE_DIR, paste0("bin", R_ARCH))
dir.create(dest, recursive = TRUE, showWarnings = FALSE)
file.copy(progs, dest, overwrite = TRUE)

files <- Sys.glob(paste0("*", SHLIB_EXT))
dest <- file.path(R_PACKAGE_DIR, paste0('libs', R_ARCH))
dir.create(dest, recursive = TRUE, showWarnings = FALSE)
file.copy(files, dest, overwrite = TRUE)
if (file.exists("symbols.rds")) {
  file.copy("symbols.rds", dest, overwrite = TRUE)
}

## nocov start
memoize <- function(fun) {
  fun
  cache <- NULL
  if (length(formals(fun)) > 0) {
    stop("Only memoizing functions without arguments")
  }
  dec <- function() {
    if (is.null(cache)) cache <<- fun()
    cache
  }
  attr(dec, "clear") <- function() cache <<- TRUE
  class(dec) <- c("memoize", class(dec))
  dec
}

`$.memoize`  <- function(x, name) {
  switch(
    name,
    "clear" = attr(x, "clear"),
    stop("unknown memoize method")
  )
}
## nocov end

format_regexp <- function() {
  "<ps::ps_handle> PID=[0-9]+, NAME=.*, AT="
}

parse_ps <- function(args) {
  out <- processx::run("ps", args)$stdout
  sub(" *$", "", strsplit(out, "\n")[[1]][[2]])
}

parse_time <- function(x) {
  x <- utils::tail(c(0, 0, 0, as.numeric(strsplit(x, ":")[[1]])), 3)
  x[1] * 60 * 60 + x[2] * 60 + x[3]
}

wait_for_status <- function(ps, status, timeout = 5) {
  limit <- Sys.time() + timeout
  while (ps_status(ps) != status && Sys.time() < limit) Sys.sleep(0.05)
}

px <- function() get_tool("px")

skip_in_rstudio <- function() {
  if (Sys.getenv("RSTUDIO") != "") skip("Cannot test in RStudio")
}

has_processx <- function() {
  requireNamespace("processx", quietly = TRUE) &&
    package_version(getNamespaceVersion("processx")) >= "3.1.0.9005"
}

skip_if_no_processx <- function() {
  if (!has_processx()) skip("Needs processx >= 3.1.0.9005 to run")
}

skip_without_program <- function(prog) {
  if (Sys.which(prog) == "") skip(paste(prog, "is not available"))
}

have_ipv6_support <- function() {
  ps_os_type()[["WINDOWS"]] ||
    !is.null(ps_env$constants$address_families$AF_INET6)
}

skip_without_ipv6 <- function() {
  if (!have_ipv6_support()) skip("Needs IPv6")
}

ipv6_url <- function() {
  paste0("https://", ipv6_host())
}

ipv6_host <- function() {
  "ipv6.test-ipv6.com"
}

have_ipv6_connection <- local({
  ok <- NULL
  myurl <- NULL
  function(url = ipv6_url()) {
    if (is.null(ok) || myurl != url) {
      myurl <<- url
      opt <- options(warn = 2)
      on.exit(options(opt), add = TRUE)
      tryCatch({
        cx <- curl::curl(url)
        open(cx)
        ok <<- TRUE
      },
      error = function(x) ok <<- FALSE,
      finally = close(cx))
    }
    ok
  }
})

skip_without_ipv6_connection <- function() {
  if (!have_ipv6_connection()) skip("Needs working IPv6 connection")
}

wait_for_string <- function(proc, string, timeout) {
  deadline <- Sys.time() + as.difftime(timeout / 1000, units = "secs")
  str <- ""
  repeat {
    left <- max(as.double(deadline - Sys.time(), units = "secs"), 0)
    pr <- processx::poll(list(proc), as.integer(left * 1000))
    str <- paste(str, proc$read_error())
    if (grepl(string, str)) return()
    if (proc$has_output_connection()) read_output()
    if (deadline < Sys.time()) stop("Cannot start proces")
    if (!proc$is_alive()) stop("Cannot start process")
  }
}

## This is not perfect, e.g. we don't check that the numbers are <255,
## but will do for our purposes

is_ipv4_address <- function(x) {
  grepl("^[0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+$", x)
}

cleanup_process <- function(p) {
  tryCatch(close(p$get_input_connection()), error = function(x) x)
  tryCatch(close(p$get_output_connection()), error = function(x) x)
  tryCatch(close(p$get_error_connection()), error = function(x) x)
  tryCatch(close(p$get_poll_connection()), error = function(x) x)
  tryCatch(p$kill(), error = function(x) x)
}

httpbin <- webfakes::new_app_process(
  webfakes::httpbin_app(),
  opts = webfakes::server_opts(num_threads = 6)
)

# nocov start

compat_vctrs <- local({
  # Modified from https://github.com/r-lib/rlang/blob/master/R/compat-vctrs.R

  # Construction ------------------------------------------------------------

  # Constructs data frames inheriting from `"tbl"`. This allows the
  # pillar package to take over printing as soon as it is loaded.
  # The data frame otherwise behaves like a base data frame.
  data_frame <- function(...) {
    new_data_frame(df_list(...), .class = "tbl")
  }

  new_data_frame <- function(.x = list(), ..., .size = NULL, .class = NULL) {
    n_cols <- length(.x)
    if (n_cols != 0 && is.null(names(.x))) {
      stop("Columns must be named.", call. = FALSE)
    }

    if (is.null(.size)) {
      if (n_cols == 0) {
        .size <- 0
      } else {
        .size <- vec_size(.x[[1]])
      }
    }

    structure(
      .x,
      class = c(.class, "data.frame"),
      row.names = .set_row_names(.size),
      ...
    )
  }

  df_list <- function(..., .size = NULL) {
    vec_recycle_common(list(...), size = .size)
  }

  # Binding -----------------------------------------------------------------

  vec_rbind <- function(...) {
    xs <- vec_cast_common(list(...))
    do.call(base::rbind, xs)
  }

  vec_cbind <- function(...) {
    xs <- list(...)

    ptype <- vec_ptype_common(lapply(xs, `[`, 0))
    class <- setdiff(class(ptype), "data.frame")

    xs <- vec_recycle_common(xs)
    out <- do.call(base::cbind, xs)
    new_data_frame(out, .class = class)
  }

  # Slicing -----------------------------------------------------------------

  vec_size <- function(x) {
    if (is.data.frame(x)) {
      nrow(x)
    } else {
      length(x)
    }
  }

  vec_rep <- function(x, times) {
    i <- rep.int(seq_len(vec_size(x)), times)
    vec_slice(x, i)
  }

  vec_recycle_common <- function(xs, size = NULL) {
    sizes <- vapply(xs, vec_size, integer(1))

    n <- unique(sizes)

    if (length(n) == 1 && is.null(size)) {
      return(xs)
    }
    n <- setdiff(n, 1L)

    ns <- length(n)

    if (ns == 0) {
      if (is.null(size)) {
        return(xs)
      }
    } else if (ns == 1) {
      if (is.null(size)) {
        size <- n
      } else if (ns != size) {
        stop("Inputs can't be recycled to `size`.", call. = FALSE)
      }
    } else {
      stop("Inputs can't be recycled to a common size.", call. = FALSE)
    }

    to_recycle <- sizes == 1L
    xs[to_recycle] <- lapply(xs[to_recycle], vec_rep, size)

    xs
  }

  vec_slice <- function(x, i) {
    if (is.logical(i)) {
      i <- which(i)
    }
    stopifnot(is.numeric(i) || is.character(i))

    if (is.null(x)) {
      return(NULL)
    }

    if (is.data.frame(x)) {
      # We need to be a bit careful to be generic. First empty all
      # columns and expand the df to final size.
      out <- x[i, 0, drop = FALSE]

      # Then fill in with sliced columns
      out[seq_along(x)] <- lapply(x, vec_slice, i)

      # Reset automatic row names to work around `[` weirdness
      if (is.numeric(attr(x, "row.names"))) {
        row_names <- .set_row_names(nrow(out))
      } else {
        row_names <- attr(out, "row.names")
      }

      return(out)
    }

    d <- vec_dims(x)
    if (d == 1) {
      if (is.object(x)) {
        out <- x[i]
      } else {
        out <- x[i, drop = FALSE]
      }
    } else if (d == 2) {
      out <- x[i, , drop = FALSE]
    } else {
      j <- rep(list(quote(expr = )), d - 1)
      out <- eval(as.call(list(
        quote(`[`),
        quote(x),
        quote(i),
        j,
        drop = FALSE
      )))
    }

    out
  }
  vec_dims <- function(x) {
    d <- dim(x)
    if (is.null(d)) {
      1L
    } else {
      length(d)
    }
  }

  vec_as_location <- function(i, n, names = NULL) {
    out <- seq_len(n)
    names(out) <- names

    # Special-case recycling to size 0
    if (is_logical(i, n = 1) && !length(out)) {
      return(out)
    }

    unname(out[i])
  }

  vec_init <- function(x, n = 1L) {
    vec_slice(x, rep_len(NA_integer_, n))
  }

  vec_assign <- function(x, i, value) {
    if (is.null(x)) {
      return(NULL)
    }

    if (is.logical(i)) {
      i <- which(i)
    }
    stopifnot(
      is.numeric(i) || is.character(i)
    )

    value <- vec_recycle(value, vec_size(i))
    value <- vec_cast(value, to = x)

    d <- vec_dims(x)

    if (d == 1) {
      x[i] <- value
    } else if (d == 2) {
      x[i, ] <- value
    } else {
      stop("Can't slice-assign arrays.", call. = FALSE)
    }

    x
  }

  vec_recycle <- function(x, size) {
    if (is.null(x) || is.null(size)) {
      return(NULL)
    }

    n_x <- vec_size(x)

    if (n_x == size) {
      x
    } else if (size == 0L) {
      vec_slice(x, 0L)
    } else if (n_x == 1L) {
      vec_slice(x, rep(1L, size))
    } else {
      stop("Incompatible lengths: ", n_x, ", ", size, call. = FALSE)
    }
  }

  # Coercion ----------------------------------------------------------------

  vec_cast_common <- function(xs, to = NULL) {
    ptype <- vec_ptype_common(xs, ptype = to)
    lapply(xs, vec_cast, to = ptype)
  }

  vec_cast <- function(x, to) {
    if (is.null(x)) {
      return(NULL)
    }
    if (is.null(to)) {
      return(x)
    }

    if (vec_is_unspecified(x)) {
      return(vec_init(to, vec_size(x)))
    }

    stop_incompatible_cast <- function(x, to) {
      stop(
        sprintf(
          "Can't convert <%s> to <%s>.",
          .rlang_vctrs_typeof(x),
          .rlang_vctrs_typeof(to)
        ),
        call. = FALSE
      )
    }

    lgl_cast <- function(x, to) {
      lgl_cast_from_num <- function(x) {
        if (any(!x %in% c(0L, 1L))) {
          stop_incompatible_cast(x, to)
        }
        as.logical(x)
      }

      switch(
        .rlang_vctrs_typeof(x),
        logical = x,
        integer = ,
        double = lgl_cast_from_num(x),
        stop_incompatible_cast(x, to)
      )
    }

    int_cast <- function(x, to) {
      int_cast_from_dbl <- function(x) {
        out <- suppressWarnings(as.integer(x))
        if (any((out != x) | xor(is.na(x), is.na(out)))) {
          stop_incompatible_cast(x, to)
        } else {
          out
        }
      }

      switch(
        .rlang_vctrs_typeof(x),
        logical = as.integer(x),
        integer = x,
        double = int_cast_from_dbl(x),
        stop_incompatible_cast(x, to)
      )
    }

    dbl_cast <- function(x, to) {
      switch(
        .rlang_vctrs_typeof(x),
        logical = ,
        integer = as.double(x),
        double = x,
        stop_incompatible_cast(x, to)
      )
    }

    chr_cast <- function(x, to) {
      switch(
        .rlang_vctrs_typeof(x),
        character = x,
        stop_incompatible_cast(x, to)
      )
    }

    list_cast <- function(x, to) {
      switch(
        .rlang_vctrs_typeof(x),
        list = x,
        stop_incompatible_cast(x, to)
      )
    }

    df_cast <- function(x, to) {
      # Check for extra columns
      if (length(setdiff(names(x), names(to))) > 0) {
        stop(
          "Can't convert data frame because of missing columns.",
          call. = FALSE
        )
      }

      # Avoid expensive [.data.frame method
      out <- as.list(x)

      # Coerce common columns
      common <- intersect(names(x), names(to))
      out[common] <- Map(vec_cast, out[common], to[common])

      # Add new columns
      from_type <- setdiff(names(to), names(x))
      out[from_type] <- lapply(to[from_type], vec_init, n = vec_size(x))

      # Ensure columns are ordered according to `to`
      out <- out[names(to)]

      new_data_frame(out)
    }

    rlib_df_cast <- function(x, to) {
      new_data_frame(df_cast(x, to), .class = "tbl")
    }
    tib_cast <- function(x, to) {
      new_data_frame(df_cast(x, to), .class = c("tbl_df", "tbl"))
    }

    switch(
      .rlang_vctrs_typeof(to),
      logical = lgl_cast(x, to),
      integer = int_cast(x, to),
      double = dbl_cast(x, to),
      character = chr_cast(x, to),
      list = list_cast(x, to),

      base_data_frame = df_cast(x, to),
      rlib_data_frame = rlib_df_cast(x, to),
      tibble = tib_cast(x, to),

      stop_incompatible_cast(x, to)
    )
  }

  vec_ptype_common <- function(xs, ptype = NULL) {
    if (!is.null(ptype)) {
      return(vec_ptype(ptype))
    }

    xs <- Filter(function(x) !is.null(x), xs)

    if (length(xs) == 0) {
      return(NULL)
    }

    if (length(xs) == 1) {
      out <- vec_ptype(xs[[1]])
    } else {
      xs <- map(xs, vec_ptype)
      out <- Reduce(vec_ptype2, xs)
    }

    vec_ptype_finalise(out)
  }

  vec_ptype_finalise <- function(x) {
    if (is.data.frame(x)) {
      x[] <- lapply(x, vec_ptype_finalise)
      return(x)
    }

    if (inherits(x, "rlang_unspecified")) {
      logical()
    } else {
      x
    }
  }

  vec_ptype <- function(x) {
    if (vec_is_unspecified(x)) {
      return(.rlang_vctrs_unspecified())
    }

    if (is.data.frame(x)) {
      out <- new_data_frame(lapply(x, vec_ptype))

      attrib <- attributes(x)
      attrib$row.names <- attr(out, "row.names")
      attributes(out) <- attrib

      return(out)
    }

    vec_slice(x, 0)
  }

  vec_ptype2 <- function(x, y) {
    stop_incompatible_type <- function(x, y) {
      stop(
        sprintf(
          "Can't combine types <%s> and <%s>.",
          .rlang_vctrs_typeof(x),
          .rlang_vctrs_typeof(y)
        ),
        call. = FALSE
      )
    }

    x_type <- .rlang_vctrs_typeof(x)
    y_type <- .rlang_vctrs_typeof(y)

    if (x_type == "unspecified" && y_type == "unspecified") {
      return(.rlang_vctrs_unspecified())
    }
    if (x_type == "unspecified") {
      return(y)
    }
    if (y_type == "unspecified") {
      return(x)
    }

    df_ptype2 <- function(x, y) {
      set_partition <- function(x, y) {
        list(
          both = intersect(x, y),
          only_x = setdiff(x, y),
          only_y = setdiff(y, x)
        )
      }

      # Avoid expensive [.data.frame
      x <- as.list(vec_slice(x, 0))
      y <- as.list(vec_slice(y, 0))

      # Find column types
      names <- set_partition(names(x), names(y))
      if (length(names$both) > 0) {
        common_types <- Map(vec_ptype2, x[names$both], y[names$both])
      } else {
        common_types <- list()
      }
      only_x_types <- x[names$only_x]
      only_y_types <- y[names$only_y]

      # Combine and construct
      out <- c(common_types, only_x_types, only_y_types)
      out <- out[c(names(x), names$only_y)]
      new_data_frame(out)
    }

    rlib_df_ptype2 <- function(x, y) {
      new_data_frame(df_ptype2(x, y), .class = "tbl")
    }
    tib_ptype2 <- function(x, y) {
      new_data_frame(df_ptype2(x, y), .class = c("tbl_df", "tbl"))
    }

    ptype <- switch(
      x_type,

      logical = switch(
        y_type,
        logical = x,
        integer = y,
        double = y,
        stop_incompatible_type(x, y)
      ),

      integer = switch(
        .rlang_vctrs_typeof(y),
        logical = x,
        integer = x,
        double = y,
        stop_incompatible_type(x, y)
      ),

      double = switch(
        .rlang_vctrs_typeof(y),
        logical = x,
        integer = x,
        double = x,
        stop_incompatible_type(x, y)
      ),

      character = switch(
        .rlang_vctrs_typeof(y),
        character = x,
        stop_incompatible_type(x, y)
      ),

      list = switch(
        .rlang_vctrs_typeof(y),
        list = x,
        stop_incompatible_type(x, y)
      ),

      base_data_frame = switch(
        .rlang_vctrs_typeof(y),
        base_data_frame = ,
        s3_data_frame = df_ptype2(x, y),
        rlib_data_frame = rlib_df_ptype2(x, y),
        tibble = tib_ptype2(x, y),
        stop_incompatible_type(x, y)
      ),

      rlib_data_frame = switch(
        .rlang_vctrs_typeof(y),
        base_data_frame = ,
        rlib_data_frame = ,
        s3_data_frame = rlib_df_ptype2(x, y),
        tibble = tib_ptype2(x, y),
        stop_incompatible_type(x, y)
      ),

      tibble = switch(
        .rlang_vctrs_typeof(y),
        base_data_frame = ,
        rlib_data_frame = ,
        tibble = ,
        s3_data_frame = tib_ptype2(x, y),
        stop_incompatible_type(x, y)
      ),

      stop_incompatible_type(x, y)
    )

    vec_slice(ptype, 0)
  }

  .rlang_vctrs_typeof <- function(x) {
    if (is.object(x)) {
      class <- class(x)

      if (identical(class, "rlang_unspecified")) {
        return("unspecified")
      }
      if (identical(class, "data.frame")) {
        return("base_data_frame")
      }
      if (identical(class, c("tbl", "data.frame"))) {
        return("rlib_data_frame")
      }
      if (identical(class, c("tbl_df", "tbl", "data.frame"))) {
        return("tibble")
      }
      if (inherits(x, "data.frame")) {
        return("s3_data_frame")
      }

      class <- paste0(class, collapse = "/")
      stop(sprintf("Unimplemented class <%s>.", class), call. = FALSE)
    }

    type <- typeof(x)
    switch(
      type,
      NULL = return("null"),
      logical = if (vec_is_unspecified(x)) {
        return("unspecified")
      } else {
        return(type)
      },
      integer = ,
      double = ,
      character = ,
      raw = ,
      list = return(type)
    )

    stop(sprintf("Unimplemented type <%s>.", type), call. = FALSE)
  }

  vec_is_unspecified <- function(x) {
    !is.object(x) &&
      typeof(x) == "logical" &&
      length(x) &&
      all(vapply(x, identical, logical(1), NA))
  }

  .rlang_vctrs_unspecified <- function(x = NULL) {
    structure(
      rep(NA, length(x)),
      class = "rlang_unspecified"
    )
  }

  .rlang_vctrs_s3_method <- function(generic, class, env = parent.frame()) {
    fn <- get(generic, envir = env)

    ns <- asNamespace(topenv(fn))
    tbl <- ns$.__S3MethodsTable__.

    for (c in class) {
      name <- paste0(generic, ".", c)
      if (exists(name, envir = tbl, inherits = FALSE)) {
        return(get(name, envir = tbl))
      }
      if (exists(name, envir = globalenv(), inherits = FALSE)) {
        return(get(name, envir = globalenv()))
      }
    }

    NULL
  }

  environment()
})

data_frame <- compat_vctrs$data_frame

as_data_frame <- function(x) {
  if (is.matrix(x)) {
    x <- as.data.frame(x, stringsAsFactors = FALSE)
  } else {
    x <- compat_vctrs$vec_recycle_common(x)
  }
  compat_vctrs$new_data_frame(x, .class = "tbl")
}

# nocov end

#' @importFrom utils read.table

psl_connections <- function(p) {
  sock_raw <- not_null(.Call(psll_connections, p))
  sock <- data_frame(
    fd = as.integer(vapply(sock_raw, "[[", character(1), 1)),
    id = vapply(sock_raw, "[[", character(1), 2)
  )

  flt <- function(x, col, values) x[x[[col]] %in% values, ]

  unix <- flt(psl__read_table("/proc/net/unix"),         "V7",  sock$id)
  tcp  <- flt(psl__read_table("/proc/net/tcp") [, 1:10], "V10", sock$id)
  tcp6 <- flt(psl__read_table("/proc/net/tcp6")[, 1:10], "V10", sock$id)
  udp  <- flt(psl__read_table("/proc/net/udp") [, 1:10], "V10", sock$id)
  udp6 <- flt(psl__read_table("/proc/net/udp6")[, 1:10], "V10", sock$id)

  ## Sockets that still existed when we queried /proc/net,
  ## because some of them might be closed already...
  sockx <- flt(sock, "id",
               c(unix$V7, tcp$V10, tcp6$V10, udp$V10, udp6$V10))

  if (length(tcp) && nrow(tcp)) {
    tcp$type <- "SOCK_STREAM"
    tcp$family <- "AF_INET"
  }
  if (length(tcp6) && nrow(tcp6)) {
    tcp6$type <- "SOCK_STREAM"
    tcp6$family <- "AF_INET6"
  }
  if (length(udp) && nrow(udp)) {
    udp$type <- "SOCK_DGRAM"
    udp$family <- "AF_INET"
  }
  if (length(udp6) && nrow(udp6)) {
    udp6$type <- "SOCK_DGRAM"
    udp6$family <- "AF_INET6"
  }
  net <- rbind(tcp, tcp6, udp, udp6)

  ## Unix socket might or might not have a path
  if (length(unix) && nrow(unix)) {
    unix$V8 <- unix$V8 %||% ""
    unix$V8[unix$V8 == ""] <- NA_character_
  }

  ## The status column is 01...09, 0A, 0B, but R might parse it as integer
  if (length(unix) && nrow(unix)) {
    unix$V6 <- str_tail(paste0('0', as.character(unix$V6)), 2)
  }
  if (length(net) && nrow(net)) {
    net$V4 <- str_tail(paste0('0', as.character(net$V4)), 2)
  }

  d <- data_frame(
    fd = integer(),
    family = character(),
    type = character(),
    laddr = character(),
    lport = integer(),
    raddr = character(),
    rport = integer(),
    state = character()
  )

  if (length(unix) &&  nrow(unix)) {
    d <- data_frame(
      fd = sockx$fd[match(unix$V7, sockx$id)],
      family = "AF_UNIX",
      type = match_names(ps_env$constants$socket_types, unix$V5),
      laddr = unix$V8,
      lport = NA_integer_,
      raddr = NA_character_,
      rport = NA_integer_,
      state = NA_character_)
  }

  if (!is.null(net)) {

    laddr <- mapply(psl__decode_address, net$V2, net$family,
                    SIMPLIFY = FALSE, USE.NAMES = FALSE)
    raddr <- mapply(psl__decode_address, net$V3, net$family,
                    SIMPLIFY = FALSE, USE.NAMES = FALSE)

    net_d <- data_frame(
      fd = sockx$fd[match(net$V10, sockx$id)],
      family = net$family,
      type = net$type,
      laddr = vapply(laddr, "[[", character(1), 1),
      lport = vapply(laddr, "[[", integer(1), 2),
      raddr = vapply(raddr, "[[", character(1), 1),
      rport = vapply(raddr, "[[", integer(1), 2),
      state = match_names(ps_env$constants$tcp_statuses, net$V4))

    d <- rbind(d, net_d)
  }

  d
}

psl__read_table <- function(file, stringsAsFactors = FALSE, header = FALSE,
                            skip = 1, fill = TRUE, ...) {
  tryCatch(
    read.table(file, stringsAsFactors = stringsAsFactors, header = header,
               skip = skip, fill = fill, ...),
    error = function(e) NULL
  )
}

psl__decode_address <- function(addr, family) {
  ipp <- strsplit(addr, ":")[[1]]
  if (length(ipp) != 2) return(list(NA_character_, NA_integer_))
  addr <- str_strip(ipp[[1]])
  port <- strtoi(ipp[[2]], 16)

  if (family == "AF_INET") {
    AF_INET <- ps_env$constants$address_families[["AF_INET"]]
    addrn <- strtoi(substring(addr, 1:4*2-1, 1:4*2), base = 16)
    if (.Platform$endian == "little") addrn <- rev(addrn)
    addrs <- .Call(ps__inet_ntop, as.raw(addrn), AF_INET) %||% NA_character_
    list(addrs, port)

  } else {
    AF_INET6 <- ps_env$constants$address_families[["AF_INET6"]]
    addrn <- strtoi(substring(addr, 1:16*2-1, 1:16*2), base = 16)
    if (.Platform$endian == "little") {
      addrn[ 1: 4] <- rev(addrn[ 1: 4])
      addrn[ 5: 8] <- rev(addrn[ 5: 8])
      addrn[ 9:12] <- rev(addrn[ 9:12])
      addrn[13:16] <- rev(addrn[13:16])
    }
    addrs <- .Call(ps__inet_ntop, as.raw(addrn), AF_INET6) %||% NA_character_
    list(addrs, port)
  }
}

psl__cpu_count_from_lscpu <- function() {
  tryCatch({
    lines <- system("lscpu -p=core", intern = TRUE)
    cores <- unique(lines[!str_starts_with(lines, "#")])
    length(cores)
  }, error = function(e) psl__cpu_count_from_cpuinfo())
}

psl__cpu_count_from_cpuinfo <- function() {
  lines <- readLines("/proc/cpuinfo")
  mapping = list()
  current = list()

  for (l in lines) {
    l <- tolower(str_strip(l))
    if (!nchar(l)) {
      if ("physical id" %in% names(current) &&
          "cpu cores" %in% names(current)) {
        mapping[[ current[["physical id"]] ]] <- current[["cpu cores"]]
      }
      current <- list()

    } else {
      if (str_starts_with(l, "physical id") ||
          str_starts_with(l, "cpu cores")) {
        kv <- strsplit(l, "\\t+:")[[1]]
        current[[ kv[[1]] ]] <- kv[[2]]
      }
    }
  }

  sum(as.integer(unlist(mapping)))
}

ps_cpu_count_physical_linux <- function() {
  if (Sys.which("lscpu") != "") {
    psl__cpu_count_from_lscpu()
  } else {
    psl__cpu_count_from_cpuinfo()
  }
}

ps__system_cpu_times_linux <- function() {
  clock_ticks <- tryCatch(
    as.numeric(system("getconf CLK_TCK", intern=TRUE)),
    error = function(e) return(250)
  )
  stat <- readLines("/proc/stat", n = 1)
  tms <- as.double(strsplit(stat, "\\s+")[[1]][-1]) / clock_ticks
  nms <- c(
    "user", "nice", "system", "idle", "iowait", "irq", "softirq", "steal",
    "guest", "guest_nice"
  )
  names(tms) <- nms[1:length(tms)]
  tms
}

ps__disk_io_counters_linux <- function() {
  # Internal disk IO counters for Linux, reads lines from diskstats if it exists
  # or /sys/block as a backup
  if (file.exists("/proc/diskstats")) {
    ps__read_procfs()
  } else if (dir.exists("/sys/block")) {
    ps__read_sysfs()
  } else {
    stop("Can't read disk IO, neither /proc/diskstats or /sys/block on this system")
  }
}

ps__is_storage_device <- function(name, including_virtual = TRUE) {
  # Whether a named drive (e.g. 'sda') is a real storage device, or a virtual one
  name <- gsub("/", "!", name, fixed=TRUE)
  if (including_virtual) {
    path <- file.path("/sys/block", name)
  } else {
    path <- file.path("/sys/block", name, "device")
  }

  return(dir.exists(path))
}

ps__read_procfs <- function() {
  # Read total disk IO stats from /proc/diskstats
  file <- readLines("/proc/diskstats")
  # Get info as list of vectors (could be different lengths)
  lines <- strsplit(trimws(file), "\\s+")
  # Pre-allocate matrix of info as NAs
  mat <- matrix(data = NA_character_, nrow = length(lines), ncol = 10)
  for (i in seq_along(lines)) {
    # For each line, check length and insert into matrix as appropriate
    line <- lines[[i]]
    flen <- length(line)
    if (flen == 15) {
      # Linux 2.4
      mat[i, 1] <- line[[4]] # name
      mat[i, 2] <- line[[3]] # reads
      mat[i, 3:9] <- line[5:11] # reads_merged, rbytes, rtime, writes, writes_merged, wbytes, wtime
      mat[i, 10] <- line[[14]] # busy_time
    } else if (flen == 14 || flen >= 18) {
      # Linux 2.6+, line referring to a disk
      mat[i, 1] <- line[[3]] # name
      mat[i, 2:9] <- line[5:12] # reads, reads_merged, rbytes, rtime, writes, writes_merged, wbytes, wtime
      mat[i, 10] <- line[[14]] # busy_time
    } else if (flen == 7) {
      # Linux 2.6+, line referring to a partition
      mat[i, 1] <- line[[2]] # name
      mat[i, 2] <- line[[4]] # reads
      mat[i, 4:5] <- line[5:6] # rbytes, writes
      mat[i, 8] <- line[7] # wbytes
    } else {
      stop("Cannot read diskstats file")
    }
  }

  # Add names and convert types as appropriate
  return(data.frame(
    name = mat[, 1],
    read_count = as.numeric(mat[, 2]),
    read_merged_count = as.numeric(mat[, 3]),
    read_bytes = as.numeric(mat[, 4]) * 512, # Multiple by disk sector size to get bytes
    read_time = as.numeric(mat[, 5]),
    write_count = as.numeric(mat[, 6]),
    write_merged_count = as.numeric(mat[, 7]),
    write_bytes = as.numeric(mat[, 8]) * 512, # Multiple by disk sector size to get bytes
    write_time = as.numeric(mat[, 9]),
    busy_time = as.numeric(mat[, 10])
  ))
}

ps__read_sysfs <- function() {
  # Read disk IO from each device folder in /sys/block

  # Get stat files for each block
  blocks <- list.dirs("/sys/block/", recursive = FALSE)
  all_files <- list.files(blocks, full.names = TRUE)
  stats <- all_files[grepl("(stat)$", all_files)]

  # Pre-allocate list for dfs
  disk_info <- vector(mode = "list", length = length(stats))
  for (i in seq_along(stats)) {
    # Read in each file, get the field
    stat <- stats[[i]]
    fields <- readLines(stat)
    fields <- unlist(strsplit(trimws(fields), "\\s+"))

    # Save all info as a dataframe
    block_info <- data.frame(
      name = strsplit(stat, "/", fixed=TRUE)[[1]][[5]], # Extract name from stat filepath
      read_count = as.numeric(fields[[1]]),
      read_merged_count = as.numeric(fields[[2]]),
      read_bytes = as.numeric(fields[[3]]) * 512, # Multiple by disk sector size to get bytes
      read_time = as.numeric(fields[[4]]),
      write_count = as.numeric(fields[[5]]),
      write_merged_count = as.numeric(fields[[6]]),
      write_bytes = as.numeric(fields[[7]]) * 512, # Multiple by disk sector size to get bytes
      write_time = as.numeric(fields[[8]]),
      busy_time = as.numeric(fields[[10]])
    )

    disk_info[[i]] <- block_info
  }
  return(do.call(rbind, disk_info))
}

#' @useDynLib ps, .registration = TRUE
NULL

#' Process table
#'
#' Data frame with the currently running processes.
#'
#' Columns shown by default, if `columns` is not given or `NULL`:
#' * `pid`: Process ID.
#' * `ppid`: Process ID of parent process.
#' * `name`: Process name.
#' * `username`: Name of the user (real uid on POSIX).
#' * `status`: I.e. *running*, *sleeping*, etc.
#' * `user`: User CPU time.
#' * `system`: System CPU time.
#' * `rss`: Resident set size, the amount of memory the process currently
#'    uses. Does not include memory that is swapped out. It does include
#'    shared libraries.
#' * `vms`: Virtual memory size. All memory the process has access to.
#' * `created`: Time stamp when the process was created.
#' * `ps_handle`: `ps_handle` objects, in a list column.
#'
#' Additional columns that can be requested via `columns`:
#' * `cmdline`: Command line, in a single string, from [ps_cmdline()].
#' * `vcmdline`: Like `cmdline`, but each command line argument in a
#'   separate string.
#' * `cwd`: Current working directory, from [ps_cwd()].
#' * `exe`: Path of the executable of the process, from [ps_exe()].
#' * `num_fds`: Number of open file descriptors, from [ps_num_fds()].
#' * `num_threads`: Number of threads, from [ps_num_threads()].
#' * `cpu_children_user`: See [ps_cpu_times()].
#' * `cpu_children_system`: See [ps_cpu_times()].
#' * `terminal`: Terminal device, from [ps_terminal()].
#' * `uid_real`: Real user id, from [ps_uids()].
#' * `uid_effective`: Effective user id, from [ps_uids()].
#' * `uid_saved`: Saved user id, from [ps_uids()].
#' * `gid_real`: Real group id, from [ps_gids()].
#' * `gid_effective`: Effective group id, from [ps_gids()].
#' * `gid_saved`: Saved group id, from [ps_gids()].
#' * `mem_shared`: See [ps_memory_info()].
#' * `mem_text`: See [ps_memory_info()].
#' * `mem_data`: See [ps_memory_info()].
#' * `mem_lib`: See [ps_memory_info()].
#' * `mem_dirty`: See [ps_memory_info()].
#' * `mem_pfaults`: See [ps_memory_info()].
#' * `mem_pageins`: See [ps_memory_info()].
#' * `mem_maxrss`: See [ps_memory_full_info()].
#' * `mem_uss`: See [ps_memory_full_info()].
#' * `mem_pss`: See [ps_memory_full_info()].
#' * `mem_swap`: See [ps_memory_full_info()].
#'
#' Use `"*"` in `columns` to include all columns.
#'
#' @param user Username, to filter the results to matching processes.
#' @param after Start time (`POSIXt`), to filter the results to processes
#'   that started after this.
#' @param columns Columns to include in the result. If `NULL` (the default),
#'   then a default set of columns are returned, see below. The columns are
#'   shown in the same order they are specified in `columns`, but each
#'   column is included at most once. Use `"*"` to include all possible
#'   columns, and prefix a column name with `-` to remove it.
#' @return Data frame, see columns below.
#'
#' @export

ps <- function(user = NULL, after = NULL, columns = NULL) {
  if (!is.null(user)) assert_string(user)
  if (!is.null(after)) assert_time(after)

  columns <- unique(columns %||% ps_default_columns)
  if ("*" %in% columns) {
    columns <- append(columns, ps_all_columns, which(columns == "*"))
    columns <- columns[columns != "*"]
  }
  if (any(bad <- !columns %in% ps_all_columns)) {
    stop(
      "Unknown column", if (sum(bad) > 1) "s", " requested: ",
      paste(columns[bad], collapse = ", ")
    )
  }

  pids <- ps_pids()
  processes <- not_null(lapply(pids, function(p) {
    tryCatch(ps_handle(p), error = function(e) NULL) }))

  ct <- NULL
  if (!is.null(after)) {
    ct <- lapply(processes, ps_create_time)
    selected <- ct >= after
    processes <- processes[selected]
    ct <- ct[selected]
  }

  us <- NULL
  if (!is.null(user)) {
    us <- map_chr(processes, function(p)
      fallback(ps_username(p), NA_character_))
    selected <- !is.na(us) & us == user
    processes <- processes[selected]
    us <- us[selected]
    ct <- ct[selected]
  }

  ct <- ct %||% lapply(processes, function(p)
    fallback(ps_create_time(p), NA_time()))

  if (c_username <- "username" %in% columns) {
    us <- us %||% map_chr(processes, function(p)
      fallback(ps_username(p), NA_character_))
  }
  if (c_pid <- "pid" %in% columns) {
    pd <- map_int(processes, function(p) fallback(ps_pid(p), NA_integer_))
  }
  if (c_ppid <- "ppid" %in% columns) {
    pp <- map_int(processes, function(p) fallback(ps_ppid(p), NA_integer_))
  }
  if (c_name <- "name" %in% columns) {
    nm <- map_chr(processes, function(p) fallback(ps_name(p), NA_character_))
  }

  if (c_status <- "status" %in% columns) {
    opt <- options(ps.no_external_ps = TRUE)
    on.exit(options(opt), add = TRUE)
    st <- map_chr(processes, function(p) fallback(ps_status(p), NA_character_))
    options(opt)
    if (ps_os_type()[["MACOS"]] && !isTRUE(getOption("ps.no_external_ps")) &&
        anyNA(st[pids != 0])) {
      misspids <- map_int(processes[is.na(st)], ps_pid)
      st[is.na(st)] <- ps_status_macos_ps(misspids)
    }
  }

  c_user <- "user" %in% columns
  c_system <- "system" %in% columns
  c_cpu_children_user <- "cpu_children_user" %in% columns
  c_cpu_children_system <- "cpu_children_system" %in% columns
  if (c_user || c_system || c_cpu_children_user || c_cpu_children_system) {
    time <- lapply(processes, function(p) fallback(ps_cpu_times(p), NULL))
    cpt <- map_dbl(time, function(x) x[["user"]] %||% NA_real_)
    cps <- map_dbl(time, function(x) x[["system"]] %||% NA_real_)
    cpct <- map_dbl(time, function(x) x[["children_user"]] %||% NA_real_)
    cpcs <- map_dbl(time, function(x) x[["children_system"]] %||% NA_real_)
  }
  c_rss <- "rss" %in% columns
  c_vms <- "vms" %in% columns
  c_mem_shared <- "mem_shared" %in% columns
  c_mem_text <- "mem_text" %in% columns
  c_mem_data <- "mem_data" %in% columns
  c_mem_lib <- "mem_lib" %in% columns
  c_mem_dirty <- "mem_dirty" %in% columns
  c_mem_pfaults <- "mem_pfaults" %in% columns
  c_mem_pageins <- "mem_pageins" %in% columns
  c_mem_maxrss <- "mem_maxrss" %in% columns
  c_mem_uss <- "mem_uss" %in% columns
  c_mem_pss <- "mem_pss" %in% columns
  c_mem_swap <- "mem_swap" %in% columns
  if (c_rss || c_vms || c_mem_shared || c_mem_text || c_mem_data ||
      c_mem_lib || c_mem_dirty || c_mem_pfaults || c_mem_pageins ||
      c_mem_maxrss || c_mem_uss || c_mem_pss || c_mem_swap) {
    if (c_mem_maxrss || c_mem_uss || c_mem_pss || c_mem_swap) {
      mem <- lapply(processes, function(p) fallback(ps_memory_full_info(p), NULL))
    } else {
      mem <- lapply(processes, function(p) fallback(ps_memory_info(p), NULL))
    }
    rss <- map_dbl(mem, function(x) x[["rss"]] %||% NA_real_)
    vms <- map_dbl(mem, function(x) x[["vms"]] %||% NA_real_)
    mem_shared <- map_dbl(mem, function(x) x["shared"] %||% NA_real_)
    mem_text <- map_dbl(mem, function(x) x["text"] %||% NA_real_)
    mem_data <- map_dbl(mem, function(x) x["data"] %||% NA_real_)
    mem_lib <- map_dbl(mem, function(x) x["lib"] %||% NA_real_)
    mem_dirty <- map_dbl(mem, function(x) x["dirty"] %||% NA_real_)
    mem_pfaults <- map_dbl(mem, function(x) x["pfaults"] %||% NA_real_)
    mem_pageins <- map_dbl(mem, function(x) x["pageins"] %||% NA_real_)
    mem_maxrss<- map_dbl(mem, function(x) x["maxrss"] %||% NA_real_)
    mem_uss <- map_dbl(mem, function(x) x["uss"] %||% NA_real_)
    mem_pss <- map_dbl(mem, function(x) x["pss"] %||% NA_real_)
    mem_swap <- map_dbl(mem, function(x) x["swap"] %||% NA_real_)
  }
  if (c_ps_handle <- "ps_handle" %in% columns) {
    ps_handle <- I(processes)
  }
  c_cmdline <- "cmdline" %in% columns
  c_vcmdline <- "vcmdline" %in% columns
  if (c_cmdline || c_vcmdline) {
    vcmdline <- I(lapply(
      processes,
      function(p) fallback(ps_cmdline(p), NULL)
    ))
    cmdline <- map_chr(
      vcmdline,
      function(x) {
        if (is.null(x)) NA_character_ else paste(x, collapse = " ")
      }
    )
  }

  if (c_cwd <- "cwd" %in% columns) {
    cwd <- map_chr(processes, function(x) fallback(ps_cwd(x), NA_character_))
  }
  if (c_exe <- "exe" %in% columns) {
    exe <- map_chr(processes, function(x) fallback(ps_exe(x), NA_character_))
  }
  if (c_num_fds <- "num_fds" %in% columns) {
    num_fds <- map_int(processes, function(x) fallback(ps_num_fds(x), NA_integer_))
  }
  if (c_num_threads <- "num_threads" %in% columns) {
    num_threads <- map_int(processes, function(x) fallback(ps_num_threads(x), NA_integer_))
  }
  if (c_terminal <- "terminal" %in% columns) {
    terminal <- map_chr(processes, function(x) fallback(ps_terminal(x), NA_character_))
  }
  c_uid_real <- "uid_real" %in% columns
  c_uid_effective <- "uid_effective" %in% columns
  c_uid_saved <- "uid_saved" %in% columns
  if (c_uid_real || c_uid_effective || c_uid_saved) {
    uids <- lapply(processes, function(x) fallback(ps_uids(x), NULL))
    uid_real <- map_int(uids, function(x) if (is.null(x)) NA_integer_ else x[["real"]])
    uid_effective <- map_int(uids, function(x) if (is.null(x)) NA_integer_ else x[["effective"]])
    uid_saved <- map_int(uids, function(x) if (is.null(x)) NA_integer_ else x[["saved"]])
  }
  c_gid_real <- "gid_real" %in% columns
  c_gid_effective <- "gid_effective" %in% columns
  c_gid_saved <- "gid_saved" %in% columns
  if (c_gid_real || c_gid_effective || c_gid_saved) {
    gids <- lapply(processes, function(x) fallback(ps_gids(x), NULL))
    gid_real <- map_int(gids, function(x) if (is.null(x)) NA_integer_ else x[["real"]])
    gid_effective <- map_int(gids, function(x) if (is.null(x)) NA_integer_ else x[["effective"]])
    gid_saved <- map_int(gids, function(x) if (is.null(x)) NA_integer_ else x[["saved"]])
  }

  c_created <- "created" %in% columns
  created <- format_unix_time(unlist(ct))

  pss <- as_data_frame(not_null(list(
    # default
    pid       = if (c_pid)       pd,
    ppid      = if (c_ppid)      pp,
    name      = if (c_name)      nm,
    username  = if (c_username)  us,
    status    = if (c_status)    st,
    user      = if (c_user)      cpt,
    system    = if (c_system)    cps,
    rss       = if (c_rss)       rss,
    vms       = if (c_vms)       vms,
    created   = if (c_created)   created,
    ps_handle = if (c_ps_handle) ps_handle,

    # optional
    cmdline   = if (c_vcmdline)  cmdline,
    vcmdline  = if (c_vcmdline)  vcmdline,
    cwd       = if (c_cwd)       cwd,
    exe       = if (c_exe)       exe,
    num_fds   = if (c_num_fds)   num_fds,
    num_threads = if (c_num_threads) num_threads,
    cpu_children_user = if (c_cpu_children_user) cpct,
    cpu_children_system = if (c_cpu_children_system) cpcs,
    terminal            = if (c_terminal)            terminal,
    uid_real            = if (c_uid_real)            uid_real,
    uid_effective       = if (c_uid_effective)       uid_effective,
    uid_saved           = if (c_uid_saved)           uid_saved,
    gid_real            = if (c_gid_real)            gid_real,
    gid_effective       = if (c_gid_effective)       gid_effective,
    gid_saved           = if (c_gid_saved)           gid_saved,
    mem_shared          = if (c_mem_shared)          mem_shared,
    mem_text            = if (c_mem_text)            mem_text,
    mem_data            = if (c_mem_data)            mem_data,
    mem_lib             = if (c_mem_lib)             mem_lib,
    mem_dirty           = if (c_mem_dirty)           mem_dirty,
    mem_pfaults         = if (c_mem_pfaults)         mem_pfaults,
    mem_pageins         = if (c_mem_pageins)         mem_pageins,
    mem_maxrss          = if (c_mem_maxrss)          mem_maxrss,
    mem_uss             = if (c_mem_uss)             mem_uss,
    mem_pss             = if (c_mem_pss)             mem_pss,
    mem_swap            = if (c_mem_swap)            mem_swap,

    NULL
  )))

  pss <- pss[order(-as.numeric(created)), ]
  pss <- pss[, columns]

  pss
}

ps_default_columns <- c(
  "pid",
  "ppid",
  "name",
  "username",
  "status",
  "user",
  "system",
  "rss",
  "vms",
  "created",
  "ps_handle",
  NULL
)

ps_all_columns <- c(
  ps_default_columns,
  "cmdline",                    # ps_cmdline
  "vcmdline",
  "cwd",                        # ps_cwd
  "exe",
  "num_fds",                    # ps_num_fds
  "num_threads",                # ps_num_threads
  "cpu_children_user",          # ps_cpu_times
  "cpu_children_system",
  "terminal",                   # ps_terminal
  "uid_real",                   # ps_uids
  "uid_effective",
  "uid_saved",
  "gid_real",                   # ps_gids
  "gid_effective",
  "gid_saved",
  "mem_shared",                 # ps_memory_info
  "mem_text",
  "mem_data",
  "mem_lib",
  "mem_dirty",
  "mem_pfaults",
  "mem_pageins",
  "mem_maxrss",                 # ps_full_memory_info
  "mem_uss",
  "mem_pss",
  "mem_swap",
  NULL
)

#' Encode a `ps_handle` as a short string
#'
#' A convenient format for passing between processes, naming semaphores, or
#' using as a directory/file name. Will always be 12 alphanumeric characters,
#' with the first character guarantied to be a letter. Encodes the pid and
#' creation time for a process.
#'
#' @param p Process handle.
#'
#' @return A process string (scalar character), that can be passed to
#' `ps_handle()` in place of a pid.
#'
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' (p <- ps_handle())
#' (str <- ps_string(p))
#' ps_handle(pid = str)

ps_string <- function(p = ps_handle()) {
  assert_ps_handle(p)
  ps__str_encode(p)
}


ps__str_encode <- function(p) {

  # Assumptions:
  #   - Date < 8888-12-02 (Windows only).
  #   - System uptime < 6918 years (Unix only).
  #   - PID <= 768,369,472 (current std max = 4,194,304).
  #   - PIDs are not reused within the same millisecond.

  # Surprisingly, `ps_boot_time()` is not constant from process to process, and
  # `ps_create_time()` is derived from `ps_boot_time()` on Unix. Therefore:
  #   - On Windows, encode `ps_create_time()`
  #   - On Unix, encode `ps_create_time() - `ps_boot_time()`

  pid  <- ps_pid(p)
  time <- as.numeric(ps_create_time(p))

  if (.Platform$OS.type == "unix")
    time <- time - as.numeric(ps_boot_time())

  time <- round(time, 3) * 1000 # millisecond resolution

  map <- c(letters, LETTERS, 0:9)
  paste(
    collapse = '',
    map[
      1 +
        c(
          floor(pid  / 62^(3:0)) %% 62,
          floor(time / 62^(7:0)) %% 62
        )
    ]
  )
}


ps__str_decode <- function(str) {

  map <- structure(0:61, names = c(letters, LETTERS, 0:9))
  val <- map[strsplit(str, '', fixed = TRUE)[[1]]]
  pid <- sum(val[01:04] * 62^(3:0))

  tryCatch(
    expr = {
      p <- ps_handle(pid = pid)
      stopifnot(str == ps__str_encode(p))
      p
    },
    error = function(e) {

      time <- sum(val[05:12] * 62^(7:0)) / 1000

      if (.Platform$OS.type == "unix")
        time <- time + as.numeric(ps_boot_time())

      ps_handle(pid = pid, time = format_unix_time(time))
    }
  )
}

#' Mark a process and its (future) child tree
#'
#' `ps_mark_tree()` generates a random environment variable name and sets
#' it in the  current R process. This environment variable will be (by
#' default) inherited by all child (and grandchild, etc.) processes, and
#' will help finding these processes, even if and when they are (no longer)
#' related to the current R process. (I.e. they are not connected in the
#' process tree.)
#'
#' `ps_find_tree()` finds the processes that set the supplied environment
#' variable and returns them in a list.
#'
#' `ps_kill_tree()` finds the processes that set the supplied environment
#' variable, and kills them (or sends them the specified signal on Unix).
#'
#' `with_process_cleanup()` evaluates an R expression, and cleans up all
#' external processes that were started by the R process while evaluating
#' the expression. This includes child processes of child processes, etc.,
#' recursively. It returns a list with entries: `result` is the result of
#' the expression, `visible` is TRUE if the expression should be printed
#' to the screen, and `process_cleanup` is a named integer vector of the
#' cleaned pids, names are the process names.
#'
#' If `expr` throws an error, then so does `with_process_cleanup()`, the
#' same error. Nevertheless processes are still cleaned up.
#'
#' @section Note:
#' Note that `with_process_cleanup()` is problematic if the R process is
#' multi-threaded and the other threads start subprocesses.
#' `with_process_cleanup()` cleans up those processes as well, which is
#' probably not what you want. This is an issue for example in RStudio.
#' Do not use `with_process_cleanup()`, unless you are sure that the
#' R process is single-threaded, or the other threads do not start
#' subprocesses. E.g. using it in package test cases is usually fine,
#' because RStudio runs these in a separate single-threaded process.
#'
#' The same holds for manually running `ps_mark_tree()` and then
#' `ps_find_tree()` or `ps_kill_tree()`.
#'
#' A safe way to use process cleanup is to use the processx package to
#' start subprocesses, and set the `cleanup_tree = TRUE` in
#' [processx::run()] or the [processx::process] constructor.
#'
#' @return `ps_mark_tree()` returns the name of the environment variable,
#' which can be used as the `marker` in `ps_kill_tree()`.
#'
#' `ps_find_tree()` returns a list of `ps_handle` objects.
#'
#' `ps_kill_tree()` returns the pids of the killed processes, in a named
#' integer vector. The names are the file names of the executables, when
#' available.
#'
#' `with_process_cleanup()` returns the value of the evaluated expression.
#'
#' @rdname ps_kill_tree
#' @export

ps_mark_tree <- function() {
  id <- get_id()
  do.call(Sys.setenv, structure(list("YES"), names = id))
  id
}

get_id <- function() {
  paste0(
    "PS",
    paste(
      sample(c(LETTERS, 0:9), 10, replace = TRUE),
      collapse = ""
    ),
    "_",
    as.integer(Internal(Sys.time()))
  )
}

#' @param expr R expression to evaluate in the new context.
#'
#' @rdname ps_kill_tree
#' @export

with_process_cleanup <- function(expr) {
  id <- ps_mark_tree()
  stat <- NULL
  do <- function() {
    on.exit(stat <<- ps_kill_tree(id), add = TRUE)
    withVisible(expr)
  }

  res <- do()

  ret <- list(
    result = res$value,
    visible = res$visible,
    process_cleanup = stat)
  class(ret) <- "with_process_cleanup"
  ret
}

#' @export

print.with_process_cleanup <- function(x, ...) {
  if (x$visible) print(x$result)
  if (length(x$process_cleanup)) {
    cat("!! Cleaned up the following processes:\n")
    print(x$process_cleanup)
  } else {
    cat("-- No leftover processes to clean up.\n")
  }
  invisible(x)
}


#' @rdname ps_kill_tree
#' @export

ps_find_tree <- function(marker) {
  assert_string(marker)
  after <- as.numeric(strsplit(marker, "_", fixed = TRUE)[[1]][2])

  pids <- setdiff(ps_pids(), Sys.getpid())

  not_null(lapply(pids, function(p) {
    tryCatch(
      .Call(ps__find_if_env, marker, after, p),
      error = function(e) NULL
    )
  }))
}

#' @param marker String scalar, the name of the environment variable to
#' use to find the marked processes.
#' @param sig The signal to send to the marked processes on Unix. On
#' Windows this argument is ignored currently.
#' @param grace Grace period, in milliseconds, used on Unix, if `sig` is
#'   `SIGKILL`.  If it is not zero, then `ps_kill_tree()` first sends a
#'   `SIGTERM` signal to all processes. If some proccesses do not
#'   terminate within `grace` milliseconds after the `SIGTERM` signal,
#'   `ps_kill_tree()` kills them by sending `SIGKILL` signals.
#'
#' @rdname ps_kill_tree
#' @export

ps_kill_tree <- function(marker, sig = signals()$SIGKILL, grace = 200) {

  # NULL on Windows
  if (!ps_os_type()[["WINDOWS"]]) {
    sig <- assert_integer(sig)
  }

  procs <- ps_find_tree(marker)
  pids <- map_int(procs, ps_pid)
  nms <- map_chr(
    procs,
    function(p) tryCatch(ps_name(p), error = function(e) "???")
  )

  if (!ps_os_type()[["WINDOWS"]] && sig == signals()$SIGKILL) {
    ps_send_signal(procs, sig)
  } else {
    ps_kill(procs, grace = grace)
  }

  structure(pids, names = nms)
}

test_that("ps_mark_tree", {
  id <- ps_mark_tree()
  on.exit(Sys.unsetenv(id), add = TRUE)
  expect_true(is.character(id))
  expect_true(length(id) == 1)
  expect_false(is.na(id))
  expect_false(Sys.getenv(id) == "")
})

test_that("kill_tree",  {
  skip_on_cran()
  skip_in_rstudio()

  res <- ps_kill_tree(get_id())
  expect_equal(length(res), 0)
  expect_true(is.integer(res))

  ## Child processes
  id <- ps_mark_tree()
  on.exit(Sys.unsetenv(id), add = TRUE)
  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  p <- lapply(1:5, function(x) {
    out <- file.path(tmp, basename(tempfile()))
    processx::process$new(
      px(),
      c("outln", "ready", "sleep", "10"),
      stdout = out
    )
  })
  on.exit(lapply(p, function(x) x$kill()), add = TRUE)

  timeout <- Sys.time() + 5
  while (sum(file_size(dir(tmp, full.names = TRUE)) > 0) < 5 &&
         Sys.time() < timeout) Sys.sleep(0.1)

  expect_true(Sys.time() < timeout)

  res <- ps_kill_tree(id)
  res <- res[names(res) %in% c("px", "px.exe")]
  expect_equal(length(res), 5)
  expect_equal(
    sort(as.integer(res)),
    sort(map_int(p, function(x) x$get_pid())))

  ## We need to wait a bit here, potentially, because the process
  ## might be a zombie, which is technically alive.
  now <- Sys.time()
  timeout <- now + 5
  while (any(map_lgl(p, function(pp) pp$is_alive())) &&
         Sys.time() < timeout) Sys.sleep(0.1)

  expect_true(Sys.time() < timeout)
  lapply(p, function(pp) expect_false(pp$is_alive()))
})

test_that("kill_tree, grandchild", {
  skip_on_cran()
  skip_in_rstudio()

  id <- ps_mark_tree()
  on.exit(Sys.unsetenv(id), add = TRUE)

  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  N <- 3
  p <- lapply(1:N, function(x) {
    callr::r_bg(
      function(d) {
        cat("OK\n", file = file.path(d, Sys.getpid()))
        # We ignore error from the grandchild, in case it gets
        # killed first. The child still runs on, because of the sleep.
        try(callr::r(
          function(d) {
            cat("OK\n", file = file.path(d, Sys.getpid()))
            Sys.sleep(5)
          },
          args = list(d = d)))
        Sys.sleep(5)
      },
      args = list(d = tmp),
      cleanup = FALSE
    )
  })
  on.exit(lapply(p, function(x) x$kill()), add = TRUE)

  timeout <- Sys.time() + 10
  while (length(dir(tmp)) < 2*N && Sys.time() < timeout) Sys.sleep(0.1)

  expect_true(Sys.time() < timeout)

  res <- ps_kill_tree(id)

  ## Older processx versions do not close the connections on kill,
  ## so the cleanup reporter picks them up
  lapply(p, function(pp) {
    close(pp$get_output_connection())
    close(pp$get_error_connection())
  })

  res <- res[names(res) %in% c("R", "Rterm.exe")]

  ## We might miss some processes, because grandchildren can be
  ## are in the same job object and they are cleaned up automatically.
  ## To fix the, processx would need an option _not_ to create a job
  ## object.
  expect_true(length(res) <= N * 2)
  expect_true(all(names(res) %in% c("R", "Rterm.exe")))
  cpids <- map_int(p, function(x) x$get_pid())
  expect_true(all(cpids %in% res))
  ccpids <- as.integer(dir(tmp))

  ## Again, the opposite might not be true, because we might miss some
  ## grandchildren.
  expect_true(all(res %in% ccpids))

  ## Nevertheless none of them should be alive.
  ## (Taking the risk of pid reuse here...)
  timeout <- Sys.time() + 5
  while (any(ccpids %in% ps_pids()) && Sys.time() < timeout) Sys.sleep(0.1)
  expect_true(Sys.time() < timeout)
})

test_that("kill_tree, orphaned grandchild", {
  skip_on_cran()
  skip_in_rstudio()

  id <- ps_mark_tree()
  on.exit(Sys.unsetenv(id), add = TRUE)

  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  cmdline <- paste(px(), "sleep 5")

  N <- 3
  lapply(1:N, function(x) {
    system2(px(), c("outln", "ok","sleep", "5"),
            stdout = file.path(tmp, x), wait = FALSE)
  })

  timeout <- Sys.time() + 10
  while (sum(file_size(dir(tmp, full.names = TRUE)) > 0) < N &&
         Sys.time() < timeout) Sys.sleep(0.1)

  res <- ps_kill_tree(id)
  res <- res[names(res) %in% c("px", "px.exe")]
  expect_equal(length(res), N)
  expect_true(all(names(res) %in% c("px", "px.exe")))
})

test_that("with_process_cleanup", {
  skip_on_cran()
  skip_in_rstudio()

  p <- NULL
  with_process_cleanup({
    p <- lapply(1:3, function(x) {
      processx::process$new(px(), c("sleep", "10"))
    })
    expect_equal(length(p), 3)
    lapply(p, function(pp) expect_true(pp$is_alive()))
  })

  expect_equal(length(p), 3)

  ## We need to wait a bit here, potentially, because the process
  ## might be a zombie, which is technically alive.
  now <- Sys.time()
  timeout <- now + 5
  while (any(map_lgl(p, function(pp) pp$is_alive())) &&
         Sys.time() < timeout) Sys.sleep(0.05)

  lapply(p, function(pp) expect_false(pp$is_alive()))
  rm(p)
})

test_that("find_tree",  {
  skip_on_cran()
  skip_in_rstudio()
  skip_if_no_processx()

  res <- ps_find_tree(get_id())
  expect_equal(length(res), 0)
  expect_true(is.list(res))

  ## Child processes
  id <- ps_mark_tree()
  on.exit(Sys.unsetenv(id), add = TRUE)
  p <- lapply(1:5, function(x) processx::process$new(px(), c("sleep", "10")))
  on.exit(lapply(p, function(x) x$kill()), add = TRUE)
  res <- ps_find_tree(id)
  names <- not_null(lapply(res, function(p) fallback(ps_name(p), NULL)))
  res <- res[names %in% c("px", "px.exe")]
  expect_equal(length(res), 5)
  expect_equal(
    sort(map_int(res, ps_pid)),
    sort(map_int(p, function(x) x$get_pid())))

  lapply(p, function(x) x$kill())
})

test_that("find_tree, grandchild", {
  skip_on_cran()
  skip_in_rstudio()

  id <- ps_mark_tree()
  on.exit(Sys.unsetenv(id), add = TRUE)

  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  N <- 3
  p <- lapply(1:N, function(x) {
    callr::r_bg(
      function(d) {
        callr::r(
          function(d) {
            cat("OK\n", file = file.path(d, Sys.getpid()))
            Sys.sleep(5)
          },
          args = list(d = d))
      },
      args = list(d = tmp))
  })
  on.exit(lapply(p, function(x) x$kill()), add = TRUE)
  on.exit(ps_kill_tree(id), add = TRUE)

  timeout <- Sys.time() + 10
  while (length(dir(tmp)) < N && Sys.time() < timeout) Sys.sleep(0.1)

  res <- ps_find_tree(id)
  names <- not_null(lapply(res, function(p) fallback(ps_name(p), NULL)))
  res <- res[names %in% c("R", "Rterm.exe")]
  expect_equal(length(res), N * 2)
  cpids <- map_int(p, function(x) x$get_pid())
  res_pids <- map_int(res, ps_pid)
  expect_true(all(cpids %in% res_pids))
  ccpids <- as.integer(dir(tmp))
  expect_true(all(ccpids %in% res_pids))

  ## Older processx versions do not close the connections on kill,
  ## so the cleanup reporter picks them up
  lapply(p, function(pp) {
    pp$kill()
    close(pp$get_output_connection())
    close(pp$get_error_connection())
  })
})

test_that("find_tree, orphaned grandchild", {
  skip_on_cran()
  skip_in_rstudio()

  id <- ps_mark_tree()
  on.exit(Sys.unsetenv(id), add = TRUE)

  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  cmdline <- paste(px(), "sleep 5")

  N <- 3
  lapply(1:N, function(x) {
    system2(px(), c("outln", "ok","sleep", "5"),
            stdout = file.path(tmp, x), wait = FALSE)
  })
  on.exit(ps_kill_tree(id), add = TRUE)

  timeout <- Sys.time() + 10
  while (sum(file_size(dir(tmp, full.names = TRUE)) > 0) < N &&
         Sys.time() < timeout) Sys.sleep(0.1)

  res <- ps_find_tree(id)
  names <- not_null(lapply(res, function(p) fallback(ps_name(p), NULL)))
  res <- res[names %in% c("px", "px.exe")]
  expect_equal(length(res), N)
})

globalVariables("private")

#' testthat reporter that checks if child processes are cleaned up in tests
#'
#' `CleanupReporter` takes an existing testthat `Reporter` object, and
#' wraps it, so it checks for leftover child processes, at the specified
#' place, see the `proc_unit` argument below.
#'
#' Child processes can be reported via a failed expectation, cleaned up
#' silently, or cleaned up and reported (the default).
#'
#' If a `test_that()` block has an error, `CLeanupReporter` does not
#' emit any expectations at the end of that block. The error will lead to a
#' test failure anyway. It will still perform the cleanup, if requested,
#' however.
#'
#' The constructor of the `CleanupReporter` class has options:
#' * `file`: the output file, if any, this is passed to `reporter`.
#' * `proc_unit`: when to perform the child process check and cleanup.
#'   Possible values:
#'     * `"test"`: at the end of each [testthat::test_that()] block
#'       (the default),
#'     * `"testsuite"`: at the end of the test suite.
#' * `proc_cleanup`: Logical scalar, whether to kill the leftover
#'   processes, `TRUE` by default.
#' * `proc_fail`: Whether to create an expectation, that fails if there
#'   are any processes alive, `TRUE` by default.
#' * `proc_timeout`: How long to wait for the processes to quit. This is
#'   sometimes needed, because even if some kill signals were sent to
#'   child processes, it might take a short time for these to take effect.
#'   It defaults to one second.
#' * `rconn_unit`: When to perform the R connection cleanup. Possible values
#'   are `"test"` and `"testsuite"`, like for `proc_unit`.
#' * `rconn_cleanup`: Logical scalar, whether to clean up leftover R
#'   connections. `TRUE` by default.
#' * `rconn_fail`: Whether to fail for leftover R connections. `TRUE` by
#'   default.
#' * `file_unit`: When to check for open files. Possible values are
#'    `"test"` and `"testsuite"`, like for `proc_unit`.
#' * `file_fail`: Whether to fail for leftover open files. `TRUE` by
#'   default.
#' * `conn_unit`: When to check for open network connections.
#'   Possible values are `"test"` and `"testsuite"`, like for `proc_unit`.
#' * `conn_fail`: Whether to fail for leftover network connections.
#'   `TRUE` by default.
#'
#' @note Some IDEs, like RStudio, start child processes frequently, and
#' sometimes crash when these are killed, only use this reporter in a
#' terminal session. In particular, you can always use it in the
#' idiomatic `testthat.R` file, that calls `test_check()` during
#' `R CMD check`.
#'
#' @param reporter A testthat reporter to wrap into a new `CleanupReporter`
#'   class.
#' @return New reporter class that behaves exactly like `reporter`,
#'   but it checks for, and optionally cleans up child processes, at the
#'   specified granularity.
#'
#' @section Examples:
#' This is how to use this reporter in `testthat.R`:
#' ```
#' library(testthat)
#' library(mypackage)
#'
#' if  (ps::ps_is_supported()) {
#'   reporter <- ps::CleanupReporter(testthat::ProgressReporter)$new(
#'     proc_unit = "test", proc_cleanup = TRUE)
#' } else {
#'   ## ps does not support this platform
#'   reporter <- "progress"
#' }
#'
#' test_check("mypackage", reporter = reporter)
#' ```
#'
#' @export

CleanupReporter <- function(reporter = testthat::ProgressReporter) {

  R6::R6Class("CleanupReporter",
    inherit = reporter,
    public = list(

      initialize = function(
        file = getOption("testthat.output_file", stdout()),
        proc_unit = c("test", "testsuite"),
        proc_cleanup = TRUE, proc_fail = TRUE, proc_timeout = 1000,
        rconn_unit = c("test", "testsuite"),
        rconn_cleanup = TRUE, rconn_fail = TRUE,
        file_unit = c("test", "testsuite"), file_fail = TRUE,
        conn_unit = c("test", "testsuite"), conn_fail = TRUE) {

        if (!ps::ps_is_supported()) {
          stop("CleanupReporter is not supported on this platform")
        }

        super$initialize(file = file)
        private$proc_unit <- match.arg(proc_unit)
        private$proc_cleanup <- proc_cleanup
        private$proc_fail <- proc_fail
        private$proc_timeout <- proc_timeout

        private$rconn_unit <- match.arg(rconn_unit)
        private$rconn_cleanup <- rconn_cleanup
        private$rconn_fail <- rconn_fail

        private$file_unit <- match.arg(file_unit)
        private$file_fail <- file_fail

        private$conn_unit <- match.arg(conn_unit)
        private$conn_fail <- conn_fail

        invisible(self)
      },

      start_test = function(context, test) {
        private$has_error <- FALSE
        super$start_test(context, test)
        if (private$file_unit == "test") private$files <- ps_open_files(ps_handle())
        if (private$rconn_unit == "test") private$rconns <- showConnections()
        if (private$proc_unit == "test") private$tree_id <- ps::ps_mark_tree()
        if (private$conn_unit == "test") private$conns <- ps_connections(ps_handle())
      },

      end_test = function(context, test) {
        if (private$proc_unit == "test") self$do_proc_cleanup(test)
        if (private$rconn_unit == "test") self$do_rconn_cleanup(test)
        if (private$file_unit == "test") self$do_file_cleanup(test)
        if (private$conn_unit == "test") self$do_conn_cleanup(test)
        super$end_test(context, test)
      },

      add_result = function(context, test, result) {
        if (inherits(result, "expectation_error")) {
          private$has_error <- TRUE
        }
        super$add_result(context, test, result)
      },

      start_reporter = function() {
        super$start_reporter()
        if (private$file_unit == "testsuite") private$files <- ps_open_files(ps_handle())
        if (private$rconn_unit == "testsuite") private$rconns <- showConnections()
        if (private$proc_unit == "testsuite") private$tree_id <- ps::ps_mark_tree()
        if (private$conn_unit == "testsuite") private$conns <- ps_connections(ps_handle())
      },

      end_reporter = function() {
        super$end_reporter()
        if (private$proc_unit  == "testsuite") {
          self$do_proc_cleanup("testsuite", quote = "")
        }
        if (private$rconn_unit  == "testsuite") {
          self$do_rconn_cleanup("testsuite", quote = "")
        }
        if (private$file_unit  == "testsuite") {
          self$do_file_cleanup("testsuite", quote = "")
        }
        if (private$conn_unit  == "testsuite") {
          self$do_conn_cleanup("testsuite", quote = "")
        }
      },

      do_proc_cleanup = function(test, quote = "'") {
        Sys.unsetenv(private$tree_id)
        deadline <- Sys.time() + private$proc_timeout / 1000
        if (private$proc_fail) {
          while (length(ret <- ps::ps_find_tree(private$tree_id)) &&
                 Sys.time() < deadline) Sys.sleep(0.05)
          # maybe gc() will clean up something
          if (length(ret) > 0) {
            gc()
            ret <- ps::ps_find_tree(private$tree_id)
          }
        }
        if (private$proc_cleanup) {
          ret <- ps::ps_kill_tree(private$tree_id)
        }
        if (private$proc_fail && !private$has_error)  {
          testthat::with_reporter(self, start_end_reporter = FALSE, {
            self$expect_cleanup(test, ret, quote)
          })
        }
      },

      do_rconn_cleanup = function(test, quote = "'") {
        old <- private$rconns
        new <- showConnections()
        private$rconns <- NULL
        leftover <- ! new[, "description"] %in% old[, "description"]

        # maybe gc() will clean up some
        if (sum(leftover) > 0) {
          gc()
          new <- showConnections()
          leftover <- ! new[, "description"] %in% old[, "description"]
        }

        if (private$rconn_cleanup) {
          for (no in as.integer(rownames(new)[leftover])) {
            tryCatch(close(getConnection(no)), error = function(e) NULL)
          }
        }

        if (private$rconn_fail && !private$has_error) {
          act <- testthat::quasi_label(rlang::enquo(test), test)
          testthat::expect(
            sum(leftover) == 0,
            sprintf(
              "%s did not close R connections: %s",
              encodeString(act$lab, quote = quote),
              paste0(encodeString(new[leftover, "description"], quote = "'"),
                     " (", rownames(new)[leftover], ")", collapse = ",  ")))
        }
      },

      do_file_cleanup = function(test, quote = "'") {
        old <- private$files
        new <- ps_open_files(ps_handle())
        private$files <- NULL
        leftover <- ! new$path %in% old$path

        ## Need to ignore some open files:
        ## * /dev/urandom might be opened internally by curl, openssl, etc.
        leftover <- leftover & new$path != "/dev/urandom"

        # maybe gc() will clean up some
        if (sum(leftover) > 0) {
          gc()
          new <- ps_open_files(ps_handle())
          leftover <- ! new$path %in% old$path
          leftover <- leftover & new$path != "/dev/urandom"
        }

        if (private$file_fail && !private$has_error) {
          act <- testthat::quasi_label(rlang::enquo(test), test)
          testthat::expect(
            sum(leftover) == 0,
            sprintf(
              "%s did not close open files: %s",
              encodeString(act$lab, quote = quote),
              paste0(encodeString(new$path[leftover], quote = "'"),
                     collapse = ",  ")))
        }
      },

      do_conn_cleanup = function(test, quote = "'") {
        old <- private$conns[, 1:6]
        private$conns <- NULL

        ## On windows, sometimes it takes time to remove the connection
        ## from the processes connection tables, so we try waiting a bit.
        ## We haven't seen issues with this on other OSes yet.
        deadline <- Sys.time() + as.difftime(0.5, units = "secs")
        done <- FALSE
        repeat {
          new <- ps_connections(ps_handle())[, 1:6]
          ## This is a connection that is used internally on macOS,
          ## for DNS resolution. We'll just ignore it. Looks like this:
          ## # A data frame: 2 x 6
          ##    fd family  type        laddr lport raddr
          ## <int> <chr>   <chr>       <chr> <int> <chr>
          ##     7 AF_UNIX SOCK_STREAM <NA>     NA /var/run/mDNSResponder
          ##    10 AF_UNIX SOCK_STREAM <NA>     NA /var/run/mDNSResponder
          new <- new[
            new$family != "AF_UNIX" |
            new$type != "SOCK_STREAM" |
            is.na(new$raddr) |
            paste(tolower(basename(new$raddr))) != "mdnsresponder", ]

          leftover <- ! apply(new, 1, paste, collapse = "&") %in%
            apply(old, 1, paste, collapse = "&")

          # is this the final try, or are we all clean?
          if (done || sum(leftover) == 0) break

          # if Unix, then try again after gc()
          # on Windows, gc() after a timeout, then quit
          if (!ps_os_type()[["WINDOWS"]] || Sys.time() >= deadline) {
            gc()
            done <- TRUE
            next
          }

          Sys.sleep(0.05)
        }

        if (private$conn_fail && !private$has_error) {
          left <- new[leftover,]
          act <- testthat::quasi_label(rlang::enquo(test), test)
          testthat::expect(
            sum(leftover) == 0,
            sprintf(
              "%s did not close network connections: \n%s",
              encodeString(act$lab, quote = quote),
              paste(format(left), collapse = "\n")))
        }
      },

      expect_cleanup = function(test, pids, quote) {
        act <- testthat::quasi_label(rlang::enquo(test), test)
        act$pids <- length(pids)
        testthat::expect(
          length(pids) == 0,
          sprintf("%s did not clean up processes: %s",
                  encodeString(act$lab, quote = quote),
                  paste0(encodeString(names(pids), quote = "'"),
                         " (", pids, ")", collapse = ", ")))

        invisible(act$val)
      }
    ),

    private = list(
      proc_unit = NULL,
      proc_cleanup = NULL,
      proc_fail = NULL,
      proc_timeout = NULL,

      rconn_unit = NULL,
      rconn_cleanup = NULL,
      rconn_fail = NULL,
      rconns = NULL,

      file_unit = NULL,
      file_fail = NULL,
      files = NULL,

      conn_unit = NULL,
      conn_fail = NULL,
      conns = NULL,

      tree_id = NULL,

      has_error = FALSE
    )
  )
}

#' Query the type of the OS
#'
#' @return `ps_os_type` returns a named logical vector. The rest of the
#' functions return a logical scalar.
#'
#' `ps_is_supported()` returns `TRUE` if ps supports the current platform.
#'
#' @export
#' @examples
#' ps_os_type()
#' ps_is_supported()

ps_os_type <- function() {
  if (is.null(ps_env$os_type)) ps_env$os_type <- .Call(ps__os_type)
  ps_env$os_type
}

ps_os_name <- function() {
  os <- ps_os_type()
  os <- os[setdiff(names(os), c("BSD", "POSIX"))]
  names(os)[which(os)]
}

#' @rdname ps_os_type
#' @export

ps_is_supported <- function() {
  os <- ps_os_type()
  if (os[["LINUX"]]) {
    # On Linux we need to check if /proc is readable
    supported <- FALSE
    tryCatch({
      readLines("/proc/stat", warn = FALSE, n = 1)
      supported <- TRUE
    }, error = function(e) e)
    supported

  } else {
    os <- os[setdiff(names(os), c("BSD", "POSIX"))]
    any(os)
  }
}

supported_str <- function() {
  os <- ps_os_type()
  os <- os[setdiff(names(os), c("BSD", "POSIX"))]
  paste(caps(names(os)), collapse = ", ")
}

test_that("dummy", {
  expect_true(TRUE)
})

if (ps_os_type()[["LINUX"]]) {
  fun <- function() {
    withr::local_envvar(PS_WAIT_FORCE_INOTIFY = "true")
    testthat::source_file(test_path("test-wait.R"), env = environment())
  }
  fun()
}

test_that("issue #129", {
  if (!ps_os_type()[["POSIX"]]) return()
  pss <- ps(user = "root", after = as.POSIXct('2022-05-15', tz = "GMT"))
  expect_s3_class(pss, "tbl")
})

test_that("can select columns", {
  skip_on_cran()
  expect_silent(ps(user = ps_username(), columns = c("pid", "username")))
  expect_silent(ps(user = ps_username(), columns = "*"))
})

`%||%` <-  function(l, r) if (is.null(l)) r else l

`%&&%` <- function(l, r) if (is.null(l)) NULL else r

not_null <- function(x) x[!map_lgl(x, is.null)]

not_zchar <- function(x) x[x != ""]

map_chr <- function(.x, .f, ...) {
  vapply(X = .x, FUN = .f, FUN.VALUE = character(1), ...)
}

map_lgl <- function(.x, .f, ...) {
  vapply(X = .x, FUN = .f, FUN.VALUE = logical(1), ...)
}

map_int <- function(.x, .f, ...) {
  vapply(X = .x, FUN = .f, FUN.VALUE = integer(1), ...)
}

map_dbl <- function(.x, .f, ...) {
  vapply(X = .x, FUN = .f, FUN.VALUE = double(1), ...)
}

parse_envs <- function(x) {
  x <- enc2utf8(x)
  x <- strsplit(x, "=", fixed = TRUE)
  nms <- map_chr(x, "[[", 1)
  vls <- map_chr(x, function(x) paste(x[-1], collapse = "="))
  ord <- order(nms)
  structure(vls[ord], names = nms[ord], class = "Dlist")
}

## These two are fully vectorized

str_starts_with <- function(x, p) {
  ncp <- nchar(p)
  substr(x, 1, nchar(p)) == p
}

str_strip <- function(x) {
  sub("\\s+$", "", sub("^\\s+", "", x))
}

str_tail <- function(x, num) {
  nc  <- nchar(x)
  substr(x, pmax(nc - num + 1, 1), nc)
}

r_version <- function(x) {
  v <- paste0(version[["major"]], ".", version[["minor"]])
  package_version(v)
}

file_size <- function(x) {
  if (r_version() >= "3.2.0") {
    file.info(x, extra_cols = FALSE)$size
  } else {
    file.info(x)$size
  }
}

format_unix_time <- function(z) {
  structure(z, class = c("POSIXct", "POSIXt"), tzone = "GMT")
}

NA_time <- function() {
  x <- Sys.time()
  x[] <- NA
  x
}

fallback <- function(expr, alternative) {
  tryCatch(
    expr,
    error = function(e) alternative
  )
}

read_lines <- function(path) {
  suppressWarnings(con <- file(path, open = "r"))
  on.exit(close(con), add = TRUE)
  suppressWarnings(readLines(con))
}

## We need to wait until the child becomes a zombie, otherwise
## it might still be in a running state

zombie <- function() {
  if (ps_os_type()[["POSIX"]]) {
    pid <- .Call(psp__zombie)
    ps <- ps_handle(pid)
    timeout <- Sys.time() + 5
    while (ps_status(ps) != "zombie" && Sys.time() < timeout)  {
      Sys.sleep(0.05)
    }
    if (ps_status(ps) == "zombie")  pid else stop("Cannot create zombie")
  }
}

waitpid <- function(pid) {
  if (ps_os_type()[["POSIX"]]) .Call(psp__waitpid, as.integer(pid))
}

caps <- function(x) {
  paste0(toupper(substr(x, 1, 1)), tolower(substr(x, 2, nchar(x))))
}

assert_string <- function(x) {
  if (is.character(x) && length(x) == 1 && !is.na(x)) return()
  stop(ps__invalid_argument(match.call()$x,
                            " is not a string (character scalar)"))
}

assert_integer <- function(x) {
  x <- tryCatch(
    suppressWarnings(as.integer(x)),
    error = function(e) x
  )
  if (is.integer(x) && length(x) == 1 && !is.na(x)) return(x)
  stop(ps__invalid_argument(match.call()$x,
                            " is not a scalar integer"))
}

assert_character <- function(x) {
  if (is.character(x)) return()
  stop(ps__invalid_argument(match.call()$x,
                            " is not of type character"))
}

assert_pid <- function(x) {
  if (is.integer(x) && length(x) == 1 && !is.na(x)) return(x)
  if (is.numeric(x) && length(x) == 1 && !is.na(x) &&
      as.integer(x) == x) {
    return(as.integer(x))
  }
  stop(ps__invalid_argument(match.call()$x,
                            " is not a process id (integer scalar)"))
}

assert_grace <- function(x) {
  if (is.integer(x) && length(x) == 1 && !is.na(x) && x >= 0) return(x)
  if (is.numeric(x) && length(x) == 1 && !is.na(x) && x >= 0) {
    xi <- as.integer(x)
    # if x is non-zero, then return non-zero
    if (xi == 0 && x > 0) return(1)
    return(as.integer(x))
  }
  stop(ps__invalid_argument(match.call()$x,
                            " is not a non-negative integer"))
}

assert_time <- function(x) {
  if (inherits(x, "POSIXct")) return()
  stop(ps__invalid_argument(match.call()$x,
                            " must be a time stamp (POSIXt)"))
}

assert_ps_handle <- function(x) {
  if (inherits(x, "ps_handle")) return()
  stop(ps__invalid_argument(match.call()$x,
                            " must be a process handle (ps_handle)"))
}

assert_ps_handle_list <- function(x) {
  if (all(map_lgl(x, inherits, "ps_handle"))) return()
  stop(ps__invalid_argument(
    match.call()$x,
    " must be a process handle (ps_handle) or a list of process handles"
  ))
}

assert_ps_handle_or_handle_list <- function(p) {
  if (!is.list(p)) {
    assert_ps_handle(p)
    p <- list(p)
  } else {
    assert_ps_handle_list(p)
  }
  p
}

assert_flag <- function(x) {
  if (is.logical(x) && length(x) == 1 && !is.na(x)) return()
  stop(ps__invalid_argument(match.call()$x,
                            " is not a flag (logical scalar)"))
}

assert_signal <- function(x) {
  if (is.integer(x) && length(x) == 1 && !is.na(x) &&
      x %in% unlist(signals())) return()
  stop(ps__invalid_argument(match.call()$x,
                            " is not a signal number (see ?signals())"))
}

assert_nice_value <- function(x) {
  if (ps_os_type()[["POSIX"]]) {
    if (is.integer(x) && length(x) == 1 && !is.na(x) && x <= 20) return()
    stop(ps__invalid_argument(match.call()$x,
                              " is not a valid priority value"))
  } else {
    match.arg(x, ps_windows_nice_values())
  }
}

realpath <- function(x) {
  if (ps_os_type()[["WINDOWS"]])
    .Call(psw__realpath, x)
  else
    normalizePath(x)
}

get_tool <- function(prog) {
  if (ps_os_type()[["WINDOWS"]]) prog <- paste0(prog, ".exe")
  exe <- system.file(package = "ps", "bin", .Platform$r_arch, prog)
  if (exe == "") {
    pkgpath <- system.file(package = "ps")
    if (basename(pkgpath) == "inst") pkgpath <- dirname(pkgpath)
    exe <- file.path(pkgpath, "src", prog)
    if (!file.exists(exe)) return("")
  }
  exe
}

match_names <- function(map, x) {
  names(map)[match(x, map)]
}

is_cran_check <- function() {
  if (identical(Sys.getenv("NOT_CRAN"), "true")) {
    FALSE
  } else {
    Sys.getenv("_R_CHECK_PACKAGE_NAME_", "") != ""
  }
}

test_that("unit: test, mode: cleanup-fail", {
  out <- list()
  on.exit(if (!is.null(out$p)) out$p$kill(), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(proc_unit = "test"),
      {
        test_that("foobar", {
          out$p <<- processx::process$new(px(), c("sleep", "5"))
          out$running <<- out$p$is_alive()
        })
      }
    )
  )
  expect_true(status$n_failure > 0)
  fails <- map_chr(status$failures, conditionMessage)
  expect_match(fails, "did not clean up processes", all = FALSE)

  expect_true(out$running)
  deadline <- Sys.time() + 2
  while (out$p$is_alive() && Sys.time() < deadline) {
    Sys.sleep(0.05)
  }
  expect_true(Sys.time() < deadline)
  expect_false(out$p$is_alive())
})

test_that("unit: test, multiple processes", {
  out <- list()
  on.exit(if (!is.null(out$p1)) out$p1$kill(), add = TRUE)
  on.exit(if (!is.null(out$p2)) out$p2$kill(), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(proc_unit = "test"),
      {
        test_that("foobar", {
          out$p1 <<- processx::process$new(px(), c("sleep", "5"))
          out$p2 <<- processx::process$new(px(), c("sleep", "5"))
          out$running <<- out$p1$is_alive() && out$p2$is_alive()
        })
      }
    )
  )
  expect_true(status$n_failure > 0)
  fails <- map_chr(status$failures, conditionMessage)
  expect_match(fails, "did not clean up processes.*px.*px", all = FALSE)

  expect_true(out$running)
  expect_false(out$p1$is_alive())
  expect_false(out$p2$is_alive())
})

test_that("on.exit() works", {
  out <- list()
  on.exit(if (!is.null(out$p)) out$p$kill(), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(proc_unit = "test"),
      {
        test_that("foobar", {
          out$p <<- processx::process$new(px(), c("sleep", "5"))
          on.exit(out$p$kill(), add = TRUE)
          out$running <<- out$p$is_alive()
        })
      }
    )
  )
  expect_true(status$n_failure == 0)
  expect_true(status$n_success > 0)

  expect_true(out$running)
  expect_false(out$p$is_alive())
})

test_that("only report", {
  out <- list()
  on.exit(if (!is.null(out$p)) out$p$kill(), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(
        proc_unit = "test",
        proc_cleanup = FALSE,
        proc_fail = TRUE
      ),
      {
        test_that("foobar", {
          out$p <<- processx::process$new(px(), c("sleep", "5"))
          out$running <<- out$p$is_alive()
        })
      }
    )
  )
  expect_true(status$n_failure > 0)
  fails <- map_chr(status$failures, conditionMessage)
  expect_match(fails, "did not clean up processes", all = FALSE)

  expect_true(out$running)
  expect_true(out$p$is_alive())
  out$p$kill()
})

test_that("only kill", {
  out <- list()
  on.exit(if (!is.null(out$p)) out$p$kill(), add = TRUE)
  with_reporter(
    CleanupReporter(testthat::SilentReporter)$new(
      proc_unit = "test",
      proc_cleanup = TRUE,
      proc_fail = FALSE,
      conn_fail = FALSE
    ),
    {
      test_that("foobar", {
        out$p <<- processx::process$new(px(), c("sleep", "5"))
        out$running <<- out$p$is_alive()
      })
      ## It must be killed by now
      test_that("foobar2", {
        deadline <- Sys.time() + 3
        while (out$p$is_alive() && Sys.time() < deadline) {
          Sys.sleep(0.05)
        }
        out$running2 <<- out$p$is_alive()
      })
    }
  )

  expect_true(out$running)
  expect_false(out$running2)
  expect_false(out$p$is_alive())
})

test_that("unit: testsuite", {
  out <- list()
  on.exit(if (!is.null(out$p)) out$p$kill(), add = TRUE)
  expect_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(
        proc_unit = "testsuite",
        rconn_fail = FALSE,
        file_fail = FALSE,
        conn_fail = FALSE
      ),
      {
        test_that("foobar", {
          out$p <<- processx::process$new(px(), c("sleep", "5"))
          out$running <<- out$p$is_alive()
        })
        test_that("foobar2", {
          ## Still alive
          out$running2 <<- out$p$is_alive()
        })
      }
    ),
    "did not clean up processes"
  )

  expect_true(out$running)
  expect_true(out$running2)
  deadline <- Sys.time() + 3
  while (out$p$is_alive() && Sys.time() < deadline) {
    Sys.sleep(0.05)
  }
  expect_false(out$p$is_alive())
})

test_that("R connection cleanup, test, close, fail", {
  out <- list()
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  on.exit(try(close(out$conn), silent = TRUE), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(proc_fail = FALSE),
      {
        test_that("foobar", {
          out$conn <<- file(tmp, open = "w")
          out$open <<- isOpen(out$conn)
        })
      }
    )
  )
  expect_true(status$n_failure > 0)
  fails <- map_chr(status$failures, conditionMessage)
  expect_match(fails, "did not close R connections", all = FALSE)

  expect_true(out$open)
  expect_error(isOpen(out$conn))
})

test_that("R connection cleanup, test, do not close, fail", {
  out <- list()
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  on.exit(try(close(out$conn), silent = TRUE), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(
        proc_fail = FALSE,
        rconn_cleanup = FALSE
      ),
      {
        test_that("foobar", {
          out$conn <<- file(tmp, open = "w")
          out$open <<- isOpen(out$conn)
        })
      }
    )
  )
  expect_true(status$n_failure > 0)
  fails <- map_chr(status$failures, conditionMessage)
  expect_match(fails, "did not close R connections", all = FALSE)

  expect_true(out$open)
  expect_true(isOpen(out$conn))
  expect_silent(close(out$conn))
})

test_that("R connection cleanup, test, close, do not fail", {
  out <- list()
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  on.exit(try(close(out$conn), silent = TRUE), add = TRUE)
  with_reporter(
    CleanupReporter(testthat::SilentReporter)$new(
      proc_fail = FALSE,
      rconn_fail = FALSE
    ),
    {
      test_that("foobar", {
        out$conn <<- file(tmp, open = "w")
        out$open <<- isOpen(out$conn)
      })
    }
  )

  expect_true(out$open)
  expect_error(isOpen(out$conn))
})

test_that("R connections, unit: testsuite", {
  out <- list()
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  on.exit(try(close(out$conn), silent = TRUE), add = TRUE)
  expect_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(
        rconn_unit = "testsuite",
        proc_fail = FALSE,
        file_fail = FALSE,
        conn_fail = FALSE
      ),
      {
        test_that("foobar", {
          out$conn <<- file(tmp, open = "w")
          out$open <<- isOpen(out$conn)
        })
        test_that("foobar2", {
          ## Still alive
          out$open2 <<- isOpen(out$conn)
        })
      }
    ),
    "did not close R connections"
  )

  expect_true(out$open)
  expect_true(out$open2)
  expect_error(isOpen(out$conn))
})

test_that("connections already open are ignored", {
  tmp2 <- tempfile()
  on.exit(unlink(tmp2), add = TRUE)
  conn <- file(tmp2, open = "w")
  on.exit(try(close(conn), silent = TRUE), add = TRUE)

  out <- list()
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  on.exit(try(close(out$conn), silent = TRUE), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(proc_fail = FALSE),
      {
        test_that("foobar", {
          out$conn <<- file(tmp, open = "w")
          out$open <<- isOpen(out$conn)
          close(out$conn)
        })
      }
    )
  )
  expect_true(status$n_failure == 0)
  expect_true(status$n_success > 0)

  expect_error(isOpen(out$conn))
  expect_true(isOpen(conn))
  expect_silent(close(conn))
})

test_that("File cleanup, test, fail", {
  out <- list()
  tmp <- tempfile()
  cat("data\ndata2\n", file = tmp)
  on.exit(unlink(tmp), add = TRUE)
  on.exit(try(close(out$conn), silent = TRUE), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(
        proc_fail = FALSE,
        rconn_cleanup = FALSE,
        rconn_fail = FALSE
      ),
      {
        test_that("foobar", {
          out$conn <<- file(tmp, open = "r")
          out$open <<- isOpen(out$conn)
        })
      }
    )
  )
  expect_true(status$n_failure > 0)
  fails <- map_chr(status$failures, conditionMessage)
  expect_match(fails, "did not close open files", all = FALSE)

  expect_true(out$open)
  expect_true(isOpen(out$conn))
  close(out$conn)
})

test_that("File cleanup, unit: testsuite", {
  out <- list()
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  on.exit(try(close(out$conn), silent = TRUE), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(
        file_unit = "testsuite",
        proc_fail = FALSE,
        rconn_fail = FALSE,
        rconn_cleanup = FALSE,
        conn_fail = FALSE
      ),
      {
        test_that("foobar", {
          out$conn <<- file(tmp, open = "w")
          out$open <<- isOpen(out$conn)
        })
        test_that("foobar2", {
          ## Still alive
          out$open2 <<- isOpen(out$conn)
        })
      }
    )
  )
  expect_true(status$n_failure > 0)
  fails <- map_chr(status$failures, conditionMessage)
  expect_match(fails, "did not close open files", all = FALSE)

  expect_true(out$open)
  expect_true(out$open2)
  expect_true(isOpen(out$conn))
  close(out$conn)
})

test_that("files already open are ignored", {
  tmp2 <- tempfile()
  on.exit(unlink(tmp2), add = TRUE)
  conn <- file(tmp2, open = "w")
  on.exit(try(close(conn), silent = TRUE), add = TRUE)

  out <- list()
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  on.exit(try(close(out$conn), silent = TRUE), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(
        proc_fail = FALSE,
        rconn_fail = FALSE,
        rconn_cleanup = FALSE
      ),
      {
        test_that("foobar", {
          out$conn <<- file(tmp, open = "w")
          out$open <<- isOpen(out$conn)
          close(out$conn)
        })
      }
    )
  )
  expect_true(status$n_failure == 0)
  expect_true(status$n_success > 0)

  expect_error(isOpen(out$conn))
  expect_true(isOpen(conn))
  expect_silent(close(conn))
})

conn <- curl::curl(httpbin$url(), open = "r")
close(conn)

test_that("Network cleanup, test, fail", {
  skip_on_cran()
  out <- list()
  on.exit(
    {
      try(close(out$conn), silent = TRUE)
    },
    add = TRUE
  )
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(
        proc_fail = FALSE,
        rconn_cleanup = FALSE,
        rconn_fail = FALSE,
        file_fail = FALSE
      ),
      {
        test_that("foobar", {
          out$conn <<- curl::curl(httpbin$url("/drip"), open = "r")
          out$open <<- isOpen(out$conn)
        })
      }
    )
  )
  expect_true(status$n_failure > 0)
  fails <- map_chr(status$failures, conditionMessage)
  expect_match(fails, "did not close network", all = FALSE)

  expect_true(out$open)
  expect_true(isOpen(out$conn))
  close(out$conn)
})

test_that("Network cleanup, unit: testsuite", {
  skip_on_cran()
  out <- list()
  on.exit(try(close(out$conn), silent = TRUE), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(
        conn_unit = "testsuite",
        proc_fail = FALSE,
        rconn_fail = FALSE,
        rconn_cleanup = FALSE,
        file_fail = FALSE
      ),
      {
        test_that("foobar", {
          out$conn <<- curl::curl(httpbin$url("/drip"), open = "r")
          out$open <<- isOpen(out$conn)
        })
        test_that("foobar2", {
          ## Still alive
          out$open2 <<- isOpen(out$conn)
        })
      }
    )
  )
  expect_true(status$n_failure > 0)
  fails <- map_chr(status$failures, conditionMessage)
  expect_match(fails, "did not close network connections", all = FALSE)

  expect_true(out$open)
  expect_true(out$open2)
  expect_true(isOpen(out$conn))
  close(out$conn)
})

test_that("Network connections already open are ignored", {
  skip_on_cran()
  conn <- curl::curl(httpbin$url(), open = "r")
  on.exit(try(close(conn), silent = TRUE), add = TRUE)

  out <- list()
  on.exit(try(close(out$conn), silent = TRUE), add = TRUE)
  status <- capture_success_failure(
    with_reporter(
      CleanupReporter(testthat::SilentReporter)$new(
        proc_fail = FALSE,
        rconn_fail = FALSE,
        rconn_cleanup = FALSE
      ),
      {
        test_that("foobar", {
          out$conn <<- curl::curl(httpbin$url(), open = "r")
          out$open <<- isOpen(out$conn)
          close(out$conn)
        })
      }
    )
  )
  expect_true(status$n_failure == 0)
  expect_true(status$n_success > 0)

  expect_error(isOpen(out$conn))
  expect_true(isOpen(conn))
  expect_silent(close(conn))
})

# https://github.com/r-lib/ps/issues/163
test_that("errors still cause a failure", {
  rep <- CleanupReporter(testthat::SilentReporter)$new()
  expect_error(
    test_dir(
      reporter = rep,
      test_path("fixtures/cleanup-error"),
      stop_on_failure = TRUE
    )
  )
})

re_match <- function(text, pattern, perl = TRUE, ...) {
  text <- as.character(text)

  match <- regexpr(pattern, text, perl = perl, ...)

  start  <- as.vector(match)
  length <- attr(match, "match.length")
  end    <- start + length - 1L

  matchstr <- substring(text, start, end)
  matchstr[ start == -1 ] <- NA_character_

  res <- data_frame(.text = text, .match = matchstr)

  if (!is.null(attr(match, "capture.start"))) {

    gstart  <- attr(match, "capture.start")
    glength <- attr(match, "capture.length")
    gend    <- gstart + glength - 1L

    groupstr <- substring(text, gstart, gend)
    groupstr[ gstart == -1 ] <- NA_character_
    dim(groupstr) <- dim(gstart)

    res <- cbind(groupstr, res, stringsAsFactors = FALSE)
  }
  names(res) <- c(attr(match, "capture.names"), ".text", ".match")
  class(res) <- c("tbl", class(res))
  res
}

#' Ids of all processes on the system
#'
#' @return Integer vector of process ids.
#' @export

ps_pids <- function() {
  os <- ps_os_type()
  pp <- if (os[["MACOS"]])
    ps_pids_macos()
  else if (os[["LINUX"]])
    ps_pids_linux()
  else if (os[["WINDOWS"]])
    ps_pids_windows()
  else
    stop("Not implemented for this platform")

  sort(pp)
}

ps_pids_windows <- function() {
  sort(.Call(ps__pids))
}

ps_pids_macos <- function() {
  ls <- .Call(ps__pids)
  ## 0 is missing from the list, usually, even though it is a process
  if (! 0L %in% ls && ps_pid_exists_macos(0L)) {
    ls <- c(ls, 0L)
  }
  ls
}

ps_pid_exists_macos <- function(pid) {
  .Call(psp__pid_exists, as.integer(pid))
}

ps_pids_linux <- function() {
  sort(as.integer(dir("/proc", pattern = "^[0-9]+$")))
}

#' Boot time of the system
#'
#' @return A `POSIXct` object.
#'
#' @export

ps_boot_time <- function() {
  format_unix_time(.Call(ps__boot_time))
}

#' List users connected to the system
#'
#' @return A data frame with columns
#'  `username`, `tty`, `hostname`, `start_time`, `pid`. `tty` and `pid`
#'  are `NA` on Windows. `pid` is the process id of the login process.
#'  For local users the `hostname` column is the empty string.
#'
#' @export

ps_users <- function() {
  l <- not_null(.Call(ps__users))

  d <- data_frame(
    username = vapply(l, "[[", character(1), 1),
    tty = vapply(l, "[[", character(1), 2),
    hostname = vapply(l, "[[", character(1), 3),
    start_time = format_unix_time(vapply(l, "[[", double(1), 4)),
    pid = vapply(l, "[[", integer(1),  5)
  )

  d
}

#' Number of logical or physical CPUs
#'
#' If cannot be determined, it returns `NA`. It also returns `NA` on older
#' Windows systems, e.g. Vista or older and Windows Server 2008 or older.
#'
#' @param logical Whether to count logical CPUs.
#' @return Integer scalar.
#'
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' ps_cpu_count(logical = TRUE)
#' ps_cpu_count(logical = FALSE)

ps_cpu_count <- function(logical = TRUE) {
  assert_flag(logical)
  if (logical) ps_cpu_count_logical() else ps_cpu_count_physical()
}

 ps_cpu_count_logical <- function() {
   .Call(ps__cpu_count_logical)
 }

ps_cpu_count_physical <- function() {
  if (ps_os_type()[["LINUX"]]) {
    ps_cpu_count_physical_linux()
  } else {
    .Call(ps__cpu_count_physical)
  }
}

#' Query the size of the current terminal
#'
#' If the standard output of the current R process is not a terminal,
#' e.g. because it is redirected to a file, or the R process is running in
#' a GUI, then it will throw an error. You need to handle this error if
#' you want to use this function in a package.
#'
#' If an error happens, the error message is different depending on
#' what type of device the standard output is. Some common error messages
#' are:
#' * "Inappropriate ioctl for device."
#' * "Operation not supported on socket."
#' * "Operation not supported by device."
#'
#' Whatever the error message, `ps_tty_size` always fails with an error of
#' class `ps_unknown_tty_size`, which you can catch.
#'
#' @export
#' @examples
#' # An example that falls back to the 'width' option
#' tryCatch(
#'   ps_tty_size(),
#'   ps_unknown_tty_size = function(err) {
#'     c(width = getOption("width"), height = NA_integer_)
#'   }
#' )

ps_tty_size <- function() {
  tryCatch(
    ret <- .Call(ps__tty_size),
    error = function(err) {
      class(err) <- c("ps_unknown_tty_size", class(err))
      stop(err)
    }
  )
  c(width = ret[1], height = ret[2])
}

#' List all processes that loaded a shared library
#'
#' @details
#' ## Notes:
#' This function currently only works on Windows.
#'
#' On Windows, a 32 bit R process can only list other 32 bit processes.
#' Similarly, a 64 bit R process can only list other 64 bit processes.
#' This is a limitation of the Windows API.
#'
#' Even though Windows file systems are (almost always) case
#' insensitive, the matching of `paths`, `user` and also `filter`
#' are case sensitive. This might change in the future.
#'
#' This function can be very slow on Windows, because it needs to
#' enumerate all shared libraries of all processes in the system,
#' unless the `filter` argument is set. Make sure you set `filter`
#' if you can.
#'
#' If you want to look up multiple shared libraries, list all of them
#' in `paths`, instead of calling `ps_shared_lib_users` for each
#' individually.
#'
#' If you are after libraries loaded by R processes, you might want to
#' set `filter` to `c("Rgui.exe", "Rterm.exe", "rsession.exe")` The
#' last one is for RStudio.
#'
#' @param paths Character vector of paths of shared libraries to
#' look up. They must be absolute paths. They don't need to exist.
#' Forward slashes are converted to backward slashes on Windows, and
#' the output will always have backward slashes in the paths.
#' @param user Character scalar or `NULL`. If not `NULL`, then only
#' the processes of this user are considered. It defaults to the
#' current user.
#' @param filter Character vector or `NULL`. If not NULL, then it is
#' a vector of glob expressions, used to filter the process names.
#' @return A data frame with columns:
#' * `dll`: the file name of the dll file, without the path,
#' * `path`: path to the shared library,
#' * `pid`: process ID of the process,
#' * `name`: name of the process,
#' * `username`: username of process owner,
#' * `ps_handle`: `ps_handle` object, that can be used to further
#'   query and manipulate the process.
#'
#' @export
#' @family shared library tools
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check() && ps::ps_os_type()[["WINDOWS"]]
#' dlls <- vapply(getLoadedDLLs(), "[[", character(1), "path")
#' psdll <- dlls[["ps"]][[1]]
#' r_procs <- c("Rgui.exe", "Rterm.exe", "rsession.exe")
#' ps_shared_lib_users(psdll, filter = r_procs)

ps_shared_lib_users <- function(paths, user = ps_username(),
                                filter = NULL) {
  os <- ps_os_type()
  if (!os[["WINDOWS"]]) {
    stop("`ps_shared_lib_users()` currently only works on Windows")
  }
  assert_character(paths)
  if (!is.null(user)) assert_string(user)
  if (!is.null(filter)) assert_character(filter)
  if (os[["WINDOWS"]]) paths <- gsub("/", "\\", paths, fixed = TRUE)

  pids <- ps_pids()
  processes <- not_null(lapply(pids, function(p) {
    tryCatch(ps_handle(p), error = function(e) NULL) }))

  nm <- map_chr(processes, function(p)
    fallback(ps_name(p), NA_character_))

  if (!is.null(filter)) {
    selected <- glob$test_any(filter, nm)
    processes <- processes[selected]
    nm <- nm[selected]
  }

  us <- map_chr(processes, function(p)
    fallback(ps_username(p), NA_character_))

  if (!is.null(user)) {
    us2 <- short_username(us)
    selected <- (!is.na(us) & us == user) | (!is.na(us2) & us2 == user)
    processes <- processes[selected]
    nm <- nm[selected]
    us <- us[selected]
  }

  libs <- lapply(processes, function(p) {
    tryCatch(ps_shared_libs(p)$path, error = function(e) character())
  })

  # TODO: handle case insensitive OS/FS

  match <- lapply(libs, intersect, paths)
  match_len <- map_int(match, length)
  match_processes <- processes[match_len > 0]
  match_username <- us[match_len > 0]
  match_name <- nm[match_len > 0]
  match_len <- match_len[match_len > 0]
  match_pids <- map_int(match_processes, ps_pid)

  d <- data_frame(
    dll = basename(unlist(match)),
    path = unlist(match),
    pid = rep(match_pids, match_len),
    name = rep(match_name, match_len),
    username = rep(match_username, match_len),
    ps_handle = I(rep(match_processes, match_len))
  )

  # The ones without name probably finished already.
  d <- d[!is.na(d$name), , drop = FALSE]

  d
}

short_username <- function(x) {
  xs <- strsplit(x, "\\", fixed = TRUE)
  p1 <- map_chr(xs, "[", 1)
  p2 <- map_chr(xs, "[", 2)
  ifelse(!is.na(p2), p2, x)
}

# Docs from psutil, thanks!

#' Return the average system load over the last 1, 5 and 15 minutes as a
#' tuple.
#'
#' The “load” represents the processes which are in a runnable
#' state, either using the CPU or waiting to use the CPU (e.g. waiting for
#' disk I/O). On Windows this is emulated by using a Windows API that
#' spawns a thread which keeps running in background and updates results
#' every 5 seconds, mimicking the UNIX behavior. Thus, on Windows, the
#' first time this is called and for the next 5 seconds it will return a
#' meaningless (0.0, 0.0, 0.0) vector. The numbers returned only make sense
#' if related to the number of CPU cores installed on the system. So, for
#' instance, a value of 3.14 on a system with 10 logical CPUs means that
#' the system load was 31.4% percent over the last N minutes.
#'
#' @return Numeric vector of length 3.
#'
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' ps_loadavg()

ps_loadavg <- function() {
  if (is.null(ps_env$counter_name)) {
    if (ps_os_type()[["WINDOWS"]]) {
      ps_env$counter_name <- find_loadavg_counter()
    } else {
      ps_env$counter_name <- ""
    }
  }

  .Call(ps__loadavg, ps_env$counter_name)
}

find_loadavg_counter <- function() {
  key <- paste0(
    "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Perflib\\",
    "CurrentLanguage"
  )
  tryCatch({
    pc <- utils::readRegistry(key)
    idx <- seq(2, length(pc$Counter), by = 2)
    cnt <- structure(pc$Counter[idx], names = pc$Counter[idx - 1])
    nm <- paste0("\\", cnt["2"], "\\", cnt["44"])
    Encoding(nm) <- ""
    enc2utf8(nm)
  }, error = function(e) "\\System\\Processor Queue Length")
}

#' System CPU times.
#'
#' Every attribute represents the seconds the CPU has spent in the given
#' mode. The attributes availability varies depending on the platform:
#' * `user`: time spent by normal processes executing in user mode;
#'   on Linux this also includes guest time.
#' * `system`: time spent by processes executing in kernel mode.
#' * `idle`: time spent doing nothing.
#'
#' Platform-specific fields:
#'
#' * `nice` (UNIX): time spent by niced (prioritized) processes executing
#'   in user mode; on Linux this also includes guest_nice time.
#' * `iowait` (Linux): time spent waiting for I/O to complete. This is not
#'   accounted in idle time counter.
#' * `irq` (Linux): time spent for servicing hardware interrupts.
#' * `softirq` (Linux): time spent for servicing software interrupts.
#' * `steal` (Linux 2.6.11+): time spent by other operating systems
#'   running in a virtualized environment.
#' * `guest` (Linux 2.6.24+): time spent running a virtual CPU for guest
#'   operating systems under the control of the Linux kernel.
#' * `guest_nice` (Linux 3.2.0+): time spent running a niced guest
#'   (virtual CPU for guest operating systems under the control of the
#'   Linux kernel).
#'
#' @return Named list
#'
#' @export
#' @examplesIf ps::ps_is_supported()
#' ps_system_cpu_times()

ps_system_cpu_times <- function() {
  os <- ps_os_name()
  if (os == "LINUX") {
    ps__system_cpu_times_linux()
  } else {
    .Call(ps__system_cpu_times)
  }
}

#' List currently running applications
#'
#' This function currently only works on macOS.
#'
#' @return A data frame with columns:
#'   - `pid`: integer process id.
#'   - `name`: process name.
#'   - `bundle_identifier`: bundle identifier, e.g. `com.apple.dock`.
#'   - `bundle_url`: bundle URL, a `file://` URL to the app bundle.
#'   - `arch`: executable architecture, possible values are
#'     `r paste(macos_archs$name, collapse = ", ")`.
#'   - `executable_url`: `file://` URL to the executable file.
#'   - `launch_date`: launch time stamp, a `POSIXct` object, may be `NA`.
#'   - `finished_launching`: whether the app has finished launching.
#'   - `active`: whether the app is active.
#'   - `activation_policy`: one of the following values:
#'     * `regular`: the application is an ordinary app that appears in the
#'       Dock and may have a user interface.
#'     * `accessory`: the application doesn’t appear in the Dock and
#'       doesn’t have a menu bar, but it may be activated programmatically
#'       or by clicking on one of its windows.
#'     * `prohibited`: the application doesn’t appear in the Dock and may
#'       not create windows or be activated.
#'
#' @export
#' @examplesIf ps_is_supported() && ps_os_type()[["MACOS"]] && !ps:::is_cran_check()
#' ps_apps()

ps_apps <- function() {
  if (!ps_os_type()[["MACOS"]]) {
    stop("'ps_apps()' is only implemented on macOS")
  }
  tab <- as_data_frame(.Call(ps__list_apps))

  tab <- tab[, c("pid", setdiff(names(tab), "pid"))]
  tab[["arch"]] <- macos_archs$name[match(tab[["arch"]], macos_archs$code)]
  tab[["launch_date"]] <- parse_iso_8601(
    sub(" +", "+", fixed = TRUE, tab[["launch_date"]])
  )
  # drop the ones without a pid, they are clearly (?) not running
  tab <- tab[tab$pid != -1, ]
  tab
}

macos_archs <- data.frame(
  name = c("arm64",    "i386",     "x86_64",   "ppc",      "ppc64"),
  code = c(0x0100000c, 0x00000007, 0x01000007, 0x00000012, 0x01000012)
)

ps_status_macos_ps <- function(pids) {
  stopifnot(is.integer(pids))
  suppressWarnings(tryCatch({
    out <- system2(
      "/bin/ps",
      c("-o", "pid,stat", "-p", paste(pids, collapse = ",")),
      stdout = TRUE,
      stderr = FALSE
    )
    out <- out[-1]
    out <- trimws(out)
    out2 <- strsplit(out, " ")
    opids <- map_int(out2, function(x) as.integer(x[[1]]))
    state <- map_chr(out2, function(x) substr(x[[2]], 1, 1))
    state <- macos_process_states[state]
    unname(state[match(pids, opids)])
  }, error = function(e) rep(NA_character_, length(pids))))
}

# From `man 1 ps`
# I       Marks a process that is idle (sleeping for longer than about 20 seconds).
# R       Marks a runnable process.
# S       Marks a process that is sleeping for less than about 20 seconds.
# T       Marks a stopped process.
# U       Marks a process in uninterruptible wait.
# Z       Marks a dead process (a “zombie”).

macos_process_states <- c(
  I = "idle",
  R = "running",
  S = "sleeping",
  T = "stopped",
  U = "uninterruptible",
  Z = "zombie"
)

ps_env <- new.env(parent = emptyenv())

Internal <- NULL

## nocov start
.onLoad <- function(libname, pkgname) {
  ps_env$constants <- new.env(parent  = emptyenv())
  .Call(ps__init, asNamespace("ps"), ps_env$constants)
  if (!is.null(ps_env$constants$signals)) {
    ps_env$constants$signals <- as.list(ps_env$constants$signals)
  }
  if (!is.null(ps_env$constants$errno)) {
    ps_env$constants$errno <- as.list(ps_env$constants$errno)
  }
  if (!is.null(ps_env$constants$address_families)) {
    ps_env$constants$address_families <-
      as.list(ps_env$constants$address_families)
  }
  if (!is.null(ps_env$constants$socket_types)) {
    ps_env$constants$socket_types <-
      as.list(ps_env$constants$socket_types)
  }

  Internal <<- get(".Internal", asNamespace("base"))

  ps_boot_time <<- memoize(ps_boot_time)
  ps_cpu_count_logical <<- memoize(ps_cpu_count_logical)  
  ps_cpu_count_physical <<- memoize(ps_cpu_count_physical)  
  get_terminal_map <<- memoize(get_terminal_map)
  NA_time <<- memoize(NA_time)
}
## nocov end

utils::globalVariables(c("self", "super"))

milliseconds <- function(x) as.difftime(as.numeric(x) / 1000, units = "secs")
seconds <- function(x) as.difftime(as.numeric(x), units = "secs")
minutes <- function(x) as.difftime(as.numeric(x), units = "mins")
hours <- function(x) as.difftime(as.numeric(x), units = "hours")
days <- function(x) as.difftime(as.numeric(x), units = "days")
weeks <- function(x) as.difftime(as.numeric(x), units = "weeks")
wday <- function(x) as.POSIXlt(x, tz = "UTC")$wday + 1
with_tz <- function(x, tzone = "") as.POSIXct(as.POSIXlt(x, tz = tzone))
ymd <- function(x) as.POSIXct(x, format = "%Y %m %d", tz = "UTC")
yj <- function(x) as.POSIXct(x, format = "%Y %j", tz = "UTC")

parse_iso_8601 <- function(dates, default_tz = "UTC") {
  if (default_tz == "") default_tz <- Sys.timezone()
  dates <- as.character(dates)
  match <- re_match(dates, iso_regex)
  matching <- !is.na(match$.match)
  result <- rep(.POSIXct(NA_real_, tz = ""), length.out = length(dates))
  result[matching] <- parse_iso_parts(match[matching, ], default_tz)
  class(result) <- c("POSIXct", "POSIXt")
  with_tz(result, "UTC")
}

parse_iso_parts <- function(mm, default_tz) {

  num <- nrow(mm)

  ## -----------------------------------------------------------------
  ## Date first

  date <- .POSIXct(rep(NA_real_, num), tz = "")

  ## Years-days
  fyd <- is.na(date) & mm$yearday != ""
  date[fyd] <- yj(paste(mm$year[fyd], mm$yearday[fyd]))

  ## Years-weeks-days
  fywd <- is.na(date) & mm$week != "" & mm$weekday != ""
  date[fywd] <- iso_week(mm$year[fywd], mm$week[fywd], mm$weekday[fywd])

  ## Years-weeks
  fyw <- is.na(date) & mm$week != ""
  date[fyw] <- iso_week(mm$year[fyw], mm$week[fyw], "1")

  ## Years-months-days
  fymd <- is.na(date) & mm$month != "" & mm$day != ""
  date[fymd] <- ymd(paste(mm$year[fymd], mm$month[fymd], mm$day[fymd]))

  ## Years-months
  fym <- is.na(date) & mm$month != ""
  date[fym] <- ymd(paste(mm$year[fym], mm$month[fym], "01"))

  ## Years
  fy <- is.na(date)
  date[fy] <- ymd(paste(mm$year, "01", "01"))

  ## -----------------------------------------------------------------
  ## Now the time

  th <- mm$hour != ""
  date[th] <- date[th] + hours(mm$hour[th])

  tm <- mm$min != ""
  date[tm] <- date[tm] + minutes(mm$min[tm])

  ts <- mm$sec != ""
  date[ts] <- date[ts] + seconds(mm$sec[ts])

  ## -----------------------------------------------------------------
  ## Fractional time

  frac <- as.numeric(sub(",", ".", mm$frac))

  tfs <- !is.na(frac) & mm$sec != ""
  date[tfs] <- date[tfs] + milliseconds(round(frac[tfs] * 1000))

  tfm <- !is.na(frac) & mm$sec == "" & mm$min != ""
  sec <- trunc(frac[tfm] * 60)
  mil <- round((frac[tfm] * 60 - sec) * 1000)
  date[tfm] <- date[tfm] + seconds(sec) + milliseconds(mil)

  tfh <- !is.na(frac) & mm$sec == "" & mm$min == ""
  min <- trunc(frac[tfh] * 60)
  sec <- trunc((frac[tfh] * 60 - min) * 60)
  mil <- round((((frac[tfh] * 60) - min) * 60 - sec) * 1000)
  date[tfh] <- date[tfh] + minutes(min) + seconds(sec) + milliseconds(mil)

  ## -----------------------------------------------------------------
  ## Time zone

  ftzpm <- mm$tzpm != ""
  m <- ifelse(mm$tzpm[ftzpm] == "+", -1, 1)
  ftzpmh <- ftzpm & mm$tzhour != ""
  date[ftzpmh] <- date[ftzpmh] + m * hours(mm$tzhour[ftzpmh])
  ftzpmm <- ftzpm & mm$tzmin != ""
  date[ftzpmm] <- date[ftzpmm] + m * minutes(mm$tzmin[ftzpmm])

  ftzz <- mm$tz == "Z"
  date[ftzz] <- as.POSIXct(date[ftzz], "UTC")

  ftz <- mm$tz != "Z" & mm$tz != ""
  date[ftz] <- as.POSIXct(date[ftz], mm$tz[ftz])

  if (default_tz != "UTC") {
    ftna <- mm$tzpm == "" & mm$tz == ""
    if (any(ftna)) {
      dd <- as.POSIXct(format_iso_8601(date[ftna]),
                       "%Y-%m-%dT%H:%M:%S+00:00", tz = default_tz)
      date[ftna] <- dd
    }
  }

  as.POSIXct(date, "UTC")
}

iso_regex <- paste0(
  "^\\s*",
  "(?<year>[\\+-]?\\d{4}(?!\\d{2}\\b))",
  "(?:(?<dash>-?)",
   "(?:(?<month>0[1-9]|1[0-2])",
    "(?:\\g{dash}(?<day>[12]\\d|0[1-9]|3[01]))?",
    "|W(?<week>[0-4]\\d|5[0-3])(?:-?(?<weekday>[1-7]))?",
    "|(?<yearday>00[1-9]|0[1-9]\\d|[12]\\d{2}|3",
      "(?:[0-5]\\d|6[1-6])))",
   "(?<time>[T\\s](?:(?:(?<hour>[01]\\d|2[0-3])",
            "(?:(?<colon>:?)(?<min>[0-5]\\d))?|24\\:?00)",
           "(?<frac>[\\.,]\\d+(?!:))?)?",
    "(?:\\g{colon}(?<sec>[0-5]\\d)(?:[\\.,]\\d+)?)?",
    "(?<tz>[zZ]|(?<tzpm>[\\+-])",
     "(?<tzhour>[01]\\d|2[0-3]):?(?<tzmin>[0-5]\\d)?)?)?)?$"
  )

iso_week <- function(year, week, weekday) {

  wdmon <- function(date) { (wday(date) + 5L) %% 7L }
  thu <- function(date) { date - days(wdmon(date) - 3L) }

  thu(ymd(paste(year, "01", "04"))) + weeks(as.numeric(week) - 1L) +
    days(as.numeric(weekday) - 4L)
}

format_iso_8601 <- function(date) {
  format(as.POSIXlt(date, tz = "UTC"), "%Y-%m-%dT%H:%M:%S+00:00")
}

#' List all mounted partitions
#'
#' The output is similar the Unix `mount` and `df` commands.
#'
#' @param all Whether to list virtual devices as well. If `FALSE`, on
#' Linux it will still list `overlay` and `grpcfuse` file systems, to
#' provide some useful information in Docker containers.
#' @return A data frame with columns `device`, `mountpoint`,
#' `fstype` and `options`.
#'
#' @family disk functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' ps_disk_partitions(all = TRUE)
#' ps_disk_partitions()

ps_disk_partitions <- function(all = FALSE) {
  assert_flag(all)
  l <- not_null(.Call(ps__disk_partitions, all))

  d <- data_frame(
    device = vapply(l, "[[", character(1), 1),
    mountpoint = vapply(l, "[[", character(1), 2),
    fstype = vapply(l, "[[", character(1), 3),
    options = vapply(l, "[[", character(1), 4)
  )

  if (!all) d <- ps__disk_partitions_filter(d)

  d
}

#' @importFrom utils read.delim

ps__disk_partitions_filter <- function(pt) {
  os <- ps_os_name()

  if (os == "LINUX") {
    fs <- read.delim("/proc/filesystems", header = FALSE, sep = "\t")
    goodfs <- c(fs[[2]][fs[[1]] != "nodev"], "zfs")
    ok <- pt$device != "none" & file.exists(pt$device) & pt$fstype %in% goodfs
    ok <- ok | pt$device %in% c("overlay", "grpcfuse")
    pt <- pt[ok, , drop = FALSE]

  } else if (os == "MACOS") {
    ok <- substr(pt$device, 1, 1) == "/" & file.exists(pt$device)
    pt <- pt[ok, , drop = FALSE]
  }

  pt
}

#' Disk usage statistics, per partition
#'
#' The output is similar to the Unix `df` command.
#'
#'
#' Note that on Unix a small percentage of the disk space (5% typically)
#' is reserved for the superuser. `ps_disk_usage()` returns the space
#' available to the calling user.
#'
#' @param paths The mounted file systems to list. By default all file
#' systems returned by [ps_disk_partitions()] is listed.
#' @return A data frame with columns `mountpoint`, `total`, `used`,
#' `available` and `capacity`.
#'
#' @family disk functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' ps_disk_usage()


ps_disk_usage <- function(paths = ps_disk_partitions()$mountpoint) {
  assert_character(paths)
  l <- .Call(ps__disk_usage, paths)
  os <- ps_os_name()
  if (os == "WINDOWS") {
    ps__disk_usage_format_windows(paths, l)
  } else {
    ps__disk_usage_format_posix(paths, l)
  }
}

ps__disk_usage_format_windows <- function(paths, l) {
  total <- vapply(l, "[[", double(1), 1)
  free <- vapply(l, "[[", double(1), 2)
  freeuser <- vapply(l, "[[", double(1), 3)
  used <- total - free

  d <- data_frame(
    mountpoint = paths,
    total = total,
    used = used,
    available = freeuser,
    capacity = used / total
  )

  d
}

ps__disk_usage_format_posix <- function(paths, l) {
  l2 <- lapply(l, function(fs) {
    total <- fs[[5]] * fs[[1]]
    avail_to_root <- fs[[6]] * fs[[1]]
    avail = fs[[7]] * fs[[1]]
    used  <-  total - avail_to_root
    total_user <- used + avail
    usage_percent <- used / total_user
    list(total = total, used = used, free = avail, percent = usage_percent)
  })

  d <- data_frame(
    mountpoint = paths,
    total = vapply(l2, "[[", double(1), "total"),
    used = vapply(l2, "[[", double(1), "used"),
    available = vapply(l2, "[[", double(1), "free"),
    capacity = vapply(l2, "[[", double(1), "percent")
  )

  d
}

#' System-wide disk I/O counters
#'
#' Returns a data.frame of system-wide disk I/O counters.
#'
#' Includes the following non-NA fields for all supported platforms:
#' * `read_count`: number of reads
#' * `write_count`: number of writes
#' * `read_bytes`: number of bytes read
#' * `write_bytes`: number of bytes written
#'
#' And for only some platforms:
#' * `read_time`: time spent reading from disk (in milliseconds)
#' * `write_time`: time spent writing to disk (in milliseconds)
#' * `busy_time`: time spent doing actual I/Os (in milliseconds)
#' * `read_merged_count`: number of merged reads (see iostats doc)
#' * `write_merged_count`: number of merged writes (see iostats doc)
#'
#' @return A data frame of one row per disk of I/O stats, with columns
#' `name`, `read_count` `read_merged_count` `read_bytes`, `read_time`,
#' `write_count`, `write_merged_count`, `write_bytes` `write_time`, and
#' `busy_time`.
#'
#' @family disk functions
#' @export
#' @examplesIf ps::ps_is_supported() && ps:::ps_os_name() %in% c("LINUX", "WINDOWS") && !ps:::is_cran_check()
#' ps_disk_io_counters()
ps_disk_io_counters <- function() {
  os <- ps_os_name()
  tab <- if (os == "LINUX") {
    ps__disk_io_counters_linux()
  } else if (os == "WINDOWS") {
    ps__disk_io_counters_windows()
  } else {
    ps__disk_io_counters_macos()
  }

  class(tab) <- c("tbl", "data.frame")
  tab
}

ps__disk_io_counters_windows <- function() {
  l <- .Call(ps__disk_io_counters)
  disk_info <- data_frame(
    name = l[[1]],
    read_count = l[[2]],
    read_merged_count = NA,
    read_bytes = l[[3]],
    read_time = l[[4]],
    write_count = l[[5]],
    write_merged_count = NA,
    write_bytes = l[[6]],
    write_time = l[[7]],
    busy_time = NA
  )

  disk_info[disk_info$name != "",]
}

ps__disk_io_counters_macos <- function() {
  tab <- not_null(.Call(ps__disk_io_counters))
  data_frame(
    name = names(tab),
    read_count = map_dbl(tab, "[[", 1),
    read_merged_count = NA,
    read_bytes = map_dbl(tab, "[[", 3),
    read_time = map_dbl(tab, "[[", 5),
    write_count = map_dbl(tab, "[[", 2),
    write_merged_count = NA,
    write_bytes = map_dbl(tab, "[[", 4),
    write_time = map_dbl(tab, "[[", 6),
    busy_time = NA
  )
}

#' File system information for files
#'
#' @param paths A path or a vector of paths. `ps_fs_info()` returns
#'   information about the file systems of all paths. `path` may contain
#'   direcories as well.
#' @return Data frame with file system information for each
#'   path in `paths`, one row per path. Common columns for all
#'   operating systems:
#'   * `path`: The input paths, i.e. the `paths` argument.
#'   * `mountpoint`: Directory where the file system is mounted.
#'     On Linux there is a small chance that it was not possible to
#'     look this up, and it is `NA_character_`. This is the drive letter
#'     or the mount directory on Windows, with a trailing `\`.
#'   * `name`: Device name.
#'     On Linux there is a small chance that it was not possible to
#'     look this up, and it is `NA_character_`. On Windows this is the
#'     volume GUID path of the form `\\?\Volume{GUID}\`.
#'   * `type`: File system type (character).
#'     On Linux there is a tiny chance that it was not possible to
#'     look this up, and it is `NA_character_`.
#'   * `block_size`: File system block size. This is the sector size on
#'     Windows, in bytes.
#'   * `transfer_block_size`: Pptimal transfer block size. On Linux it is
#'     currently always the same as `block_size`. This is the cluster size
#'     on Windows, in bytes.
#'   * `total_data_blocks`: Total data blocks in file system. On Windows
#'     this is the number of sectors.
#'   * `free_blocks`: Free blocks in file system. On Windows this is the
#'     number of free sectors.
#'   * `free_blocks_non_superuser`: Free blocks for a non-superuser, which
#'     might be different on Unix. On Windows this is the number of free
#'     sectors for the calling user.
#'   * `id`: File system id. This is a raw vector. On Linux it is
#'     often all zeros. It is always `NULL` on Windows.
#'   * `owner`: User that mounted the file system. On Linux and Windows
#'     this is currently always `NA_real_`.
#'   * `type_code`: Type of file system, a numeric code. On Windows this
#'     this is `NA_real_`.
#'   * `subtype_code`: File system subtype (flavor). On Linux and Windows
#'     this is always `NA_real_`.
#'
#'   The rest of the columns are flags, and they are operating system
#'   dependent.
#'
#'   macOS:
#'
#'   * `RDONLY`: A read-only filesystem.
#'   * `SYNCHRONOUS`: File system is written to synchronously.
#'   * `NOEXEC`: Can't exec from filesystem.
#'   * `NOSUID`: Setuid bits are not honored on this filesystem.
#'   * `NODEV`: Don't interpret special files.
#'   * `UNION`: Union with underlying filesysten.
#'   * `ASYNC`: File system written to asynchronously.
#'   * `EXPORTED`: File system is exported.
#'   * `LOCAL`: File system is stored locally.
#'   * `QUOTA`: Quotas are enabled on this file system.
#'   * `ROOTFS`: This file system is the root of the file system.
#'   * `DOVOLFS`: File system supports volfs.
#'   * `DONTBROWSE`: File system is not appropriate path to user data.
#'   * `UNKNOWNPERMISSIONS`:  VFS will ignore ownership information on
#'     filesystem filesystemtem objects.
#'   * `AUTOMOUNTED`: File system was mounted by automounter.
#'   * `JOURNALED`: File system is journaled.
#'   * `DEFWRITE`: File system should defer writes.
#'   * `MULTILABEL`: MAC support for individual labels.
#'   * `CPROTECT`: File system supports per-file encrypted data protection.
#'
#'   Linux:
#'
#'   * `MANDLOCK`: Mandatory locking is permitted on the filesystem
#'     (see `fcntl(2)`).
#'   * `NOATIME`: Do not update access times; see `mount(2)`.
#'   * `NODEV`: Disallow access to device special files on this filesystem.
#'   * `NODIRATIME`: Do not update directory access times; see mount(2).
#'   * `NOEXEC`: Execution of programs is disallowed on this filesystem.
#'   * `NOSUID`: The set-user-ID and set-group-ID bits are ignored by
#'      `exec(3)` for executable files on this filesystem
#'   * `RDONLY`: This filesystem is mounted read-only.
#'   * `RELATIME`: Update atime relative to mtime/ctime; see `mount(2)`.
#'   * `SYNCHRONOUS`: Writes are synched to the filesystem immediately
#'     (see the description of `O_SYNC` in `open(2)``).
#'   * `NOSYMFOLLOW`: Symbolic links are not followed when resolving paths;
#'     see `mount(2)``.
#'
#'   Windows:
#'
#'   * `CASE_SENSITIVE_SEARCH`: Supports case-sensitive file names.
#'   * `CASE_PRESERVED_NAMES`: Supports preserved case of file names when
#'      it places a name on disk.
#'   * `UNICODE_ON_DISK`: Supports Unicode in file names as they appear on
#'      disk.
#'   * `PERSISTENT_ACLS`: Preserves and enforces access control lists
#'      (ACL). For example, the NTFS file system preserves and enforces
#'      ACLs, and the FAT file system does not.
#'   * `FILE_COMPRESSION`: Supports file-based compression.
#'   * `VOLUME_QUOTAS`: Supports disk quotas.
#'   * `SUPPORTS_SPARSE_FILES`: Supports sparse files.
#'   * `SUPPORTS_REPARSE_POINTS`: Supports reparse points.
#'   * `SUPPORTS_REMOTE_STORAGE`: Supports remote storage.
#'   * `RETURNS_CLEANUP_RESULT_INFO`: On a successful cleanup operation,
#'      the file system returns information that describes additional
#'      actions taken during cleanup, such as deleting the file. File
#'      system filters can examine this information in their post-cleanup
#'      callback.
#'   * `SUPPORTS_POSIX_UNLINK_RENAME`: Supports POSIX-style delete and
#'      rename operations.
#'   * `VOLUME_IS_COMPRESSED`: It is a compressed volume, for example, a
#'      DoubleSpace volume.
#'   * `SUPPORTS_OBJECT_IDS`: Supports object identifiers.
#'   * `SUPPORTS_ENCRYPTION`: Supports the Encrypted File System (EFS).
#'   * `NAMED_STREAMS`: Supports named streams.
#'   * `READ_ONLY_VOLUME`: It is read-only.
#'   * `SEQUENTIAL_WRITE_ONCE`: Supports a single sequential write.
#'   * `SUPPORTS_TRANSACTIONS`: Supports transactions.
#'   * `SUPPORTS_HARD_LINKS`: The volume supports hard links.
#'   * `SUPPORTS_EXTENDED_ATTRIBUTES`: Supports extended attributes.
#'   * `SUPPORTS_OPEN_BY_FILE_ID`: Supports open by FileID.
#'   * `SUPPORTS_USN_JOURNAL`: Supports update sequence number (USN)
#'      journals.
#'   * `SUPPORTS_INTEGRITY_STREAMS`: Supports integrity streams.
#'   * `SUPPORTS_BLOCK_REFCOUNTING`: The volume supports sharing logical
#'      clusters between files on the same volume.
#'   * `SUPPORTS_SPARSE_VDL`: The file system tracks whether each cluster
#'      of a file contains valid data (either from explicit file writes or
#'      automatic zeros) or invalid data (has not yet been written to or
#'      zeroed).
#'   * `DAX_VOLUME`: The volume is a direct access (DAX) volume.
#'   * `SUPPORTS_GHOSTING`: Supports ghosting.
#'
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' ps_fs_info(c("/", "~", "."))

ps_fs_info <- function(paths = "/") {
  assert_character(paths)
  abspaths <- normalizePath(paths, mustWork = TRUE)
  mps <- ps_fs_mount_point(paths)
  res <- .Call(ps__fs_info, paths, abspaths, mps)
  df <- as_data_frame(res)

  # this should not happen in practice, but just in case
  if (ps_os_type()[["LINUX"]] && any(is.na(df$type))) {
    miss <- which(is.na(df$type))
    df$type[miss] <- linux_fs_types$name[match(df$type_code[miss], linux_fs_types$id)]
  }

  df
}

linux_fs_types <- utils::read.table(
  "tools/linux-fs-types.txt",
  header = TRUE,
  stringsAsFactors = FALSE
)

posix_stat_types <- c(
  "regular file",
  "directory",
  "character device",
  "block device",
  "FIFO",
  "symbolic link",
  "socket"
)

#' File status
#'
#' This function is currently not implemented on Windows.
#'
#' @param paths Paths to files, directories, devices, etc. They must
#'   exist. They are expanded using [base::path.expand()].
#' @param follow Whether to follow symbolic links. If `FALSE` it returns
#'   information on the links themselves.
#' @return Data frame with one row for each path in `paths`. Columns:
#'   * `path`: Expanded `paths`.
#'   * `dev_major`: Major device ID of the device the path resides on.
#'   * `dev_minor`: Minor device ID of the device the path resodes on.
#'   * `inode`: Inode number.
#'   * `mode`: File type and mode (permissions). It is easier to use the
#'     `type` and `permissions` columns.
#'   * `type`: File type, character. One of
#'     `r paste(posix_stat_types, collapse = ", ")`.
#'   * `permissions`: Permissions, numeric code in an integer column.
#'   * `nlink`: Number of hard links.
#'   * `uid`: User id of owner.
#'   * `gid`: Group id of owner.
#'   * `rdev_major`: If the path is a device, its major device id,
#'     otherwise `NA_integer_`.
#'   * `rdev_minor`: IF the path is a device, its minor device id,
#'     otherwise `NA_integer_`.
#'   * `size`: File size in bytes.
#'   * `block_size`: Block size for filesystem I/O.
#'   * `blocks`: Number of 512B blocks allocated.
#'   * `access_time`: Time of last access.
#'   * `modification_time`: Time of last modification.
#'   * `change_time`: Time of last status change.
#'
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check() && ps_os_type()[["POSIX"]]
#' ps_fs_stat(c(".", tempdir()))

ps_fs_stat <- function(paths, follow = TRUE) {
  assert_character(paths)
  paths <- path.expand(paths)
  res <- .Call(ps__stat, paths, follow)
  res[["type"]] <- posix_stat_types[res[["type"]]]
  res[["access_time"]] <- .POSIXct(res[["access_time"]], "UTC")
  res[["modification_time"]] <- .POSIXct(res[["modification_time"]], "UTC")
  res[["change_time"]] <- .POSIXct(res[["change_time"]], "UTC")
  as_data_frame(res)
}

#' Find the mount point of a file or directory
#'
#' @param paths Paths to files, directories, devices, etc. They must
#'   exist. They are normalized using [base::normalizePath()].
#' @return Character vector, paths to the mount points of the input
#'   `paths`.
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' ps_fs_mount_point(".")

ps_fs_mount_point <- function(paths) {
  assert_character(paths)
  paths <- normalizePath(paths, mustWork = TRUE)
  call_with_cleanup(ps__mount_point, paths)
}

#' List of 'errno' error codes
#'
#' For the errors that are not used on the current platform, `value` is
#' `NA_integer_`.
#'
#' A data frame with columns: `name`, `value`, `description`.
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' errno()

errno <- function() {
  err <- as.list(ps_env$constants$errno)
  err <- err[order(names(err))]
  data_frame(
    name = names(err),
    value = vapply(err, "[[", integer(1), 1),
    description = vapply(err, "[[", character(1), 2)
  )
}

#' Statistics about system memory usage
#'
#' @return Named list. All numbers are in bytes:
#' * `total`: total physical memory (exclusive swap).
#' * `avail` the memory that can be given instantly to processes without
#'   the system going into swap. This is calculated by summing different
#'   memory values depending on the platform and it is supposed to be used
#'   to monitor actual memory usage in a cross platform fashion.
#' * `percent`: Percentage of memory that is taken.
#' * `used`: memory used, calculated differently depending on
#'   the platform and designed for informational purposes only.
#'   `total` - `free` does not necessarily match `used`.
#' * `free`: memory not being used at all (zeroed) that is
#'   readily available; note that this doesn’t reflect the actual memory
#'   available (use `available` instead). `total` - `used` does not
#'   necessarily match `free`.
#' * `active`: (Unix only) memory currently in use or very recently used,
#'   and so it is in RAM.
#' * `inactive`: (Unix only) memory that is marked as not used.
#' * `wired`: (macOS only) memory that is marked to always stay in RAM. It
#'   is never moved to disk.
#' * `buffers`: (Linux only) cache for things like file system metadata.
#' * `cached`: (Linux only) cache for various things.
#' * `shared`: (Linux only) memory that may be simultaneously accessed by
#'   multiple processes.
#' * `slab`:  (Linux only) in-kernel data structures cache.
#'
#' @family memory functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' ps_system_memory()

ps_system_memory <- function() {
  os <- ps_os_name()

  if (os == "MACOS") {
    l <- .Call(ps__system_memory)
    l$avail <- l$inactive + l$free
    l$used <- l$active + l$wired
    l$free <- l$free - l$speculative
    l$percent <- (l$total - l$avail) / l$total * 100
    l[c("total", "avail", "percent", "used", "free",
        "active", "inactive", "wired")]

  } else if (os == "LINUX") {
    ps__system_memory_linux()

  } else if (os == "WINDOWS") {
    l <- .Call(ps__system_memory)[c("total", "avail")]
    l$free <- l$avail
    l$used <- l$total - l$avail
    l$percent <- (l$total - l$avail) * 100 / l$total
    l[c("total", "avail", "percent", "used", "free")]

  } else {
    stop("ps is not supported in this platform")
  }
}

ps__system_memory_linux <- function() {
  tab <- read.table("/proc/meminfo", header = FALSE, fill = TRUE)
  mems <- structure(
    as.list(tab[[2]] * 1024),
    names = tolower(sub(":$", "", tab[[1]]))
  )

  total <- mems[["memtotal"]]
  free <- mems[["memfree"]]

  buffers <- mems[["buffers"]] %||% NA_real_
  cached <- mems[["cached"]] %||% NA_real_ + mems[["sreclaimable"]] %||% 0
  shared <- mems[["shmem"]] %||% mems[["memshared"]] %||% NA_real_
  active <- mems[["active"]] %||% NA_real_

  inactive <- mems[["inactive"]]
  if (is.null(inactive)) {
    inactive <-
      mems[["inact_dirty"]] + mems[["inact_clean"]] + mems[["inact_laundry"]]
  }
  if (length(inactive) == 0) inactive <- NA_real_

  slab <- mems[["slab"]] %||% 0

  used <- total - free - cached - buffers
  if (used < 0 || is.na(used)) {
    # May be symptomatic of running within a LCX container where such
    # values will be dramatically distorted over those of the host.
    used <- total - free
  }

  avail <- mems[["memavailable"]] %||% NA_real_

  # If avail is greater than total or our calculation overflows,
  # that's symptomatic of running within a LCX container where such
  # values will be dramatically distorted over those of the host.
  # https://gitlab.com/procps-ng/procps/blob/
  #     24fd2605c51fccc375ab0287cec33aa767f06718/proc/sysinfo.c#L764
  if (!is.na(avail) && avail > total) avail <- free

  percent <- (total - avail) / total * 100

  list(total = total, avail = avail, percent = percent, used = used,
       free = free, active = active, inactive = inactive, buffers = buffers,
       cached = cached, shared = shared, slab = slab)
}

#' System swap memory statistics
#'
#' @return Named list. All numbers are in bytes:
#' * `total`: total swap memory.
#' * `used`: used swap memory.
#' * `free`: free swap memory.
#' * `percent`: the percentage usage.
#' * `sin`: the number of bytes the system has swapped in from disk
#'   (cumulative). This is `NA` on Windows.
#' * `sout`: the number of bytes the system has swapped out from disk
#'   (cumulative). This is `NA` on Windows.
#'
#' @family memory functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' ps_system_swap()

ps_system_swap <- function() {
  os <- ps_os_name()

  if (os == "MACOS") {
    l <- .Call(ps__system_swap)
    l$percent <- l$used / l$total * 100
    l[c("total", "used", "free", "percent", "sin", "sout")]
  } else if (os == "LINUX") {
    ps__system_swap_linux()

  } else if (os == "WINDOWS") {
    l <- .Call(ps__system_memory)
    total <- l[[3]]
    free <- l[[4]]
    used <- total - free
    percent <- used / total
    list(total = total, used = used, free = free,
         percent = percent, sin = NA_real_, sout = NA_real_)

  } else {
    stop("ps is not supported in this platform")
  }
}

ps__system_swap_linux <- function() {
  tab <- read.table("/proc/meminfo", header = FALSE, fill = TRUE)
  mems <- structure(
    as.list(tab[[2]] * 1024),
    names = tolower(sub(":$", "", tab[[1]]))
  )

  total <- mems[["swaptotal"]]
  free <- mems[["swapfree"]]
  used <- total - free
  percent <- used / total * 100
  sin <- sout <- 0

  tryCatch({
    tab2 <- read.table("/proc/vmstat", header = FALSE)
    vms <- structure(as.list(tab2[[2]] * 4 * 1024), names = tab2[[1]])
    sin <- vms[["pswpin"]] %||% 0
    sout <- vms[["pswpout"]] %||% 0
    }, error = function(e) NULL
  )

  list(total = total, used = used, free = free, percent = percent,
       sin = sin, sout = sout)
}

#' Create a process handle
#'
#' @param pid Process id. Integer scalar. `NULL` means the current R
#'   process.
#' @param time Start time of the process. Usually `NULL` and ps will query
#'   the start time.
#' @return `ps_handle()` returns a process handle (class `ps_handle`).
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p

ps_handle <- function(pid = NULL, time = NULL) {
  if (!is.null(pid)) pid <- assert_pid(pid)
  if (!is.null(time)) assert_time(time)
  .Call(psll_handle, pid, time)
}

#' @rdname ps_handle
#' @export

as.character.ps_handle <- function(x, ...) {
  pieces <- .Call(psll_format, x)
  paste0("<ps::ps_handle> PID=", pieces[[2]], ", NAME=", pieces[[1]],
         ", AT=", format_unix_time(pieces[[3]]))
}

#' @param x Process handle.
#' @param ... Not used currently.
#'
#' @rdname ps_handle
#' @export

format.ps_handle <- function(x, ...) {
  as.character(x, ...)
}

#' @rdname ps_handle
#' @export

print.ps_handle <- function(x, ...)  {
  cat(format(x, ...),  "\n", sep = "")
  invisible(x)
}

#' Pid of a process handle
#'
#' This function works even if the process has already finished.
#'
#' @param p Process handle.
#' @return Process id.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_pid(p)
#' ps_pid(p) == Sys.getpid()

ps_pid <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_pid, p)
}

#' Start time of a process
#'
#' The pid and the start time pair serves as the identifier of the process,
#' as process ids might be reused, but the chance of starting two processes
#' with identical ids within the resolution of the timer is minimal.
#'
#' This function works even if the process has already finished.
#'
#' @param p Process handle.
#' @return `POSIXct` object, start time, in GMT.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_create_time(p)

ps_create_time <- function(p = ps_handle()) {
  assert_ps_handle(p)
  format_unix_time(.Call(psll_create_time, p))
}

#' Checks whether a process is running
#'
#' It returns `FALSE` if the process has already finished.
#'
#' It uses the start time of the process to work around pid reuse. I.e.
#  it returns the correct answer, even if the process has finished and
#  its pid was reused.
#'
#' @param p Process handle.
#' @return Logical scalar.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_is_running(p)

ps_is_running <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_is_running, p)
}

#' Parent pid or parent process of a process
#'
#' `ps_ppid()` returns the parent pid, `ps_parent()` returns a `ps_handle`
#' of the parent.
#'
#' On POSIX systems, if the parent process terminates, another process
#' (typically the pid 1 process) is marked as parent. `ps_ppid()` and
#' `ps_parent()` will return this process then.
#'
#' Both `ps_ppid()` and `ps_parent()` work for zombie processes.
#'
#' @param p Process handle.
#' @return `ps_ppid()` returns and integer scalar, the pid of the parent
#'   of `p`. `ps_parent()` returns a `ps_handle`.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_ppid(p)
#' ps_parent(p)

ps_ppid <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_ppid, p)
}

#' @rdname  ps_ppid
#' @export

ps_parent <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_parent, p)
}

#' Process name
#'
#' The name of the program, which is typically the name of the executable.
#'
#' On Unix this can change, e.g. via an exec*() system call.
#'
#' `ps_name()` works on zombie processes.
#'
#' @param p Process handle.
#' @return Character scalar.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_name(p)
#' ps_exe(p)
#' ps_cmdline(p)

ps_name <- function(p = ps_handle()) {
  assert_ps_handle(p)
  n <- .Call(psll_name, p)
  if (nchar(n) >= 15) {
    ## On UNIX the name gets truncated to the first 15 characters.
    ## If it matches the first part of the cmdline we return that
    ## one instead because it's usually more explicative.
    ## Examples are "gnome-keyring-d" vs. "gnome-keyring-daemon".

    ## In addition, under qemu (e.g. in cross-platform Docker), the
    ## first entry is qemu and the second entry is the file name
    cmdline <- tryCatch(
      ps_cmdline(p),
      error = function(e) NULL
    )
    if (!is.null(cmdline) && length(cmdline) > 0L) {
      exname <- basename(cmdline[1])
      if (str_starts_with(exname, n)) {
        n <- exname
      } else if (grepl("qemu", exname) && length(cmdline) >= 2 &&
                 str_starts_with(exname2 <- basename(cmdline[2]), n)) {
        n <- exname2
      }
    }
  }
  n
}

#' Full path of the executable of a process
#'
#' Path to the executable of the process. May also be an empty string or
#' `NA` if it cannot be determined.
#'
#' For a zombie process it throws a `zombie_process` error.
#'
#' @param p Process handle.
#' @return Character scalar.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_name(p)
#' ps_exe(p)
#' ps_cmdline(p)

ps_exe <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_exe, p)
}

#' Command line of the process
#'
#' Command line of the process, i.e. the executable and the command line
#' arguments, in a character vector. On Unix the program might change its
#' command line, and some programs actually do it.
#'
#' For a zombie process it throws a `zombie_process` error.
#'
#' @param p Process handle.
#' @return Character vector.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_name(p)
#' ps_exe(p)
#' ps_cmdline(p)

ps_cmdline <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_cmdline, p)
}

#' Current process status
#'
#' One of the following:
#' * `"idle"`: Process being created by fork, or process has been sleeping
#'     for a long time. macOS only.
#' * `"running"`: Currently runnable on macOS and Windows. Actually
#'     running on Linux.
#' * `"sleeping"` Sleeping on a wait or poll.
#' * `"disk_sleep"` Uninterruptible sleep, waiting for an I/O operation
#'    (Linux only).
#' * `"stopped"` Stopped, either by a job control signal or because it
#'    is being traced.
#' * `"uninterruptible"` Process is in uninterruptible wait. macOS only.
#' * `"tracing_stop"` Stopped for tracing (Linux only).
#' * `"zombie"` Zombie. Finished, but parent has not read out the exit
#'    status yet.
#' * `"dead"` Should never be seen (Linux).
#' * `"wake_kill"` Received fatal signal (Linux only).
#' * `"waking"` Paging (Linux only, not valid since the 2.6.xx kernel).
#'
#' It might return `NA_character_` on macOS.
#'
#' Works for zombie processes.
#'
#' @section Note on macOS:
#' On macOS `ps_status()` often falls back to calling the external `ps`
#' program, because macOS does not let R access the status of most other
#' processes. Notably, it is usually able to access the status of other R
#' processes.
#'
#' The external `ps` program always runs as the root user, and
#' it also has special entitlements, so it can typically access the status
#' of most processes.
#'
#' If this behavior is problematic for you, e.g. because calling an
#' external program is too slow, set the `ps.no_external_ps` option to
#' `TRUE`:
#' ```
#' options(ps.no_external_ps = TRUE)
#' ```
#' Note that setting this option to `TRUE` will cause `ps_status()` to
#' return `NA_character_` for most processes.
#'
#' @param p Process handle.
#' @return Character scalar.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_status(p)

ps_status <- function(p = ps_handle()) {
  assert_ps_handle(p)
  ret <- .Call(psll_status, p)
  if (is.na(ret) && ps_os_type()[["MACOS"]] &&
      !isTRUE(getOption("ps.no_external_ps"))) {
    ret <- ps_status_macos_ps(ps_pid(p))
  }
  ret
}

#' Owner of the process
#'
#' The name of the user that owns the process. On Unix it is calculated
#' from the real user id.
#'
#' On Unix, a numeric uid id returned if the uid is not in the user
#' database, thus a username cannot be determined.
#'
#' Works for zombie processes.
#'
#' @param p Process handle.
#' @return String scalar.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_username(p)

ps_username <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_username, p)
}

#' Process current working directory as an absolute path.
#'
#' For a zombie process it throws a `zombie_process` error.
#'
#' @param p Process handle.
#' @return String scalar.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_cwd(p)

ps_cwd <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_cwd, p)
}

#' User ids and group ids of the process
#'
#' User ids and group ids of the process. Both return integer vectors with
#' names: `real`, `effective` and `saved`.
#'
#' Both work for zombie processes.
#'
#' They are not implemented on Windows, they throw a `not_implemented`
#' error.
#'
#' @param p Process handle.
#' @return Named integer vector of length 3, with names: `real`,
#'   `effective` and `saved`.
#'
#' @seealso [ps_username()] returns a user _name_ and works on all
#'   platforms.
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ps::ps_os_type()["POSIX"] && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_uids(p)
#' ps_gids(p)

ps_uids <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_uids, p)
}

#' @rdname ps_uids
#' @export

ps_gids <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_gids, p)
}

#' Terminal device of the process
#'
#' Returns the terminal of the process. Not implemented on Windows, always
#' returns `NA_character_`. On Unix it returns `NA_character_` if the
#' process has no terminal.
#'
#' Works for zombie processes.
#'
#' @param p Process handle.
#' @return Character scalar.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_terminal(p)

ps_terminal <- function(p = ps_handle()) {
  assert_ps_handle(p)
  ttynr <- .Call(psll_terminal, p)
  if (is.character(ttynr)) {
    ttynr
  } else if (is.na(ttynr)) {
    NA_character_
  } else {
    tmap <- get_terminal_map()
    tmap[[as.character(ttynr)]]
  }
}

#' Environment variables of a process
#'
#' `ps_environ()` returns the environment variables of the process, in a
#' named vector, similarly to the return value of `Sys.getenv()`
#' (without arguments).
#'
#' Note: this usually does not reflect changes made after the process
#' started.
#'
#' `ps_environ_raw()` is similar to `p$environ()` but returns the
#' unparsed `"var=value"` strings. This is faster, and sometimes good
#' enough.
#'
#' These functions throw a `zombie_process` error for zombie processes.
#'
#' @section macOS issues:
#'
#' `ps_environ()` usually does not work on macOS nowadays. This is because
#' macOS does not allow reading the environment variables of another
#' process. Accoding to the Darwin source code, `ps_environ` will work is
#' one of these conditions hold:
#'
#' * You are running a development or debug kernel, i.e. if you are
#'   debugging the macOS kernel itself.
#' * The target process is same as the calling process.
#' * SIP if off.
#' * The target process is not restricted, e.g. it is running a binary
#'   that was not signed.
#' * The calling process has the
#'   `com.apple.private.read-environment-variables` entitlement. However
#'   adding this entitlement to the R binary makes R crash on startup.
#'
#' Otherwise `ps_environ` will return an empty set of environment variables
#' on macOS.
#'
#' Issue 121 might have more information about this.
#'
#' @param p Process handle.
#' @return `ps_environ()` returns a named character vector (that has a
#' `Dlist` class, so it is printed nicely), `ps_environ_raw()` returns a
#' character vector.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' env <- ps_environ(p)
#' env[["R_HOME"]]

ps_environ <- function(p = ps_handle()) {
  assert_ps_handle(p)
  parse_envs(.Call(psll_environ, p))
}

#' @rdname ps_environ
#' @export

ps_environ_raw <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_environ, p)
}

#' Number of threads
#'
#' Throws a `zombie_process()` error for zombie processes.
#'
#' @param p Process handle.
#' @return Integer scalar.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_num_threads(p)

ps_num_threads <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_num_threads, p)
}

#' CPU times of the process
#'
#' All times are measured in seconds:
#' * `user`: Amount of time that this process has been scheduled in user
#'   mode.
#' * `system`: Amount of time that this process has been scheduled in
#'   kernel mode
#' * `children_user`: On Linux, amount of time that this process's
#'   waited-for children have been scheduled in user mode.
#' * `children_system`: On Linux, Amount of time that this process's
#'   waited-for children have been scheduled in kernel mode.
#'
#' Throws a `zombie_process()` error for zombie processes.
#'
#' @param p Process handle.
#' @return Named real vector or length four: `user`, `system`,
#'   `children_user`,  `children_system`. The last two are `NA` on
#'   non-Linux systems.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_cpu_times(p)
#' proc.time()

ps_cpu_times <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_cpu_times, p)
}

#' Memory usage information
#'
#' @details
#'
#' `ps_memory_info()` returns information about memory usage.
#'
#' It returns a named vector. Portable fields:
#' * `rss`: "Resident Set Size", this is the non-swapped physical memory a
#'   process has used (bytes). On UNIX it matches "top"‘s 'RES' column (see doc). On
#'   Windows this is an alias for `wset` field and it matches "Memory"
#'   column of `taskmgr.exe`.
#' * `vmem`: "Virtual Memory Size", this is the total amount of virtual
#'   memory used by the process (bytes). On UNIX it matches "top"‘s 'VIRT' column
#'   (see doc). On Windows this is an alias for the `pagefile` field and
#'   it matches the "Working set (memory)" column of `taskmgr.exe`.
#'
#' Non-portable fields:
#' * `shared`: (Linux) memory that could be potentially shared with other
#'   processes (bytes). This matches "top"‘s 'SHR' column (see doc).
#' * `text`: (Linux): aka 'TRS' (text resident set) the amount of memory
#'   devoted to executable code (bytes). This matches "top"‘s 'CODE' column (see
#'   doc).
#' * `data`: (Linux): aka 'DRS' (data resident set) the amount of physical
#'   memory devoted to other than executable code (bytes). It matches "top"‘s
#'   'DATA' column (see doc).
#' * `lib`: (Linux): the memory used by shared libraries (bytes).
#' * `dirty`: (Linux): the amount of memory in dirty pages (bytes).
#' * `pfaults`: (macOS): number of page faults.
#' * `pageins`: (macOS): number of actual pageins.
#'
#' For the explanation of Windows fields see the
#' [PROCESS_MEMORY_COUNTERS_EX](https://learn.microsoft.com/en-us/windows/win32/api/psapi/ns-psapi-process_memory_counters_ex)
#' structure.
#'
#' `ps_memory_full_info()` returns all fields as `ps_memory_info()`, plus
#' additional information, but typically takes slightly longer to run, and
#' might not have access to some processes that `ps_memory_info()` can
#' query:
#'
#' * `maxrss` maximum resident set size over the process's lifetime. This
#'   only works for the calling process, otherwise it is `NA_real_`.
#' * `uss`: Unique Set Size, this is the memory which is unique to a
#'   process and which would be freed if the process was terminated right
#'   now.
#' * `pss` (Linux only): Proportional Set Size, is the amount of memory
#'   shared with other processes, accounted in a way that the amount is
#'   divided evenly between the processes that share it. I.e. if a process
#'   has 10 MBs all to itself and 10 MBs shared with another process its
#'   PSS will be 15 MBs.
#' * `swap` (Linux only): amount of memory that has been swapped out to
#'   disk.
#'
#' They both throw a `zombie_process()` error for zombie processes.
#'
#' @param p Process handle.
#' @return Named real vector.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' p
#' ps_memory_info(p)
#' ps_memory_full_info(p)

ps_memory_info <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_memory_info, p)
}

#' @export
#' @rdname ps_memory_info

ps_memory_full_info <- function(p = ps_handle()) {
  assert_ps_handle(p)
  info <- ps_memory_info(p)
  info[["maxrss"]] <- if (Sys.getpid() == ps_pid(p)) {
    .Call(psll_memory_maxrss, p)
  } else {
    NA_real_
  }

  type <- ps_os_type()
  if (type[["LINUX"]]) {
    match <- function(re) {
      mt <- gregexpr(re, smaps, perl = TRUE)[[1]]
      st <- substring(
        smaps,
        attr(mt, "capture.start"),
        attr(mt, "capture.start") + attr(mt, "capture.length") - 1
      )
      sum(as.integer(st), na.rm = TRUE) * 1024
    }

    smaps <- .Call(ps__memory_maps, p)
    info[["uss"]] <- match("\nPrivate.*:\\s+(\\d+)")
    info[["pss"]] <- match("\nPss:\\s+(\\d+)")
    info[["swap"]] <- match("\nSwap:\\s+(\\d+)")

  } else if (type[["MACOS"]]) {
    info[["uss"]] <- .Call(psll_memory_uss, p)

    } else if (type[["WINDOWS"]]) {
    info[["uss"]] <- .Call(psll_memory_uss, p)
  }
  info
}

process_signal_result <- function(p, res, err_msg) {
  ok <- map_lgl(res, function(x) is.character(x) || is.null(x))
  if (all(ok)) {
    unlist(res)
  } else {
    for (i in which(!ok)) {
      class(res[[i]]) <- res[[i]][[2]]
    }
    pids <- map_int(res[!ok], function(x) x[["pid"]] %||% NA_integer_)
    nms <- map_chr(p[!ok], function(pp) {
      tryCatch(ps_name(pp), error = function(e) "???")
    })
    pmsg <- paste0(pids, " (", nms, ")", collapse = ", ")
    # put these classes at the end
    common <- c("ps_error", "error", "condition")
    cls <- c(
      unique(setdiff(unlist(lapply(res[!ok], function(x) class(x))), common)),
      common
    )
    err <- structure(
      list(
        message = paste0(
          err_msg,
          if (length(p) == 1) ": " else " some processes: ",
          pmsg
        ),
        results = res,
        pid = pids
      ),
      class = cls
    )
    stop(err)
  }
}

#' Send signal to a process
#'
#' Send a signal to the process. Not implemented on Windows. See
#' [signals()] for the list of signals on the current platform.
#'
#' It checks if the process is still running, before sending the signal,
#' to avoid signalling the wrong process, because of pid reuse.
#'
#' @param p Process handle, or a list of process handles.
#' @param sig Signal number, see [signals()].
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ps::ps_os_type()["POSIX"] && ! ps:::is_cran_check()
#' px <- processx::process$new("sleep", "10")
#' p <- ps_handle(px$get_pid())
#' p
#' ps_send_signal(p, signals()$SIGINT)
#' p
#' ps_is_running(p)
#' px$get_exit_status()

ps_send_signal <- function(p = ps_handle(), sig) {
  p <- assert_ps_handle_or_handle_list(p)
  assert_signal(sig)
  res <- lapply(p, function(pp) {
    tryCatch(
      .Call(psll_send_signal, pp, sig),
      error = function(e) e
    )
  })
  process_signal_result(p, res, "Failed to send signal to")
}

#' Suspend (stop) the process
#'
#' Suspend process execution with `SIGSTOP` preemptively checking
#' whether PID has been reused. On Windows this has the effect of
#' suspending all process threads.
#'
#' @param p Process handle or a list of process handles.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ps::ps_os_type()["POSIX"] && ! ps:::is_cran_check()
#' px <- processx::process$new("sleep", "10")
#' p <- ps_handle(px$get_pid())
#' p
#' ps_suspend(p)
#' ps_status(p)
#' ps_resume(p)
#' ps_status(p)
#' ps_kill(p)

ps_suspend <- function(p = ps_handle()) {
  p <- assert_ps_handle_or_handle_list(p)
  res <- lapply(p, function(pp) {
    tryCatch(
      .Call(psll_suspend, pp),
      error = function(e) e
    )
  })
  process_signal_result(p, res, "Failed to suspend")
}

#' Resume (continue) a stopped process
#'
#' Resume process execution with SIGCONT preemptively checking
#' whether PID has been reused. On Windows this has the effect of resuming
#' all process threads.
#'
#' @param p Process handle or a list of process handles.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ps::ps_os_type()["POSIX"] && ! ps:::is_cran_check()
#' px <- processx::process$new("sleep", "10")
#' p <- ps_handle(px$get_pid())
#' p
#' ps_suspend(p)
#' ps_status(p)
#' ps_resume(p)
#' ps_status(p)
#' ps_kill(p)

ps_resume <- function(p = ps_handle()) {
  p <- assert_ps_handle_or_handle_list(p)
  res <- lapply(p, function(pp) {
    tryCatch(
      .Call(psll_resume, pp),
      error = function(e) e
    )
  })
  process_signal_result(p, res, "Failed to resume")
}

#' Terminate a Unix process
#'
#' Send a `SIGTERM` signal to the process. Not implemented on Windows.
#'
#' Checks if the process is still running, to work around pid reuse.
#'
#' @param p Process handle or a list of process handles.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ps::ps_os_type()["POSIX"] && ! ps:::is_cran_check()
#' px <- processx::process$new("sleep", "10")
#' p <- ps_handle(px$get_pid())
#' p
#' ps_terminate(p)
#' p
#' ps_is_running(p)
#' px$get_exit_status()

ps_terminate <- function(p = ps_handle()) {
  p <- assert_ps_handle_or_handle_list(p)
  res <- lapply(p, function(pp) {
    tryCatch(
      .Call(psll_terminate, pp),
      error = function(e) e
    )
  })
  process_signal_result(p, res, "Failed to terminate")
}

#' Kill one or more processes
#'
#' Kill the process with SIGKILL preemptively checking whether PID has
#' been reused. On Windows it uses `TerminateProcess()`.
#'
#' Note that since ps version 1.8, `ps_kill()` does not error if the
#' `p` process (or some processes if `p` is a list) are already terminated.
#'
#' @param p Process handle, or a list of process handles.
#' @param grace Grace period, in milliseconds, used on Unix. If it is not
#'   zero, then `ps_kill()` first sends a `SIGTERM` signal to all processes
#'   in `p`. If some proccesses do not terminate within `grace`
#'   milliseconds after the `SIGTERM` signal, `ps_kill()` kills them by
#'   sending `SIGKILL` signals.
#' @return Character vector, with one element for each process handle in
#'   `p`. If the process was already dead before `ps_kill()` tried to kill
#'   it, the corresponding return value is `"dead"`. If `ps_kill()` just
#'   killed it, it is `"killed"`.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ps::ps_os_type()["POSIX"] && ! ps:::is_cran_check()
#' px <- processx::process$new("sleep", "10")
#' p <- ps_handle(px$get_pid())
#' p
#' ps_kill(p)
#' p
#' ps_is_running(p)
#' px$get_exit_status()

ps_kill <- function(p = ps_handle(), grace = 200) {
  p <- assert_ps_handle_or_handle_list(p)
  grace <- assert_grace(grace)
  if (ps_os_type()[["WINDOWS"]]) {
    res <- lapply(p, function(pp) {
      tryCatch({
        if (ps_is_running(pp)) {
          .Call(psll_kill, pp, 0L)
          "killed"
        } else {
          "dead"
        }
      }, error = function(e) {
        if (inherits(e, "no_such_process")) "dead" else e
      })
    })
  } else {
    res <- call_with_cleanup(psll_kill, p, grace)
  }

  process_signal_result(p, res, "Failed to kill")
}

#' List of child processes (process objects) of the process. Note that
#' this typically requires enumerating all processes on the system, so
#' it is a costly operation.
#'
#' @param p Process handle.
#' @param recursive Whether to include the children of the children, etc.
#' @return List of `ps_handle` objects.
#'
#' @family process handle functions
#' @export
#' @importFrom utils head tail
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_parent(ps_handle())
#' ps_children(p)

ps_children <- function(p = ps_handle(), recursive = FALSE) {
  assert_ps_handle(p)
  assert_flag(recursive)

  mypid <- ps_pid(p)
  mytime <- ps_create_time(p)
  map <- ps_ppid_map()
  ret <- list()

  if (!recursive) {
    for (i in seq_len(nrow(map))) {
      if (map$ppid[i] == mypid) {
        tryCatch({
          child  <- ps_handle(map$pid[i])
          if (mytime <= ps_create_time(child)) {
            ret <- c(ret, child)
          } },
          no_such_process = function(e) NULL,
          zombie_process = function(e) NULL)
      }
    }

  } else {
    seen <- integer()
    stack <- mypid
    while (length(stack)) {
      pid <- tail(stack, 1)
      stack <- head(stack, -1)
      if (pid %in% seen) next           # nocov (happens _very_ rarely)
      seen <- c(seen, pid)
      child_pids <- map[ map[,2] ==  pid, 1]
      for (child_pid in child_pids) {
        tryCatch({
          child <- ps_handle(child_pid)
          if (mytime <= ps_create_time(child)) {
            ret <- c(ret, child)
            stack <- c(stack, child_pid)
          } },
          no_such_process = function(e) NULL,
          zombie_process = function(e) NULL)
      }
    }
  }

  ## This will throw if p has finished
  ps_ppid(p)

  ret
}

#' Query the ancestry of a process
#'
#' Query the parent processes recursively, up to the first process.
#' (On some platforms, like Windows, the process tree is not a tree
#' and may contain loops, in which case `ps_descent()` only goes up
#' until the first repetition.)
#'
#' @param p Process handle.
#' @return A list of process handles, starting with `p`, each one
#' is the parent process of the previous one.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' ps_descent()

ps_descent <- function(p = ps_handle()) {
  assert_ps_handle(p)
  windows <- ps_os_type()[["WINDOWS"]]

  branch <- list()
  branch_pids <- integer()
  current <- p
  current_pid <- ps_pid(p)
  if (windows) current_time <- ps_create_time(p)

  while (TRUE) {
    branch <- c(branch, list(current))
    branch_pids <- c(branch_pids, current_pid)
    parent <- fallback(ps_parent(current), NULL)

    # Might fail on Windows, if the process does not exist
    if (is.null(parent)) break;

    # If the parent pid is the same, we stop.
    # Also, Windows might have loops
    parent_pid <- ps_pid(parent)
    if (parent_pid %in% branch_pids) break;

    # Need to check for pid reuse on Windows
    if (windows) {
      parent_time <- ps_create_time(parent)
      if (current_time <= parent_time) break
      current_time <- parent_time
    }

    current <- parent
    current_pid <- parent_pid
  }

  branch
}

ps_ppid_map <- function() {
  pids <- ps_pids()

  processes <- not_null(lapply(pids, function(p) {
    tryCatch(ps_handle(p), error = function(e) NULL) }))

  pids <- map_int(processes, ps_pid)
  ppids <- map_int(processes, function(p) fallback(ps_ppid(p), NA_integer_))

  ok <- !is.na(ppids)

  data_frame(
    pid = pids[ok],
    ppid = ppids[ok]
  )
}

#' Number of open file descriptors
#'
#' Note that in some IDEs, e.g. RStudio or R.app on macOS, the IDE itself
#' opens files from other threads, in addition to the files opened from the
#' main R thread.
#'
#' For a zombie process it throws a `zombie_process` error.
#'
#' @param p Process handle.
#' @return Integer scalar.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' ps_num_fds(p)
#' f <- file(tmp <- tempfile(), "w")
#' ps_num_fds(p)
#' close(f)
#' unlink(tmp)
#' ps_num_fds(p)

ps_num_fds <- function(p = ps_handle()) {
  assert_ps_handle(p)
  .Call(psll_num_fds, p)
}

#' Open files of a process
#'
#' Note that in some IDEs, e.g. RStudio or R.app on macOS, the IDE itself
#' opens files from other threads, in addition to the files opened from the
#' main R thread.
#'
#' For a zombie process it throws a `zombie_process` error.
#'
#' @param p Process handle.
#' @return Data frame with columns: `fd` and `path`. `fd` is numeric
#'    file descriptor on POSIX systems, `NA` on Windows. `path` is an
#'    absolute path to the file.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' ps_open_files(p)
#' f <- file(tmp <- tempfile(), "w")
#' ps_open_files(p)
#' close(f)
#' unlink(tmp)
#' ps_open_files(p)

ps_open_files <- function(p = ps_handle()) {
  assert_ps_handle(p)

  l <- not_null(.Call(psll_open_files, p))

  d <- data_frame(
    fd = vapply(l, "[[", integer(1), 2),
    path = vapply(l, "[[", character(1), 1))

  d
}

#' List network connections of a process
#'
#' For a zombie process it throws a `zombie_process` error.
#'
#' @param p Process handle.
#' @return Data frame, with columns:
#'    * `fd`: integer file descriptor on POSIX systems, `NA` on Windows.
#'    * `family`: Address family, string, typically `AF_UNIX`, `AF_INET` or
#'       `AF_INET6`.
#'    * `type`: Socket type, string, typically `SOCK_STREAM` (TCP) or
#'       `SOCK_DGRAM` (UDP).
#'    * `laddr`: Local address, string, `NA` for UNIX sockets.
#'    * `lport`: Local port, integer, `NA` for UNIX sockets.
#'    * `raddr`: Remote address, string, `NA` for UNIX sockets. This is
#'      always `NA` for `AF_INET` sockets on Linux.
#'    * `rport`: Remote port, integer, `NA` for UNIX sockets.
#'    * `state`: Socket state, e.g. `CONN_ESTABLISHED`, etc. It is `NA`
#'      for UNIX sockets.
#'
#' @family process handle functions
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check()
#' p <- ps_handle()
#' ps_connections(p)
#' sc <- socketConnection("httpbin.org", port = 80)
#' ps_connections(p)
#' close(sc)
#' ps_connections(p)

ps_connections <- function(p = ps_handle()) {
  assert_ps_handle(p)
  if (ps_os_type()[["LINUX"]]) return(psl_connections(p))

  l <- not_null(.Call(psll_connections, p))

  d <- data_frame(
    fd = vapply(l, "[[", integer(1), 1),
    family = match_names(ps_env$constants$address_families,
                       vapply(l, "[[", integer(1), 2)),
    type = match_names(ps_env$constants$socket_types,
                       vapply(l, "[[", integer(1), 3)),
    laddr = vapply(l, "[[", character(1), 4),
    lport = vapply(l, "[[", integer(1), 5),
    raddr = vapply(l, "[[", character(1), 6),
    rport = vapply(l, "[[", integer(1), 7),
    state = match_names(ps_env$constants$tcp_statuses,
                        vapply(l, "[[", integer(1), 8)))

  d$laddr[d$laddr == ""] <- NA_character_
  d$raddr[d$raddr == ""] <- NA_character_

  d$lport[d$lport == 0] <- NA_integer_
  d$rport[d$rport == 0] <- NA_integer_

  d
}

#' Interrupt a process
#'
#' Sends `SIGINT` on POSIX, and 'CTRL+C' or 'CTRL+BREAK' on Windows.
#'
#' @param p Process handle or a list of process handles.
#' @param ctrl_c On Windows, whether to send 'CTRL+C'. If `FALSE`, then
#'   'CTRL+BREAK' is sent. Ignored on non-Windows platforms.
#'
#' @family process handle functions
#' @export

ps_interrupt <- function(p = ps_handle(), ctrl_c = TRUE) {
  p <- assert_ps_handle_or_handle_list(p)
  assert_flag(ctrl_c)
  res <- lapply(p, function(pp) {
    tryCatch({
      if (ps_os_type()[["WINDOWS"]]) {
        interrupt <- get_tool("interrupt")
        .Call(psll_interrupt, pp, ctrl_c, interrupt)
      } else {
        .Call(psll_interrupt, pp, ctrl_c, NULL)
      }
    }, error = function(e) e)
  })
  process_signal_result(p, res, "Failed to interrupt")
}

#' @return `ps_windows_nice_values()` return a character vector of possible
#' priority values on Windows.
#' @export
#' @rdname ps_get_nice

ps_windows_nice_values <- function() {
 c("realtime",
   "high",
   "above_normal",
   "normal",
   "idle",
   "below_normal")
}

#' Get or set the priority of a process
#'
#' `ps_get_nice()` returns the current priority, `ps_set_nice()` sets a
#' new priority, `ps_windows_nice_values()` list the possible priority
#' values on Windows.
#'
#' Priority values are different on Windows and Unix.
#'
#' On Unix, priority is an integer, which is maximum 20. 20 is the lowest
#' priority.
#'
#' ## Rules:
#' * On Windows you can only set the priority of the processes the current
#'   user has `PROCESS_SET_INFORMATION` access rights to. This typically
#'   means your own processes.
#' * On Unix you can only set the priority of the your own processes.
#'   The superuser can set the priority of any process.
#' * On Unix you cannot set a higher priority, unless you are the superuser.
#'   (I.e. you cannot set a lower number.)
#' * On Unix the default priority of a process is zero.
#'
#' @param p Process handle.
#' @return `ps_get_nice()` returns a string from
#' `ps_windows_nice_values()` on Windows. On Unix it returns an integer
#' smaller than or equal to 20.
#'
#' @export

ps_get_nice <- function(p = ps_handle()) {
  assert_ps_handle(p)
  code <- .Call(psll_get_nice, p)
  if (ps_os_type()[["WINDOWS"]]) {
    ps_windows_nice_values()[code]
  } else {
    code
  }
}

#' @param value On Windows it must be a string, one of the values of
#' `ps_windows_nice_values()`. On Unix it is a priority value that is
#' smaller than or equal to 20.
#' @return `ps_set_nice()` return `NULL` invisibly.
#'
#' @export
#' @rdname ps_get_nice

ps_set_nice <- function(p = ps_handle(), value) {
  assert_ps_handle(p)
  assert_nice_value(value)
  if (ps_os_type()[["POSIX"]]) {
    value <- as.integer(value)
  } else {
    value <- match(value, ps_windows_nice_values())
  }
  invisible(.Call(psll_set_nice, p, value))
}

#' List the dynamically loaded libraries of a process
#'
#' Note: this function currently only works on Windows.
#' @param p Process handle.
#' @return Data frame with one column currently: `path`, the
#' absolute path to the loaded module or shared library. On Windows
#' the list includes the executable file itself.
#'
#' @export
#' @family process handle functions
#' @family shared library tools
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check() && ps::ps_os_type()[["WINDOWS"]]
#' # The loaded DLLs of the current process
#' ps_shared_libs()

ps_shared_libs <- function(p = ps_handle()) {
  assert_ps_handle(p)
  if (!ps_os_type()[["WINDOWS"]]) {
    stop("`ps_shared_libs()` is currently only supported on Windows")
  }

  l <- .Call(psll_dlls, p)

  d <- data_frame(
    path = map_chr(l, "[[", 1)
  )

  d
}

#' Query or set CPU affinity
#'
#' `ps_get_cpu_affinity()` queries the
#' [CPU affinity](https://www.linuxjournal.com/article/6799?page=0,0) of
#' a process. `ps_set_cpu_affinity()` sets the CPU affinity of a process.
#'
#' CPU affinity consists in telling the OS to run a process on a limited
#' set of CPUs only (on Linux cmdline, the `taskset` command is typically
#' used).
#'
#' These functions are only supported on Linux and Windows. They error on macOS.
#'
#' @param p Process handle.
#' @param affinity Integer vector of CPU numbers to restrict a process to.
#' CPU numbers start with zero, and they have to be smaller than the
#' number of (logical) CPUs, see [ps_cpu_count()].
#'
#' @return `ps_get_cpu_affinity()` returns an integer vector of CPU
#' numbers, starting with zero.
#'
#' `ps_set_cpu_affinity()` returns `NULL`, invisibly.
#'
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check() && ! ps::ps_os_type()[["MACOS"]]
#' # current
#' orig <- ps_get_cpu_affinity()
#' orig
#'
#' # restrict
#' ps_set_cpu_affinity(affinity = 0:0)
#' ps_get_cpu_affinity()
#'
#' # restore
#' ps_set_cpu_affinity(affinity = orig)
#' ps_get_cpu_affinity()

ps_get_cpu_affinity <- function(p = ps_handle()) {
  assert_ps_handle(p)
  type <- ps_os_type()
  if (!type[["LINUX"]] && !type[["WINDOWS"]]) {
    stop("`ps_cpu_affinity()` is only supported on Windows and Linux")
  }

  .Call(psll_get_cpu_aff, p)
}

#' @export
#' @rdname ps_get_cpu_affinity

ps_set_cpu_affinity <- function(p = ps_handle(), affinity) {
  assert_ps_handle(p)
  type <- ps_os_type()
  if (!type[["LINUX"]] && !type[["WINDOWS"]]) {
    stop("`ps_cpu_affinity()` is only supported on Windows and Linux")
  }

  # check affinity values
  cnt <- ps_cpu_count()
  stopifnot(is.integer(affinity), all(affinity < cnt))

  invisible(.Call(psll_set_cpu_aff, p, affinity))
}

#' Wait for one or more processes to terminate, with a timeout
#'
#' This function supports interruption with SIGINT on Unix, or CTRL+C
#' or CTRL+BREAK on Windows.
#'
#' @param p A process handle, or a list of process handles. The
#'   process(es) to wait for.
#' @param timeout Timeout in milliseconds. If -1, `ps_wait()` will wait
#'   indefinitely (or until it is interrupted). If 0, then it checks which
#'   processes have already terminated, and returns immediately.
#' @return Logical vector, with one value of each process in `p`.
#'   For processes that terminated it contains a `TRUE` value. For
#'   processes that are still running it contains a `FALSE` value.
#'
#' @export
#' @examplesIf ps::ps_is_supported() && ! ps:::is_cran_check() && ps::ps_os_type()["POSIX"]
#' # this example calls `sleep`, so it only works on Unix
#' p1 <- processx::process$new("sleep", "100")
#' p2 <- processx::process$new("sleep", "100")
#'
#' # returns c(FALSE, FALSE) immediately if p1 and p2 are running
#' ps_wait(list(p1$as_ps_handle(), p2$as_ps_handle()), 0)
#'
#' # timeouts at one second
#' ps_wait(list(p1$as_ps_handle(), p2$as_ps_handle()), 1000)
#'
#' p1$kill()
#' p2$kill()
#' # returns c(TRUE, TRUE) immediately
#' ps_wait(list(p1$as_ps_handle(), p2$as_ps_handle()), 1000)

ps_wait <- function(p, timeout = -1) {
  p <- assert_ps_handle_or_handle_list(p)
  timeout <- assert_integer(timeout)
  call_with_cleanup(psll_wait, p, timeout)
}

if (!ps_os_type()[["WINDOWS"]]) return()

test_that("uids, gids", {
  p1 <- processx::process$new(px(), c("sleep", "10"))
  on.exit(p1$kill(), add = TRUE)
  ps <- ps_handle(p1$get_pid())
  expect_true(ps_is_running(ps))

  err <- tryCatch(ps_uids(ps), error = function(e) e)
  expect_s3_class(err, "not_implemented")
  expect_s3_class(err, "ps_error")
  err <- tryCatch(ps_gids(ps), error = function(e) e)
  expect_s3_class(err, "not_implemented")
  expect_s3_class(err, "ps_error")
})

test_that("terminal", {
  p1 <- processx::process$new(px(), c("sleep", "10"))
  on.exit(p1$kill(), add = TRUE)
  ps <- ps_handle(p1$get_pid())
  expect_true(ps_is_running(ps))

  expect_identical(ps_terminal(ps), NA_character_)
})

## TODO: username
## TODO: cpu_times
## TODO: memory_info

test_that("total and available mem", {
  l <- .Call(ps__system_memory)[c("total", "avail")]
  expect_true(is.numeric(l$total))
  expect_true(is.numeric(l$avail))
  expect_lte(l$avail, l$total)
})

if (!ps_os_type()[["POSIX"]]) return()

test_that("is_running", {
  ## Zombie is running
  zpid <- zombie()
  on.exit(waitpid(zpid), add = TRUE)
  ps <- ps_handle(zpid)
  expect_true(ps_is_running(ps))
})

test_that("terminal", {
  tty <- ps_terminal(ps_handle())
  if (is.na(tty)) skip("no terminal")
  expect_true(file.exists(tty))

  ## It is a character special file
  out <- processx::run("ls", c("-l", tty))$stdout
  expect_equal(substr(out, 1, 1), "c")
})

test_that("username, uids, gids", {
  if (Sys.which("ps") == "") skip("No ps program")
  ret <- system("ps -p 1 >/dev/null 2>/dev/null")
  if (ret != 0) skip("ps does not work properly")
  p1 <- processx::process$new("sleep", "10")
  on.exit(p1$kill(), add = TRUE)
  ps <- ps_handle(p1$get_pid())
  expect_true(ps_is_running(ps))

  ps2_username <- parse_ps(c("-o", "user", "-p", ps_pid(ps)))
  expect_equal(ps_username(ps), ps2_username)

  ps2_uid <- parse_ps(c("-o", "uid", "-p", ps_pid(ps)))
  expect_equal(ps_uids(ps)[["real"]], as.numeric(ps2_uid))

  ps2_gid <- parse_ps(c("-o", "rgid", "-p", ps_pid(ps)))
  expect_equal(ps_gids(ps)[["real"]], as.numeric(ps2_gid))
})


test_that("send_signal", {
  p1 <- processx::process$new("sleep", "10")
  on.exit(p1$kill(), add = TRUE)
  ps <- ps_handle(p1$get_pid())

  ps_send_signal(ps, signals()$SIGINT)
  timeout <- Sys.time() + 60
  while (Sys.time() < timeout && p1$is_alive()) Sys.sleep(0.05)
  expect_false(p1$is_alive())
  expect_false(ps_is_running(ps))
  expect_equal(p1$get_exit_status(), - signals()$SIGINT)
})

test_that("terminate", {
  p1 <- processx::process$new("sleep", "10")
  on.exit(p1$kill(), add = TRUE)
  ps <- ps_handle(p1$get_pid())

  ps_terminate(ps)
  timeout <- Sys.time() + 60
  while (Sys.time() < timeout && p1$is_alive()) Sys.sleep(0.05)
  expect_false(p1$is_alive())
  expect_false(ps_is_running(ps))
  expect_equal(p1$get_exit_status(), - signals()$SIGTERM)
})

test_that("kill with grace", {
  p1 <- processx::process$new(
    px(),
    c("sigterm", "ignore", "outln", "setup", "sleep", "3"),
    stdout = "|"
  )
  on.exit(p1$kill(), add = TRUE)
  ph1 <- p1$as_ps_handle()

  # need to wait until the SIGTERM handler is set up in px
  expect_equal(p1$poll_io(1000)[["output"]], "ready")
  expect_equal(ps_kill(ph1), "killed")
})

test_that("kill with grace, multiple processes", {
  # ignored SIGTERM completely
  p1 <- processx::process$new(
    px(),
    c("sigterm", "ignore", "outln", "setup", "sleep", "3"),
    stdout = "|"
  )
  on.exit(p1$kill(), add = TRUE)
  ph1 <- p1$as_ps_handle()

  # exits 0.5s later after SIGTERM
  p2 <- processx::process$new(
    px(),
    c("sigterm", "sleep", "0.5", "outln", "setup", "sleep", "3"),
    stdout = "|"
  )
  on.exit(p2$kill(), add = TRUE)
  ph2 <- p2$as_ps_handle()

  # exits on SIGTERM
  p3 <- processx::process$new(px(), c("sleep", "3"))
  on.exit(p3$kill(), add = TRUE)
  ph3 <- p3$as_ps_handle()

  # wait until signal handlers are set up
  expect_equal(p1$poll_io(1000)[["output"]], "ready")
  expect_equal(p2$poll_io(1000)[["output"]], "ready")
  expect_equal(
    ps_kill(list(ph1, ph2, ph3), grace = 1000),
    c("killed", "terminated", "terminated")
  )
})

if (!ps_os_type()[["LINUX"]]) {
  return()
}

test_that("status", {
  ## Argument check
  expect_error(ps_status(123), class = "invalid_argument")

  p1 <- processx::process$new("sleep", "10")
  on.exit(p1$kill(), add = TRUE)
  ps <- ps_handle(p1$get_pid())
  expect_true(ps_is_running(ps))

  wait_for_status(ps, "sleeping")
  expect_equal(ps_status(ps), "sleeping")
  ps_suspend(ps)
  wait_for_status(ps, "stopped")
  expect_equal(ps_status(ps), "stopped")
  ps_resume(ps)
  wait_for_status(ps, "sleeping")
  expect_equal(ps_status(ps), "sleeping")
  ## TODO: rest?
})

## TODO: cpu_times ??? We apparently cannot get them from ps

## Helper: read starttime (ticks since boot) from /proc/<pid>/stat,
## handling process names that contain spaces or parentheses.
proc_starttime_ticks <- function(pid) {
  raw <- readLines(sprintf("/proc/%d/stat", pid))
  ## Everything after the last ')' is the fixed-format tail
  after_paren <- sub("^.*\\) ", "", raw)
  fields <- strsplit(after_paren, " ")[[1]]
  ## starttime is the 20th field after the name (field 22 overall)
  as.numeric(fields[20])
}

## Helper: integer boot time from /proc/stat btime (what old ps used)
proc_stat_btime <- function() {
  line <- grep("^btime ", readLines("/proc/stat"), value = TRUE)
  as.numeric(strsplit(line, " +")[[1]][2])
}

test_that("ps_handle validates legacy (integer /proc/stat btime) create_time", {
  skip_if_no_processx()

  p <- processx::process$new("sleep", "100")
  on.exit(p$kill(), add = TRUE)
  pid <- p$get_pid()

  ## Build the create_time the way old ps did: integer boot time + ticks/CLK_TCK
  btime   <- proc_stat_btime()
  ticks   <- proc_starttime_ticks(pid)
  clk_tck <- as.integer(system2("getconf", "CLK_TCK", stdout = TRUE))
  legacy_ct <- structure(
    btime + ticks / clk_tck,
    class = c("POSIXct", "POSIXt"),
    tzone = "GMT"
  )

  h <- ps_handle(pid, legacy_ct)
  expect_true(ps_is_running(h))
  expect_equal(ps_name(h), "sleep")
  ps_suspend(h)
  wait_for_status(h, "stopped")
  expect_equal(ps_status(h), "stopped")
  ps_resume(h)
  wait_for_status(h, "sleeping")
  expect_equal(ps_status(h), "sleeping")
})

test_that("ps_handle validates precise (CLOCK_REALTIME-CLOCK_MONOTONIC) create_time", {
  skip_if_no_processx()

  p <- processx::process$new("sleep", "100")
  on.exit(p$kill(), add = TRUE)
  pid <- p$get_pid()

  ## ps_handle(pid) auto-discovers the precise create_time internally
  precise_ct <- ps_create_time(ps_handle(pid))

  ## Re-create the handle with that explicit precise time
  h <- ps_handle(pid, precise_ct)
  expect_true(ps_is_running(h))
  expect_equal(ps_name(h), "sleep")
  ps_suspend(h)
  wait_for_status(h, "stopped")
  expect_equal(ps_status(h), "stopped")
  ps_resume(h)
  wait_for_status(h, "sleeping")
  expect_equal(ps_status(h), "sleeping")
})

test_that("memory_info", {
  ## Argument check
  expect_error(ps_memory_info(123), class = "invalid_argument")

  skip_on_cran()

  p1 <- processx::process$new("ls", c("-lR", "/"))
  on.exit(p1$kill(), add = TRUE)
  ps <- ps_handle(p1$get_pid())

  Sys.sleep(0.2)
  ps_suspend(ps)
  mem <- ps_memory_info(ps)
  mem2 <- scan(
    sprintf("/proc/%d/statm", ps_pid(ps)),
    what = integer(),
    quiet = TRUE
  )
  page_size <- as.integer(system2("getconf", "PAGESIZE", stdout = TRUE))

  expect_equal(mem[["vms"]], mem2[[1]] * page_size)
  expect_equal(mem[["rss"]], mem2[[2]] * page_size)
})

test_that("pid reuse", {
  ## This is simulated, because it is quite some work to force a pid
  ## reuse on some systems. So we create a handle with the pid of a
  ## running process, but wrong (earlier) create time stamp.

  z <- processx::process$new(px(), c("sleep", "600"))
  on.exit(z$kill(), add = TRUE)
  zpid <- z$get_pid()

  ctime <- Sys.time() - 60
  attr(ctime, "tzone") <- "GMT"
  p <- ps_handle(zpid, ctime)

  expect_match(format(p), format_regexp())
  expect_output(print(p), format_regexp())

  expect_equal(ps_pid(p), zpid)
  expect_equal(ps_create_time(p), ctime)
  expect_false(ps_is_running(p))

  chk <- function(expr) {
    err <- tryCatch(expr, error = function(e) e)
    expect_s3_class(err, "no_such_process")
    expect_s3_class(err, "ps_error")
    expect_equal(err$pid, zpid)
  }

  ## All these error out with "no_such_process"
  chk(ps_status(p))
  chk(ps_ppid(p))
  chk(ps_parent(p))
  chk(ps_name(p))
  if (ps_os_type()[["POSIX"]]) chk(ps_uids(p))
  chk(ps_username(p))
  if (ps_os_type()[["POSIX"]]) chk(ps_gids(p))
  chk(ps_terminal(p))

  if (ps_os_type()[["POSIX"]]) chk(ps_send_signal(p, signals()$SIGINT))
  chk(ps_suspend(p))
  chk(ps_resume(p))
  if (ps_os_type()[["POSIX"]]) chk(ps_terminate(p))
  chk(ps_kill(p))

  chk(ps_exe(p))
  chk(ps_cmdline(p))
  chk(ps_environ(p))
  chk(ps_cwd(p))
  chk(ps_memory_info(p))
  chk(ps_cpu_times(p))
  chk(ps_num_threads(p))
  chk(ps_num_fds(p))
  chk(ps_open_files(p))
  chk(ps_connections(p))
})

# Named 'test-z-string' so it will run last.
# Testing for errors that crop up due to differences
# in start (load/attach) time for `ps`.

test_that("string", {
  ps <- ps_handle()

  # Values satisfy encoding assumptions
  expect_true(all(ps_pids() < 52 * 62^3))
  expect_lt(Sys.time(), (62^8 / 1000) * 0.99)

  # Roundtrip through ps_string
  str <- expect_silent(ps_string(ps))
  ps2 <- expect_silent(ps_handle(str))

  # Got the same process back
  expect_true(ps_is_running(ps2))
  expect_identical(ps_pid(ps), ps_pid(ps2))
  expect_identical(ps_ppid(ps), ps_ppid(ps2))

  # Invalid process
  ps2 <- expect_silent(ps_handle(ps_pid(ps), ps_create_time(ps) + 1))
  expect_false(ps_is_running(ps2))
  str <- expect_silent(ps_string(ps2))
  ps2 <- expect_silent(ps_handle(str))
  expect_false(ps_is_running(ps2))

})


test_that("ipc string", {

  skip_on_cran()
  skip_on_covr()

  expect_true(
    callr::r(
      function(str) {
        ps <- ps::ps_handle(str)
        ps::ps_is_running(ps)
      },
      args = list(str = ps_string())
    )
  )

})

#' List of all supported signals
#'
#' Only the signals supported by the current platform are included.
#' @return List of integers, named by signal names.
#' 
#' @export

signals <- function() {
  as.list(ps_env$constants$signals)
}

get_terminal_map <- function() {
  ls <- c(
    dir("/dev", pattern = "^tty.*", full.names = TRUE),
    dir("/dev/pts", full.names = TRUE))
  ret <- structure(ls, names = as.character(.Call(psp__stat_st_rdev, ls)))
  ret[names(ret) != "0"]
}

glob <- local({

  to_regex <- function(glob) {
    restr <- new.env(parent = emptyenv(), size = 1003)
    idx <- 0L
    chr <- strsplit(glob, "", fixed = TRUE)[[1]]
    in_group <- FALSE

    for (c in chr) {
      if (c %in% c("/", "$", "^", "+", ".", "(", ")", "=", "!", "|")) {
        idx <- idx + 1L
        restr[[as.character(idx)]] <- paste0("\\", c)

      } else if (c == "?") {
        idx <- idx + 1L
        restr[[as.character(idx)]] <- "."

      } else if (c == "[" || c == "]") {
        idx <- idx + 1L
        restr[[as.character(idx)]] <- c

      } else if (c == "{") {
        idx <- idx + 1L
        restr[[as.character(idx)]] <- "("
        in_group <- TRUE

      } else if (c == "}") {
        idx <- idx + 1L
        restr[[as.character(idx)]] <- ")"
        in_group <- FALSE

      } else if (c ==",") {
        idx <- idx + 1L
        restr[[as.character(idx)]] <- if (in_group) "|" else paste0("\\", c)

      } else if (c == "*") {
        idx <- idx + 1L
        restr[[as.character(idx)]] <- ".*"

      } else {
        idx <- idx + 1L
        restr[[as.character(idx)]] <- c
      }
    }

    paste0(
      "^",
      paste(mget(as.character(seq_len(idx)), restr), collapse = ""),
      "$")
  }

  test <- function(glob, paths) {
    re <- to_regex(glob)
    grepl(re, paths)
  }

  test_any <- function(globs, paths) {
    if (!length(paths)) return(logical())
    res <- vapply(globs, to_regex, character(1))
    m <- matrix(
      as.logical(unlist(lapply(res, grepl, x = paths))),
      nrow = length(paths))
    apply(m, 1, any)
  }

  structure(
    list(
      .internal = environment(),
      to_regex = to_regex,
      test = test,
      test_any = test_any
    ),
    class = c("standalone_glob", "standalone")
  )
})

# `ps_handle` methods

```
method           A  C  Z
--------------   -  -  -
ps_pid           +  .  +
ps_create_time   +  .  +
ps_is_running    +  .  +
ps_format        +  .  +
-
ps_ppid          .  >  +
ps_parent        .  >  +
ps_name          .  >  +
ps_exe           .  >  Z
ps_cmdline       .  >  Z
ps_status        .  >  +
ps_username      .  >  +
ps_cwd           .  >  Z
ps_uids          .  >  +
ps_gids          .  >  +
ps_terminal      .  >  +
ps_environ       .  >  Z
ps_environ_raw   .  >  Z
ps_num_threads   .  >  Z
ps_cpu_times     .  >  Z
ps_memory_info   .  >  Z
ps_num_fds       .  >  Z
ps_open_files    .  >  Z
ps_connections   .  >  Z
ps_children      .  >  +
ps_send_signal   .  <  +
ps_suspend       .  <  +
ps_resume        .  <  +
ps_terminate     .  <  +
ps_kill          .  <  +
ps_interrupt     .  <  +
```

```
A: always works, even if the process has finished
C: <: checks if process is running, before
   >: checks if process is running, after
Z: +: works fine on a zombie
   Z: errors (zombie_process) on a zombie
```

# System API

## `ps()`

## `ps_pids()`

## `ps_boot_time()`

## Process cleanup

`ps_kill_tree()`, `ps_mark_tree()`,  `with_process_cleanup()`.

## `ps_os_type()`

## `signals()`

# kill 2

    Code
      ps_kill(list(ph5, ph6))
    Condition
      Error:
      ! preventing sending KILL signal to process with PID 0 as it would affect every process in the process group of the calling process (Sys.getpid()) instead of PID 0

---

    Code
      ps_kill(list(ph7, ph8, ph9))
    Condition
      Error:
      ! Failed to kill some processes: 1 (launchd)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system.R
\name{ps_cpu_count}
\alias{ps_cpu_count}
\title{Number of logical or physical CPUs}
\usage{
ps_cpu_count(logical = TRUE)
}
\arguments{
\item{logical}{Whether to count logical CPUs.}
}
\value{
Integer scalar.
}
\description{
If cannot be determined, it returns \code{NA}. It also returns \code{NA} on older
Windows systems, e.g. Vista or older and Windows Server 2008 or older.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
ps_cpu_count(logical = TRUE)
ps_cpu_count(logical = FALSE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testthat-reporter.R
\name{CleanupReporter}
\alias{CleanupReporter}
\title{testthat reporter that checks if child processes are cleaned up in tests}
\usage{
CleanupReporter(reporter = testthat::ProgressReporter)
}
\arguments{
\item{reporter}{A testthat reporter to wrap into a new \code{CleanupReporter}
class.}
}
\value{
New reporter class that behaves exactly like \code{reporter},
but it checks for, and optionally cleans up child processes, at the
specified granularity.
}
\description{
\code{CleanupReporter} takes an existing testthat \code{Reporter} object, and
wraps it, so it checks for leftover child processes, at the specified
place, see the \code{proc_unit} argument below.
}
\details{
Child processes can be reported via a failed expectation, cleaned up
silently, or cleaned up and reported (the default).

If a \code{test_that()} block has an error, \code{CLeanupReporter} does not
emit any expectations at the end of that block. The error will lead to a
test failure anyway. It will still perform the cleanup, if requested,
however.

The constructor of the \code{CleanupReporter} class has options:
\itemize{
\item \code{file}: the output file, if any, this is passed to \code{reporter}.
\item \code{proc_unit}: when to perform the child process check and cleanup.
Possible values:
\itemize{
\item \code{"test"}: at the end of each \code{\link[testthat:test_that]{testthat::test_that()}} block
(the default),
\item \code{"testsuite"}: at the end of the test suite.
}
\item \code{proc_cleanup}: Logical scalar, whether to kill the leftover
processes, \code{TRUE} by default.
\item \code{proc_fail}: Whether to create an expectation, that fails if there
are any processes alive, \code{TRUE} by default.
\item \code{proc_timeout}: How long to wait for the processes to quit. This is
sometimes needed, because even if some kill signals were sent to
child processes, it might take a short time for these to take effect.
It defaults to one second.
\item \code{rconn_unit}: When to perform the R connection cleanup. Possible values
are \code{"test"} and \code{"testsuite"}, like for \code{proc_unit}.
\item \code{rconn_cleanup}: Logical scalar, whether to clean up leftover R
connections. \code{TRUE} by default.
\item \code{rconn_fail}: Whether to fail for leftover R connections. \code{TRUE} by
default.
\item \code{file_unit}: When to check for open files. Possible values are
\code{"test"} and \code{"testsuite"}, like for \code{proc_unit}.
\item \code{file_fail}: Whether to fail for leftover open files. \code{TRUE} by
default.
\item \code{conn_unit}: When to check for open network connections.
Possible values are \code{"test"} and \code{"testsuite"}, like for \code{proc_unit}.
\item \code{conn_fail}: Whether to fail for leftover network connections.
\code{TRUE} by default.
}
}
\note{
Some IDEs, like RStudio, start child processes frequently, and
sometimes crash when these are killed, only use this reporter in a
terminal session. In particular, you can always use it in the
idiomatic \code{testthat.R} file, that calls \code{test_check()} during
\verb{R CMD check}.
}
\section{Examples}{

This is how to use this reporter in \code{testthat.R}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{library(testthat)
library(mypackage)

if  (ps::ps_is_supported()) \{
  reporter <- ps::CleanupReporter(testthat::ProgressReporter)$new(
    proc_unit = "test", proc_cleanup = TRUE)
\} else \{
  ## ps does not support this platform
  reporter <- "progress"
\}

test_check("mypackage", reporter = reporter)
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_windows_nice_values}
\alias{ps_windows_nice_values}
\alias{ps_get_nice}
\alias{ps_set_nice}
\title{Get or set the priority of a process}
\usage{
ps_windows_nice_values()

ps_get_nice(p = ps_handle())

ps_set_nice(p = ps_handle(), value)
}
\arguments{
\item{p}{Process handle.}

\item{value}{On Windows it must be a string, one of the values of
\code{ps_windows_nice_values()}. On Unix it is a priority value that is
smaller than or equal to 20.}
}
\value{
\code{ps_windows_nice_values()} return a character vector of possible
priority values on Windows.

\code{ps_get_nice()} returns a string from
\code{ps_windows_nice_values()} on Windows. On Unix it returns an integer
smaller than or equal to 20.

\code{ps_set_nice()} return \code{NULL} invisibly.
}
\description{
\code{ps_get_nice()} returns the current priority, \code{ps_set_nice()} sets a
new priority, \code{ps_windows_nice_values()} list the possible priority
values on Windows.
}
\details{
Priority values are different on Windows and Unix.

On Unix, priority is an integer, which is maximum 20. 20 is the lowest
priority.
\subsection{Rules:}{
\itemize{
\item On Windows you can only set the priority of the processes the current
user has \code{PROCESS_SET_INFORMATION} access rights to. This typically
means your own processes.
\item On Unix you can only set the priority of the your own processes.
The superuser can set the priority of any process.
\item On Unix you cannot set a higher priority, unless you are the superuser.
(I.e. you cannot set a lower number.)
\item On Unix the default priority of a process is zero.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_handle}
\alias{ps_handle}
\alias{as.character.ps_handle}
\alias{format.ps_handle}
\alias{print.ps_handle}
\title{Create a process handle}
\usage{
ps_handle(pid = NULL, time = NULL)

\method{as.character}{ps_handle}(x, ...)

\method{format}{ps_handle}(x, ...)

\method{print}{ps_handle}(x, ...)
}
\arguments{
\item{pid}{A process id (integer scalar) or process string (from
\code{ps_string()}). \code{NULL} means the current R process.}

\item{time}{Start time of the process. Usually \code{NULL} and ps will query
the start time.}

\item{x}{Process handle.}

\item{...}{Not used currently.}
}
\value{
\code{ps_handle()} returns a process handle (class \code{ps_handle}).
}
\description{
Create a process handle
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
p <- ps_handle()
p
\dontshow{\}) # examplesIf}
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_connections}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_create_time}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_environ}()},
\code{\link{ps_exe}()},
\code{\link{ps_interrupt}()},
\code{\link{ps_is_running}()},
\code{\link{ps_kill}()},
\code{\link{ps_memory_info}()},
\code{\link{ps_name}()},
\code{\link{ps_num_fds}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_send_signal}()},
\code{\link{ps_shared_libs}()},
\code{\link{ps_status}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}
}
\concept{process handle functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_interrupt}
\alias{ps_interrupt}
\title{Interrupt a process}
\usage{
ps_interrupt(p = ps_handle(), ctrl_c = TRUE)
}
\arguments{
\item{p}{Process handle or a list of process handles.}

\item{ctrl_c}{On Windows, whether to send 'CTRL+C'. If \code{FALSE}, then
'CTRL+BREAK' is sent. Ignored on non-Windows platforms.}
}
\description{
Sends \code{SIGINT} on POSIX, and 'CTRL+C' or 'CTRL+BREAK' on Windows.
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_connections}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_create_time}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_environ}()},
\code{\link{ps_exe}()},
\code{\link{ps_handle}()},
\code{\link{ps_is_running}()},
\code{\link{ps_kill}()},
\code{\link{ps_memory_info}()},
\code{\link{ps_name}()},
\code{\link{ps_num_fds}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_send_signal}()},
\code{\link{ps_shared_libs}()},
\code{\link{ps_status}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}
}
\concept{process handle functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{ps_string}
\alias{ps_string}
\title{Encode a \code{ps_handle} as a short string}
\usage{
ps_string(p = ps_handle())
}
\arguments{
\item{p}{Process handle.}
}
\value{
A process string (scalar character), that can be passed to
\code{ps_handle()} in place of a pid.
}
\description{
A convenient format for passing between processes, naming semaphores, or
using as a directory/file name. Will always be 12 alphanumeric characters,
with the first character guarantied to be a letter. Encodes the pid and
creation time for a process.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
(p <- ps_handle())
(str <- ps_string(p))
ps_handle(pid = str)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system.R
\name{ps_shared_lib_users}
\alias{ps_shared_lib_users}
\title{List all processes that loaded a shared library}
\usage{
ps_shared_lib_users(paths, user = ps_username(), filter = NULL)
}
\arguments{
\item{paths}{Character vector of paths of shared libraries to
look up. They must be absolute paths. They don't need to exist.
Forward slashes are converted to backward slashes on Windows, and
the output will always have backward slashes in the paths.}

\item{user}{Character scalar or \code{NULL}. If not \code{NULL}, then only
the processes of this user are considered. It defaults to the
current user.}

\item{filter}{Character vector or \code{NULL}. If not NULL, then it is
a vector of glob expressions, used to filter the process names.}
}
\value{
A data frame with columns:
\itemize{
\item \code{dll}: the file name of the dll file, without the path,
\item \code{path}: path to the shared library,
\item \code{pid}: process ID of the process,
\item \code{name}: name of the process,
\item \code{username}: username of process owner,
\item \code{ps_handle}: \code{ps_handle} object, that can be used to further
query and manipulate the process.
}
}
\description{
List all processes that loaded a shared library
}
\details{
\subsection{Notes:}{

This function currently only works on Windows.

On Windows, a 32 bit R process can only list other 32 bit processes.
Similarly, a 64 bit R process can only list other 64 bit processes.
This is a limitation of the Windows API.

Even though Windows file systems are (almost always) case
insensitive, the matching of \code{paths}, \code{user} and also \code{filter}
are case sensitive. This might change in the future.

This function can be very slow on Windows, because it needs to
enumerate all shared libraries of all processes in the system,
unless the \code{filter} argument is set. Make sure you set \code{filter}
if you can.

If you want to look up multiple shared libraries, list all of them
in \code{paths}, instead of calling \code{ps_shared_lib_users} for each
individually.

If you are after libraries loaded by R processes, you might want to
set \code{filter} to \code{c("Rgui.exe", "Rterm.exe", "rsession.exe")} The
last one is for RStudio.
}
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check() && ps::ps_os_type()[["WINDOWS"]]) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
dlls <- vapply(getLoadedDLLs(), "[[", character(1), "path")
psdll <- dlls[["ps"]][[1]]
r_procs <- c("Rgui.exe", "Rterm.exe", "rsession.exe")
ps_shared_lib_users(psdll, filter = r_procs)
\dontshow{\}) # examplesIf}
}
\seealso{
Other shared library tools: 
\code{\link{ps_shared_libs}()}
}
\concept{shared library tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/disk.R
\name{ps_disk_io_counters}
\alias{ps_disk_io_counters}
\title{System-wide disk I/O counters}
\usage{
ps_disk_io_counters()
}
\value{
A data frame of one row per disk of I/O stats, with columns
\code{name}, \code{read_count} \code{read_merged_count} \code{read_bytes}, \code{read_time},
\code{write_count}, \code{write_merged_count}, \code{write_bytes} \code{write_time}, and
\code{busy_time}.
}
\description{
Returns a data.frame of system-wide disk I/O counters.
}
\details{
Includes the following non-NA fields for all supported platforms:
\itemize{
\item \code{read_count}: number of reads
\item \code{write_count}: number of writes
\item \code{read_bytes}: number of bytes read
\item \code{write_bytes}: number of bytes written
}

And for only some platforms:
\itemize{
\item \code{read_time}: time spent reading from disk (in milliseconds)
\item \code{write_time}: time spent writing to disk (in milliseconds)
\item \code{busy_time}: time spent doing actual I/Os (in milliseconds)
\item \code{read_merged_count}: number of merged reads (see iostats doc)
\item \code{write_merged_count}: number of merged writes (see iostats doc)
}
}
\examples{
\dontshow{if (ps::ps_is_supported() && ps:::ps_os_name() \%in\% c("LINUX", "WINDOWS") && !ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
ps_disk_io_counters()
\dontshow{\}) # examplesIf}
}
\seealso{
Other disk functions: 
\code{\link{ps_disk_partitions}()},
\code{\link{ps_disk_usage}()}
}
\concept{disk functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system.R
\name{ps_tty_size}
\alias{ps_tty_size}
\title{Query the size of the current terminal}
\usage{
ps_tty_size()
}
\description{
If the standard output of the current R process is not a terminal,
e.g. because it is redirected to a file, or the R process is running in
a GUI, then it will throw an error. You need to handle this error if
you want to use this function in a package.
}
\details{
If an error happens, the error message is different depending on
what type of device the standard output is. Some common error messages
are:
\itemize{
\item "Inappropriate ioctl for device."
\item "Operation not supported on socket."
\item "Operation not supported by device."
}

Whatever the error message, \code{ps_tty_size} always fails with an error of
class \code{ps_unknown_tty_size}, which you can catch.
}
\examples{
# An example that falls back to the 'width' option
tryCatch(
  ps_tty_size(),
  ps_unknown_tty_size = function(err) {
    c(width = getOption("width"), height = NA_integer_)
  }
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/disk.R
\name{ps_fs_info}
\alias{ps_fs_info}
\title{File system information for files}
\usage{
ps_fs_info(paths = "/")
}
\arguments{
\item{paths}{A path or a vector of paths. \code{ps_fs_info()} returns
information about the file systems of all paths. \code{path} may contain
direcories as well.}
}
\value{
Data frame with file system information for each
path in \code{paths}, one row per path. Common columns for all
operating systems:
\itemize{
\item \code{path}: The input paths, i.e. the \code{paths} argument.
\item \code{mountpoint}: Directory where the file system is mounted.
On Linux there is a small chance that it was not possible to
look this up, and it is \code{NA_character_}. This is the drive letter
or the mount directory on Windows, with a trailing \verb{\\}.
\item \code{name}: Device name.
On Linux there is a small chance that it was not possible to
look this up, and it is \code{NA_character_}. On Windows this is the
volume GUID path of the form \verb{\\\\?\\Volume\{GUID\}\\}.
\item \code{type}: File system type (character).
On Linux there is a tiny chance that it was not possible to
look this up, and it is \code{NA_character_}.
\item \code{block_size}: File system block size. This is the sector size on
Windows, in bytes.
\item \code{transfer_block_size}: Pptimal transfer block size. On Linux it is
currently always the same as \code{block_size}. This is the cluster size
on Windows, in bytes.
\item \code{total_data_blocks}: Total data blocks in file system. On Windows
this is the number of sectors.
\item \code{free_blocks}: Free blocks in file system. On Windows this is the
number of free sectors.
\item \code{free_blocks_non_superuser}: Free blocks for a non-superuser, which
might be different on Unix. On Windows this is the number of free
sectors for the calling user.
\item \code{id}: File system id. This is a raw vector. On Linux it is
often all zeros. It is always \code{NULL} on Windows.
\item \code{owner}: User that mounted the file system. On Linux and Windows
this is currently always \code{NA_real_}.
\item \code{type_code}: Type of file system, a numeric code. On Windows this
this is \code{NA_real_}.
\item \code{subtype_code}: File system subtype (flavor). On Linux and Windows
this is always \code{NA_real_}.
}

The rest of the columns are flags, and they are operating system
dependent.

macOS:
\itemize{
\item \code{RDONLY}: A read-only filesystem.
\item \code{SYNCHRONOUS}: File system is written to synchronously.
\item \code{NOEXEC}: Can't exec from filesystem.
\item \code{NOSUID}: Setuid bits are not honored on this filesystem.
\item \code{NODEV}: Don't interpret special files.
\item \code{UNION}: Union with underlying filesysten.
\item \code{ASYNC}: File system written to asynchronously.
\item \code{EXPORTED}: File system is exported.
\item \code{LOCAL}: File system is stored locally.
\item \code{QUOTA}: Quotas are enabled on this file system.
\item \code{ROOTFS}: This file system is the root of the file system.
\item \code{DOVOLFS}: File system supports volfs.
\item \code{DONTBROWSE}: File system is not appropriate path to user data.
\item \code{UNKNOWNPERMISSIONS}:  VFS will ignore ownership information on
filesystem filesystemtem objects.
\item \code{AUTOMOUNTED}: File system was mounted by automounter.
\item \code{JOURNALED}: File system is journaled.
\item \code{DEFWRITE}: File system should defer writes.
\item \code{MULTILABEL}: MAC support for individual labels.
\item \code{CPROTECT}: File system supports per-file encrypted data protection.
}

Linux:
\itemize{
\item \code{MANDLOCK}: Mandatory locking is permitted on the filesystem
(see \code{fcntl(2)}).
\item \code{NOATIME}: Do not update access times; see \code{mount(2)}.
\item \code{NODEV}: Disallow access to device special files on this filesystem.
\item \code{NODIRATIME}: Do not update directory access times; see mount(2).
\item \code{NOEXEC}: Execution of programs is disallowed on this filesystem.
\item \code{NOSUID}: The set-user-ID and set-group-ID bits are ignored by
\code{exec(3)} for executable files on this filesystem
\item \code{RDONLY}: This filesystem is mounted read-only.
\item \code{RELATIME}: Update atime relative to mtime/ctime; see \code{mount(2)}.
\item \code{SYNCHRONOUS}: Writes are synched to the filesystem immediately
(see the description of \code{O_SYNC} in `open(2)``).
\item \code{NOSYMFOLLOW}: Symbolic links are not followed when resolving paths;
see `mount(2)``.
}

Windows:
\itemize{
\item \code{CASE_SENSITIVE_SEARCH}: Supports case-sensitive file names.
\item \code{CASE_PRESERVED_NAMES}: Supports preserved case of file names when
it places a name on disk.
\item \code{UNICODE_ON_DISK}: Supports Unicode in file names as they appear on
disk.
\item \code{PERSISTENT_ACLS}: Preserves and enforces access control lists
(ACL). For example, the NTFS file system preserves and enforces
ACLs, and the FAT file system does not.
\item \code{FILE_COMPRESSION}: Supports file-based compression.
\item \code{VOLUME_QUOTAS}: Supports disk quotas.
\item \code{SUPPORTS_SPARSE_FILES}: Supports sparse files.
\item \code{SUPPORTS_REPARSE_POINTS}: Supports reparse points.
\item \code{SUPPORTS_REMOTE_STORAGE}: Supports remote storage.
\item \code{RETURNS_CLEANUP_RESULT_INFO}: On a successful cleanup operation,
the file system returns information that describes additional
actions taken during cleanup, such as deleting the file. File
system filters can examine this information in their post-cleanup
callback.
\item \code{SUPPORTS_POSIX_UNLINK_RENAME}: Supports POSIX-style delete and
rename operations.
\item \code{VOLUME_IS_COMPRESSED}: It is a compressed volume, for example, a
DoubleSpace volume.
\item \code{SUPPORTS_OBJECT_IDS}: Supports object identifiers.
\item \code{SUPPORTS_ENCRYPTION}: Supports the Encrypted File System (EFS).
\item \code{NAMED_STREAMS}: Supports named streams.
\item \code{READ_ONLY_VOLUME}: It is read-only.
\item \code{SEQUENTIAL_WRITE_ONCE}: Supports a single sequential write.
\item \code{SUPPORTS_TRANSACTIONS}: Supports transactions.
\item \code{SUPPORTS_HARD_LINKS}: The volume supports hard links.
\item \code{SUPPORTS_EXTENDED_ATTRIBUTES}: Supports extended attributes.
\item \code{SUPPORTS_OPEN_BY_FILE_ID}: Supports open by FileID.
\item \code{SUPPORTS_USN_JOURNAL}: Supports update sequence number (USN)
journals.
\item \code{SUPPORTS_INTEGRITY_STREAMS}: Supports integrity streams.
\item \code{SUPPORTS_BLOCK_REFCOUNTING}: The volume supports sharing logical
clusters between files on the same volume.
\item \code{SUPPORTS_SPARSE_VDL}: The file system tracks whether each cluster
of a file contains valid data (either from explicit file writes or
automatic zeros) or invalid data (has not yet been written to or
zeroed).
\item \code{DAX_VOLUME}: The volume is a direct access (DAX) volume.
\item \code{SUPPORTS_GHOSTING}: Supports ghosting.
}
}
\description{
File system information for files
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
ps_fs_info(c("/", "~", "."))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_kill}
\alias{ps_kill}
\title{Kill one or more processes}
\usage{
ps_kill(p = ps_handle(), grace = 200)
}
\arguments{
\item{p}{Process handle, or a list of process handles.}

\item{grace}{Grace period, in milliseconds, used on Unix. If it is not
zero, then \code{ps_kill()} first sends a \code{SIGTERM} signal to all processes
in \code{p}. If some proccesses do not terminate within \code{grace}
milliseconds after the \code{SIGTERM} signal, \code{ps_kill()} kills them by
sending \code{SIGKILL} signals.}
}
\value{
Character vector, with one element for each process handle in
\code{p}. If the process was already dead before \code{ps_kill()} tried to kill
it, the corresponding return value is \code{"dead"}. If \code{ps_kill()} just
killed it, it is \code{"killed"}.
}
\description{
Kill the process with SIGKILL preemptively checking whether PID has
been reused. On Windows it uses \code{TerminateProcess()}.
}
\details{
Note that since ps version 1.8, \code{ps_kill()} does not error if the
\code{p} process (or some processes if \code{p} is a list) are already terminated.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ps::ps_os_type()["POSIX"] && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
px <- processx::process$new("sleep", "10")
p <- ps_handle(px$get_pid())
p
ps_kill(p)
p
ps_is_running(p)
px$get_exit_status()
\dontshow{\}) # examplesIf}
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_connections}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_create_time}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_environ}()},
\code{\link{ps_exe}()},
\code{\link{ps_handle}()},
\code{\link{ps_interrupt}()},
\code{\link{ps_is_running}()},
\code{\link{ps_memory_info}()},
\code{\link{ps_name}()},
\code{\link{ps_num_fds}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_send_signal}()},
\code{\link{ps_shared_libs}()},
\code{\link{ps_status}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}
}
\concept{process handle functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/disk.R
\name{ps_fs_stat}
\alias{ps_fs_stat}
\title{File status}
\usage{
ps_fs_stat(paths, follow = TRUE)
}
\arguments{
\item{paths}{Paths to files, directories, devices, etc. They must
exist. They are expanded using \code{\link[base:path.expand]{base::path.expand()}}.}

\item{follow}{Whether to follow symbolic links. If \code{FALSE} it returns
information on the links themselves.}
}
\value{
Data frame with one row for each path in \code{paths}. Columns:
\itemize{
\item \code{path}: Expanded \code{paths}.
\item \code{dev_major}: Major device ID of the device the path resides on.
\item \code{dev_minor}: Minor device ID of the device the path resodes on.
\item \code{inode}: Inode number.
\item \code{mode}: File type and mode (permissions). It is easier to use the
\code{type} and \code{permissions} columns.
\item \code{type}: File type, character. One of
regular file, directory, character device, block device, FIFO, symbolic link, socket.
\item \code{permissions}: Permissions, numeric code in an integer column.
\item \code{nlink}: Number of hard links.
\item \code{uid}: User id of owner.
\item \code{gid}: Group id of owner.
\item \code{rdev_major}: If the path is a device, its major device id,
otherwise \code{NA_integer_}.
\item \code{rdev_minor}: IF the path is a device, its minor device id,
otherwise \code{NA_integer_}.
\item \code{size}: File size in bytes.
\item \code{block_size}: Block size for filesystem I/O.
\item \code{blocks}: Number of 512B blocks allocated.
\item \code{access_time}: Time of last access.
\item \code{modification_time}: Time of last modification.
\item \code{change_time}: Time of last status change.
}
}
\description{
This function is currently not implemented on Windows.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check() && ps_os_type()[["POSIX"]]) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
ps_fs_stat(c(".", tempdir()))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_environ}
\alias{ps_environ}
\alias{ps_environ_raw}
\title{Environment variables of a process}
\usage{
ps_environ(p = ps_handle())

ps_environ_raw(p = ps_handle())
}
\arguments{
\item{p}{Process handle.}
}
\value{
\code{ps_environ()} returns a named character vector (that has a
\code{Dlist} class, so it is printed nicely), \code{ps_environ_raw()} returns a
character vector.
}
\description{
\code{ps_environ()} returns the environment variables of the process, in a
named vector, similarly to the return value of \code{Sys.getenv()}
(without arguments).
}
\details{
Note: this usually does not reflect changes made after the process
started.

\code{ps_environ_raw()} is similar to \code{p$environ()} but returns the
unparsed \code{"var=value"} strings. This is faster, and sometimes good
enough.

These functions throw a \code{zombie_process} error for zombie processes.
}
\section{macOS issues}{


\code{ps_environ()} usually does not work on macOS nowadays. This is because
macOS does not allow reading the environment variables of another
process. Accoding to the Darwin source code, \code{ps_environ} will work is
one of these conditions hold:
\itemize{
\item You are running a development or debug kernel, i.e. if you are
debugging the macOS kernel itself.
\item The target process is same as the calling process.
\item SIP if off.
\item The target process is not restricted, e.g. it is running a binary
that was not signed.
\item The calling process has the
\code{com.apple.private.read-environment-variables} entitlement. However
adding this entitlement to the R binary makes R crash on startup.
}

Otherwise \code{ps_environ} will return an empty set of environment variables
on macOS.

Issue 121 might have more information about this.
}

\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
p <- ps_handle()
p
env <- ps_environ(p)
env[["R_HOME"]]
\dontshow{\}) # examplesIf}
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_connections}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_create_time}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_exe}()},
\code{\link{ps_handle}()},
\code{\link{ps_interrupt}()},
\code{\link{ps_is_running}()},
\code{\link{ps_kill}()},
\code{\link{ps_memory_info}()},
\code{\link{ps_name}()},
\code{\link{ps_num_fds}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_send_signal}()},
\code{\link{ps_shared_libs}()},
\code{\link{ps_status}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}
}
\concept{process handle functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_is_running}
\alias{ps_is_running}
\title{Checks whether a process is running}
\usage{
ps_is_running(p = ps_handle())
}
\arguments{
\item{p}{Process handle.}
}
\value{
Logical scalar.
}
\description{
It returns \code{FALSE} if the process has already finished.
}
\details{
It uses the start time of the process to work around pid reuse. I.e.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
p <- ps_handle()
p
ps_is_running(p)
\dontshow{\}) # examplesIf}
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_connections}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_create_time}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_environ}()},
\code{\link{ps_exe}()},
\code{\link{ps_handle}()},
\code{\link{ps_interrupt}()},
\code{\link{ps_kill}()},
\code{\link{ps_memory_info}()},
\code{\link{ps_name}()},
\code{\link{ps_num_fds}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_send_signal}()},
\code{\link{ps_shared_libs}()},
\code{\link{ps_status}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}
}
\concept{process handle functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system.R
\name{ps_loadavg}
\alias{ps_loadavg}
\title{Return the average system load over the last 1, 5 and 15 minutes as a
tuple.}
\usage{
ps_loadavg()
}
\value{
Numeric vector of length 3.
}
\description{
The “load” represents the processes which are in a runnable
state, either using the CPU or waiting to use the CPU (e.g. waiting for
disk I/O). On Windows this is emulated by using a Windows API that
spawns a thread which keeps running in background and updates results
every 5 seconds, mimicking the UNIX behavior. Thus, on Windows, the
first time this is called and for the next 5 seconds it will return a
meaningless (0.0, 0.0, 0.0) vector. The numbers returned only make sense
if related to the number of CPU cores installed on the system. So, for
instance, a value of 3.14 on a system with 10 logical CPUs means that
the system load was 31.4\% percent over the last N minutes.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
ps_loadavg()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system.R
\name{ps_pids}
\alias{ps_pids}
\title{Ids of all processes on the system}
\usage{
ps_pids()
}
\value{
Integer vector of process ids.
}
\description{
Ids of all processes on the system
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/os.R
\name{ps_os_type}
\alias{ps_os_type}
\alias{ps_is_supported}
\title{Query the type of the OS}
\usage{
ps_os_type()

ps_is_supported()
}
\value{
\code{ps_os_type} returns a named logical vector. The rest of the
functions return a logical scalar.

\code{ps_is_supported()} returns \code{TRUE} if ps supports the current platform.
}
\description{
Query the type of the OS
}
\examples{
ps_os_type()
ps_is_supported()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ps-package.R
\docType{package}
\name{ps-package}
\alias{ps-package}
\title{ps: List, Query, Manipulate System Processes}
\description{
List, query and manipulate all system processes, on 'Windows', 'Linux' and 'macOS'.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/r-lib/ps}
  \item \url{https://ps.r-lib.org/}
  \item Report bugs at \url{https://github.com/r-lib/ps/issues}
}

}
\author{
\strong{Maintainer}: Gábor Csárdi \email{csardi.gabor@gmail.com}

Authors:
\itemize{
  \item Jay Loden
  \item Dave Daeschler
  \item Giampaolo Rodola'
}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kill-tree.R
\name{ps_mark_tree}
\alias{ps_mark_tree}
\alias{with_process_cleanup}
\alias{ps_find_tree}
\alias{ps_kill_tree}
\title{Mark a process and its (future) child tree}
\usage{
ps_mark_tree()

with_process_cleanup(expr)

ps_find_tree(marker)

ps_kill_tree(marker, sig = signals()$SIGKILL, grace = 200)
}
\arguments{
\item{expr}{R expression to evaluate in the new context.}

\item{marker}{String scalar, the name of the environment variable to
use to find the marked processes.}

\item{sig}{The signal to send to the marked processes on Unix. On
Windows this argument is ignored currently.}

\item{grace}{Grace period, in milliseconds, used on Unix, if \code{sig} is
\code{SIGKILL}.  If it is not zero, then \code{ps_kill_tree()} first sends a
\code{SIGTERM} signal to all processes. If some proccesses do not
terminate within \code{grace} milliseconds after the \code{SIGTERM} signal,
\code{ps_kill_tree()} kills them by sending \code{SIGKILL} signals.}
}
\value{
\code{ps_mark_tree()} returns the name of the environment variable,
which can be used as the \code{marker} in \code{ps_kill_tree()}.

\code{ps_find_tree()} returns a list of \code{ps_handle} objects.

\code{ps_kill_tree()} returns the pids of the killed processes, in a named
integer vector. The names are the file names of the executables, when
available.

\code{with_process_cleanup()} returns the value of the evaluated expression.
}
\description{
\code{ps_mark_tree()} generates a random environment variable name and sets
it in the  current R process. This environment variable will be (by
default) inherited by all child (and grandchild, etc.) processes, and
will help finding these processes, even if and when they are (no longer)
related to the current R process. (I.e. they are not connected in the
process tree.)
}
\details{
\code{ps_find_tree()} finds the processes that set the supplied environment
variable and returns them in a list.

\code{ps_kill_tree()} finds the processes that set the supplied environment
variable, and kills them (or sends them the specified signal on Unix).

\code{with_process_cleanup()} evaluates an R expression, and cleans up all
external processes that were started by the R process while evaluating
the expression. This includes child processes of child processes, etc.,
recursively. It returns a list with entries: \code{result} is the result of
the expression, \code{visible} is TRUE if the expression should be printed
to the screen, and \code{process_cleanup} is a named integer vector of the
cleaned pids, names are the process names.

If \code{expr} throws an error, then so does \code{with_process_cleanup()}, the
same error. Nevertheless processes are still cleaned up.
}
\section{Note}{

Note that \code{with_process_cleanup()} is problematic if the R process is
multi-threaded and the other threads start subprocesses.
\code{with_process_cleanup()} cleans up those processes as well, which is
probably not what you want. This is an issue for example in RStudio.
Do not use \code{with_process_cleanup()}, unless you are sure that the
R process is single-threaded, or the other threads do not start
subprocesses. E.g. using it in package test cases is usually fine,
because RStudio runs these in a separate single-threaded process.

The same holds for manually running \code{ps_mark_tree()} and then
\code{ps_find_tree()} or \code{ps_kill_tree()}.

A safe way to use process cleanup is to use the processx package to
start subprocesses, and set the \code{cleanup_tree = TRUE} in
\code{\link[processx:run]{processx::run()}} or the \link[processx:process]{processx::process} constructor.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ps.R
\name{ps}
\alias{ps}
\title{Process table}
\usage{
ps(user = NULL, after = NULL, columns = NULL)
}
\arguments{
\item{user}{Username, to filter the results to matching processes.}

\item{after}{Start time (\code{POSIXt}), to filter the results to processes
that started after this.}

\item{columns}{Columns to include in the result. If \code{NULL} (the default),
then a default set of columns are returned, see below. The columns are
shown in the same order they are specified in \code{columns}, but each
column is included at most once. Use \code{"*"} to include all possible
columns, and prefix a column name with \code{-} to remove it.}
}
\value{
Data frame, see columns below.
}
\description{
Data frame with the currently running processes.
}
\details{
Columns shown by default, if \code{columns} is not given or \code{NULL}:
\itemize{
\item \code{pid}: Process ID.
\item \code{ppid}: Process ID of parent process.
\item \code{name}: Process name.
\item \code{username}: Name of the user (real uid on POSIX).
\item \code{status}: I.e. \emph{running}, \emph{sleeping}, etc.
\item \code{user}: User CPU time.
\item \code{system}: System CPU time.
\item \code{rss}: Resident set size, the amount of memory the process currently
uses. Does not include memory that is swapped out. It does include
shared libraries.
\item \code{vms}: Virtual memory size. All memory the process has access to.
\item \code{created}: Time stamp when the process was created.
\item \code{ps_handle}: \code{ps_handle} objects, in a list column.
}

Additional columns that can be requested via \code{columns}:
\itemize{
\item \code{cmdline}: Command line, in a single string, from \code{\link[=ps_cmdline]{ps_cmdline()}}.
\item \code{vcmdline}: Like \code{cmdline}, but each command line argument in a
separate string.
\item \code{cwd}: Current working directory, from \code{\link[=ps_cwd]{ps_cwd()}}.
\item \code{exe}: Path of the executable of the process, from \code{\link[=ps_exe]{ps_exe()}}.
\item \code{num_fds}: Number of open file descriptors, from \code{\link[=ps_num_fds]{ps_num_fds()}}.
\item \code{num_threads}: Number of threads, from \code{\link[=ps_num_threads]{ps_num_threads()}}.
\item \code{cpu_children_user}: See \code{\link[=ps_cpu_times]{ps_cpu_times()}}.
\item \code{cpu_children_system}: See \code{\link[=ps_cpu_times]{ps_cpu_times()}}.
\item \code{terminal}: Terminal device, from \code{\link[=ps_terminal]{ps_terminal()}}.
\item \code{uid_real}: Real user id, from \code{\link[=ps_uids]{ps_uids()}}.
\item \code{uid_effective}: Effective user id, from \code{\link[=ps_uids]{ps_uids()}}.
\item \code{uid_saved}: Saved user id, from \code{\link[=ps_uids]{ps_uids()}}.
\item \code{gid_real}: Real group id, from \code{\link[=ps_gids]{ps_gids()}}.
\item \code{gid_effective}: Effective group id, from \code{\link[=ps_gids]{ps_gids()}}.
\item \code{gid_saved}: Saved group id, from \code{\link[=ps_gids]{ps_gids()}}.
\item \code{mem_shared}: See \code{\link[=ps_memory_info]{ps_memory_info()}}.
\item \code{mem_text}: See \code{\link[=ps_memory_info]{ps_memory_info()}}.
\item \code{mem_data}: See \code{\link[=ps_memory_info]{ps_memory_info()}}.
\item \code{mem_lib}: See \code{\link[=ps_memory_info]{ps_memory_info()}}.
\item \code{mem_dirty}: See \code{\link[=ps_memory_info]{ps_memory_info()}}.
\item \code{mem_pfaults}: See \code{\link[=ps_memory_info]{ps_memory_info()}}.
\item \code{mem_pageins}: See \code{\link[=ps_memory_info]{ps_memory_info()}}.
\item \code{mem_maxrss}: See \code{\link[=ps_memory_full_info]{ps_memory_full_info()}}.
\item \code{mem_uss}: See \code{\link[=ps_memory_full_info]{ps_memory_full_info()}}.
\item \code{mem_pss}: See \code{\link[=ps_memory_full_info]{ps_memory_full_info()}}.
\item \code{mem_swap}: See \code{\link[=ps_memory_full_info]{ps_memory_full_info()}}.
}

Use \code{"*"} in \code{columns} to include all columns.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_shared_libs}
\alias{ps_shared_libs}
\title{List the dynamically loaded libraries of a process}
\usage{
ps_shared_libs(p = ps_handle())
}
\arguments{
\item{p}{Process handle.}
}
\value{
Data frame with one column currently: \code{path}, the
absolute path to the loaded module or shared library. On Windows
the list includes the executable file itself.
}
\description{
Note: this function currently only works on Windows.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check() && ps::ps_os_type()[["WINDOWS"]]) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# The loaded DLLs of the current process
ps_shared_libs()
\dontshow{\}) # examplesIf}
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_connections}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_create_time}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_environ}()},
\code{\link{ps_exe}()},
\code{\link{ps_handle}()},
\code{\link{ps_interrupt}()},
\code{\link{ps_is_running}()},
\code{\link{ps_kill}()},
\code{\link{ps_memory_info}()},
\code{\link{ps_name}()},
\code{\link{ps_num_fds}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_send_signal}()},
\code{\link{ps_status}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}

Other shared library tools: 
\code{\link{ps_shared_lib_users}()}
}
\concept{process handle functions}
\concept{shared library tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/disk.R
\name{ps_disk_usage}
\alias{ps_disk_usage}
\title{Disk usage statistics, per partition}
\usage{
ps_disk_usage(paths = ps_disk_partitions()$mountpoint)
}
\arguments{
\item{paths}{The mounted file systems to list. By default all file
systems returned by \code{\link[=ps_disk_partitions]{ps_disk_partitions()}} is listed.}
}
\value{
A data frame with columns \code{mountpoint}, \code{total}, \code{used},
\code{available} and \code{capacity}.
}
\description{
The output is similar to the Unix \code{df} command.
}
\details{
Note that on Unix a small percentage of the disk space (5\% typically)
is reserved for the superuser. \code{ps_disk_usage()} returns the space
available to the calling user.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
ps_disk_usage()
\dontshow{\}) # examplesIf}
}
\seealso{
Other disk functions: 
\code{\link{ps_disk_io_counters}()},
\code{\link{ps_disk_partitions}()}
}
\concept{disk functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_wait}
\alias{ps_wait}
\title{Wait for one or more processes to terminate, with a timeout}
\usage{
ps_wait(p, timeout = -1)
}
\arguments{
\item{p}{A process handle, or a list of process handles. The
process(es) to wait for.}

\item{timeout}{Timeout in milliseconds. If -1, \code{ps_wait()} will wait
indefinitely (or until it is interrupted). If 0, then it checks which
processes have already terminated, and returns immediately.}
}
\value{
Logical vector, with one value of each process in \code{p}.
For processes that terminated it contains a \code{TRUE} value. For
processes that are still running it contains a \code{FALSE} value.
}
\description{
This function supports interruption with SIGINT on Unix, or CTRL+C
or CTRL+BREAK on Windows.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check() && ps::ps_os_type()["POSIX"]) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# this example calls `sleep`, so it only works on Unix
p1 <- processx::process$new("sleep", "100")
p2 <- processx::process$new("sleep", "100")

# returns c(FALSE, FALSE) immediately if p1 and p2 are running
ps_wait(list(p1$as_ps_handle(), p2$as_ps_handle()), 0)

# timeouts at one second
ps_wait(list(p1$as_ps_handle(), p2$as_ps_handle()), 1000)

p1$kill()
p2$kill()
# returns c(TRUE, TRUE) immediately
ps_wait(list(p1$as_ps_handle(), p2$as_ps_handle()), 1000)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_send_signal}
\alias{ps_send_signal}
\title{Send signal to a process}
\usage{
ps_send_signal(p = ps_handle(), sig)
}
\arguments{
\item{p}{Process handle, or a list of process handles.}

\item{sig}{Signal number, see \code{\link[=signals]{signals()}}.}
}
\description{
Send a signal to the process. Not implemented on Windows. See
\code{\link[=signals]{signals()}} for the list of signals on the current platform.
}
\details{
It checks if the process is still running, before sending the signal,
to avoid signalling the wrong process, because of pid reuse.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ps::ps_os_type()["POSIX"] && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
px <- processx::process$new("sleep", "10")
p <- ps_handle(px$get_pid())
p
ps_send_signal(p, signals()$SIGINT)
p
ps_is_running(p)
px$get_exit_status()
\dontshow{\}) # examplesIf}
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_connections}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_create_time}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_environ}()},
\code{\link{ps_exe}()},
\code{\link{ps_handle}()},
\code{\link{ps_interrupt}()},
\code{\link{ps_is_running}()},
\code{\link{ps_kill}()},
\code{\link{ps_memory_info}()},
\code{\link{ps_name}()},
\code{\link{ps_num_fds}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_shared_libs}()},
\code{\link{ps_status}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}
}
\concept{process handle functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/memory.R
\name{ps_system_swap}
\alias{ps_system_swap}
\title{System swap memory statistics}
\usage{
ps_system_swap()
}
\value{
Named list. All numbers are in bytes:
\itemize{
\item \code{total}: total swap memory.
\item \code{used}: used swap memory.
\item \code{free}: free swap memory.
\item \code{percent}: the percentage usage.
\item \code{sin}: the number of bytes the system has swapped in from disk
(cumulative). This is \code{NA} on Windows.
\item \code{sout}: the number of bytes the system has swapped out from disk
(cumulative). This is \code{NA} on Windows.
}
}
\description{
System swap memory statistics
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
ps_system_swap()
\dontshow{\}) # examplesIf}
}
\seealso{
Other memory functions: 
\code{\link{ps_system_memory}()}
}
\concept{memory functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_num_fds}
\alias{ps_num_fds}
\title{Number of open file descriptors}
\usage{
ps_num_fds(p = ps_handle())
}
\arguments{
\item{p}{Process handle.}
}
\value{
Integer scalar.
}
\description{
Note that in some IDEs, e.g. RStudio or R.app on macOS, the IDE itself
opens files from other threads, in addition to the files opened from the
main R thread.
}
\details{
For a zombie process it throws a \code{zombie_process} error.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
p <- ps_handle()
ps_num_fds(p)
f <- file(tmp <- tempfile(), "w")
ps_num_fds(p)
close(f)
unlink(tmp)
ps_num_fds(p)
\dontshow{\}) # examplesIf}
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_connections}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_create_time}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_environ}()},
\code{\link{ps_exe}()},
\code{\link{ps_handle}()},
\code{\link{ps_interrupt}()},
\code{\link{ps_is_running}()},
\code{\link{ps_kill}()},
\code{\link{ps_memory_info}()},
\code{\link{ps_name}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_send_signal}()},
\code{\link{ps_shared_libs}()},
\code{\link{ps_status}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}
}
\concept{process handle functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_get_cpu_affinity}
\alias{ps_get_cpu_affinity}
\alias{ps_set_cpu_affinity}
\title{Query or set CPU affinity}
\usage{
ps_get_cpu_affinity(p = ps_handle())

ps_set_cpu_affinity(p = ps_handle(), affinity)
}
\arguments{
\item{p}{Process handle.}

\item{affinity}{Integer vector of CPU numbers to restrict a process to.
CPU numbers start with zero, and they have to be smaller than the
number of (logical) CPUs, see \code{\link[=ps_cpu_count]{ps_cpu_count()}}.}
}
\value{
\code{ps_get_cpu_affinity()} returns an integer vector of CPU
numbers, starting with zero.

\code{ps_set_cpu_affinity()} returns \code{NULL}, invisibly.
}
\description{
\code{ps_get_cpu_affinity()} queries the
\href{https://www.linuxjournal.com/article/6799?page=0,0}{CPU affinity} of
a process. \code{ps_set_cpu_affinity()} sets the CPU affinity of a process.
}
\details{
CPU affinity consists in telling the OS to run a process on a limited
set of CPUs only (on Linux cmdline, the \code{taskset} command is typically
used).

These functions are only supported on Linux and Windows. They error on macOS.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check() && ! ps::ps_os_type()[["MACOS"]]) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# current
orig <- ps_get_cpu_affinity()
orig

# restrict
ps_set_cpu_affinity(affinity = 0:0)
ps_get_cpu_affinity()

# restore
ps_set_cpu_affinity(affinity = orig)
ps_get_cpu_affinity()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_status}
\alias{ps_status}
\title{Current process status}
\usage{
ps_status(p = ps_handle())
}
\arguments{
\item{p}{Process handle.}
}
\value{
Character scalar.
}
\description{
One of the following:
\itemize{
\item \code{"idle"}: Process being created by fork, or process has been sleeping
for a long time. macOS only.
\item \code{"running"}: Currently runnable on macOS and Windows. Actually
running on Linux.
\item \code{"sleeping"} Sleeping on a wait or poll.
\item \code{"disk_sleep"} Uninterruptible sleep, waiting for an I/O operation
(Linux only).
\item \code{"stopped"} Stopped, either by a job control signal or because it
is being traced.
\item \code{"uninterruptible"} Process is in uninterruptible wait. macOS only.
\item \code{"tracing_stop"} Stopped for tracing (Linux only).
\item \code{"zombie"} Zombie. Finished, but parent has not read out the exit
status yet.
\item \code{"dead"} Should never be seen (Linux).
\item \code{"wake_kill"} Received fatal signal (Linux only).
\item \code{"waking"} Paging (Linux only, not valid since the 2.6.xx kernel).
}
}
\details{
It might return \code{NA_character_} on macOS.

Works for zombie processes.
}
\section{Note on macOS}{

On macOS \code{ps_status()} often falls back to calling the external \code{ps}
program, because macOS does not let R access the status of most other
processes. Notably, it is usually able to access the status of other R
processes.

The external \code{ps} program always runs as the root user, and
it also has special entitlements, so it can typically access the status
of most processes.

If this behavior is problematic for you, e.g. because calling an
external program is too slow, set the \code{ps.no_external_ps} option to
\code{TRUE}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{options(ps.no_external_ps = TRUE)
}\if{html}{\out{</div>}}

Note that setting this option to \code{TRUE} will cause \code{ps_status()} to
return \code{NA_character_} for most processes.
}

\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
p <- ps_handle()
p
ps_status(p)
\dontshow{\}) # examplesIf}
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_connections}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_create_time}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_environ}()},
\code{\link{ps_exe}()},
\code{\link{ps_handle}()},
\code{\link{ps_interrupt}()},
\code{\link{ps_is_running}()},
\code{\link{ps_kill}()},
\code{\link{ps_memory_info}()},
\code{\link{ps_name}()},
\code{\link{ps_num_fds}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_send_signal}()},
\code{\link{ps_shared_libs}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}
}
\concept{process handle functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system.R
\name{ps_boot_time}
\alias{ps_boot_time}
\title{Boot time of the system}
\usage{
ps_boot_time()
}
\value{
A \code{POSIXct} object.
}
\description{
Boot time of the system
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_create_time}
\alias{ps_create_time}
\title{Start time of a process}
\usage{
ps_create_time(p = ps_handle())
}
\arguments{
\item{p}{Process handle.}
}
\value{
\code{POSIXct} object, start time, in GMT.
}
\description{
The pid and the start time pair serves as the identifier of the process,
as process ids might be reused, but the chance of starting two processes
with identical ids within the resolution of the timer is minimal.
}
\details{
This function works even if the process has already finished.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
p <- ps_handle()
p
ps_create_time(p)
\dontshow{\}) # examplesIf}
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_connections}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_environ}()},
\code{\link{ps_exe}()},
\code{\link{ps_handle}()},
\code{\link{ps_interrupt}()},
\code{\link{ps_is_running}()},
\code{\link{ps_kill}()},
\code{\link{ps_memory_info}()},
\code{\link{ps_name}()},
\code{\link{ps_num_fds}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_send_signal}()},
\code{\link{ps_shared_libs}()},
\code{\link{ps_status}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}
}
\concept{process handle functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posix.R
\name{signals}
\alias{signals}
\title{List of all supported signals}
\usage{
signals()
}
\value{
List of integers, named by signal names.
}
\description{
Only the signals supported by the current platform are included.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system.R
\name{ps_system_cpu_times}
\alias{ps_system_cpu_times}
\title{System CPU times.}
\usage{
ps_system_cpu_times()
}
\value{
Named list
}
\description{
Every attribute represents the seconds the CPU has spent in the given
mode. The attributes availability varies depending on the platform:
\itemize{
\item \code{user}: time spent by normal processes executing in user mode;
on Linux this also includes guest time.
\item \code{system}: time spent by processes executing in kernel mode.
\item \code{idle}: time spent doing nothing.
}
}
\details{
Platform-specific fields:
\itemize{
\item \code{nice} (UNIX): time spent by niced (prioritized) processes executing
in user mode; on Linux this also includes guest_nice time.
\item \code{iowait} (Linux): time spent waiting for I/O to complete. This is not
accounted in idle time counter.
\item \code{irq} (Linux): time spent for servicing hardware interrupts.
\item \code{softirq} (Linux): time spent for servicing software interrupts.
\item \code{steal} (Linux 2.6.11+): time spent by other operating systems
running in a virtualized environment.
\item \code{guest} (Linux 2.6.24+): time spent running a virtual CPU for guest
operating systems under the control of the Linux kernel.
\item \code{guest_nice} (Linux 3.2.0+): time spent running a niced guest
(virtual CPU for guest operating systems under the control of the
Linux kernel).
}
}
\examples{
\dontshow{if (ps::ps_is_supported()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
ps_system_cpu_times()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/disk.R
\name{ps_fs_mount_point}
\alias{ps_fs_mount_point}
\title{Find the mount point of a file or directory}
\usage{
ps_fs_mount_point(paths)
}
\arguments{
\item{paths}{Paths to files, directories, devices, etc. They must
exist. They are normalized using \code{\link[base:normalizePath]{base::normalizePath()}}.}
}
\value{
Character vector, paths to the mount points of the input
\code{paths}.
}
\description{
Find the mount point of a file or directory
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
ps_fs_mount_point(".")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/disk.R
\name{ps_disk_partitions}
\alias{ps_disk_partitions}
\title{List all mounted partitions}
\usage{
ps_disk_partitions(all = FALSE)
}
\arguments{
\item{all}{Whether to list virtual devices as well. If \code{FALSE}, on
Linux it will still list \code{overlay} and \code{grpcfuse} file systems, to
provide some useful information in Docker containers.}
}
\value{
A data frame with columns \code{device}, \code{mountpoint},
\code{fstype} and \code{options}.
}
\description{
The output is similar the Unix \code{mount} and \code{df} commands.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
ps_disk_partitions(all = TRUE)
ps_disk_partitions()
\dontshow{\}) # examplesIf}
}
\seealso{
Other disk functions: 
\code{\link{ps_disk_io_counters}()},
\code{\link{ps_disk_usage}()}
}
\concept{disk functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system.R
\name{ps_users}
\alias{ps_users}
\title{List users connected to the system}
\usage{
ps_users()
}
\value{
A data frame with columns
\code{username}, \code{tty}, \code{hostname}, \code{start_time}, \code{pid}. \code{tty} and \code{pid}
are \code{NA} on Windows. \code{pid} is the process id of the login process.
For local users the \code{hostname} column is the empty string.
}
\description{
List users connected to the system
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errno.R
\name{errno}
\alias{errno}
\title{List of 'errno' error codes}
\usage{
errno()
}
\description{
For the errors that are not used on the current platform, \code{value} is
\code{NA_integer_}.
}
\details{
A data frame with columns: \code{name}, \code{value}, \code{description}.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
errno()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_connections}
\alias{ps_connections}
\title{List network connections of a process}
\usage{
ps_connections(p = ps_handle())
}
\arguments{
\item{p}{Process handle.}
}
\value{
Data frame, with columns:
\itemize{
\item \code{fd}: integer file descriptor on POSIX systems, \code{NA} on Windows.
\item \code{family}: Address family, string, typically \code{AF_UNIX}, \code{AF_INET} or
\code{AF_INET6}.
\item \code{type}: Socket type, string, typically \code{SOCK_STREAM} (TCP) or
\code{SOCK_DGRAM} (UDP).
\item \code{laddr}: Local address, string, \code{NA} for UNIX sockets.
\item \code{lport}: Local port, integer, \code{NA} for UNIX sockets.
\item \code{raddr}: Remote address, string, \code{NA} for UNIX sockets. This is
always \code{NA} for \code{AF_INET} sockets on Linux.
\item \code{rport}: Remote port, integer, \code{NA} for UNIX sockets.
\item \code{state}: Socket state, e.g. \code{CONN_ESTABLISHED}, etc. It is \code{NA}
for UNIX sockets.
}
}
\description{
For a zombie process it throws a \code{zombie_process} error.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
p <- ps_handle()
ps_connections(p)
sc <- socketConnection("httpbin.org", port = 80)
ps_connections(p)
close(sc)
ps_connections(p)
\dontshow{\}) # examplesIf}
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_create_time}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_environ}()},
\code{\link{ps_exe}()},
\code{\link{ps_handle}()},
\code{\link{ps_interrupt}()},
\code{\link{ps_is_running}()},
\code{\link{ps_kill}()},
\code{\link{ps_memory_info}()},
\code{\link{ps_name}()},
\code{\link{ps_num_fds}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_send_signal}()},
\code{\link{ps_shared_libs}()},
\code{\link{ps_status}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}
}
\concept{process handle functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/low-level.R
\name{ps_memory_info}
\alias{ps_memory_info}
\alias{ps_memory_full_info}
\title{Memory usage information}
\usage{
ps_memory_info(p = ps_handle())

ps_memory_full_info(p = ps_handle())
}
\arguments{
\item{p}{Process handle.}
}
\value{
Named real vector.
}
\description{
Memory usage information
}
\details{
\code{ps_memory_info()} returns information about memory usage.

It returns a named vector. Portable fields:
\itemize{
\item \code{rss}: "Resident Set Size", this is the non-swapped physical memory a
process has used (bytes). On UNIX it matches "top"‘s 'RES' column (see doc). On
Windows this is an alias for \code{wset} field and it matches "Memory"
column of \code{taskmgr.exe}.
\item \code{vmem}: "Virtual Memory Size", this is the total amount of virtual
memory used by the process (bytes). On UNIX it matches "top"‘s 'VIRT' column
(see doc). On Windows this is an alias for the \code{pagefile} field and
it matches the "Working set (memory)" column of \code{taskmgr.exe}.
}

Non-portable fields:
\itemize{
\item \code{shared}: (Linux) memory that could be potentially shared with other
processes (bytes). This matches "top"‘s 'SHR' column (see doc).
\item \code{text}: (Linux): aka 'TRS' (text resident set) the amount of memory
devoted to executable code (bytes). This matches "top"‘s 'CODE' column (see
doc).
\item \code{data}: (Linux): aka 'DRS' (data resident set) the amount of physical
memory devoted to other than executable code (bytes). It matches "top"‘s
'DATA' column (see doc).
\item \code{lib}: (Linux): the memory used by shared libraries (bytes).
\item \code{dirty}: (Linux): the amount of memory in dirty pages (bytes).
\item \code{pfaults}: (macOS): number of page faults.
\item \code{pageins}: (macOS): number of actual pageins.
}

For the explanation of Windows fields see the
\href{https://learn.microsoft.com/en-us/windows/win32/api/psapi/ns-psapi-process_memory_counters_ex}{PROCESS_MEMORY_COUNTERS_EX}
structure.

\code{ps_memory_full_info()} returns all fields as \code{ps_memory_info()}, plus
additional information, but typically takes slightly longer to run, and
might not have access to some processes that \code{ps_memory_info()} can
query:
\itemize{
\item \code{maxrss} maximum resident set size over the process's lifetime. This
only works for the calling process, otherwise it is \code{NA_real_}.
\item \code{uss}: Unique Set Size, this is the memory which is unique to a
process and which would be freed if the process was terminated right
now.
\item \code{pss} (Linux only): Proportional Set Size, is the amount of memory
shared with other processes, accounted in a way that the amount is
divided evenly between the processes that share it. I.e. if a process
has 10 MBs all to itself and 10 MBs shared with another process its
PSS will be 15 MBs.
\item \code{swap} (Linux only): amount of memory that has been swapped out to
disk.
}

They both throw a \code{zombie_process()} error for zombie processes.
}
\examples{
\dontshow{if (ps::ps_is_supported() && ! ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
p <- ps_handle()
p
ps_memory_info(p)
ps_memory_full_info(p)
\dontshow{\}) # examplesIf}
}
\seealso{
Other process handle functions: 
\code{\link{ps_children}()},
\code{\link{ps_cmdline}()},
\code{\link{ps_connections}()},
\code{\link{ps_cpu_times}()},
\code{\link{ps_create_time}()},
\code{\link{ps_cwd}()},
\code{\link{ps_descent}()},
\code{\link{ps_environ}()},
\code{\link{ps_exe}()},
\code{\link{ps_handle}()},
\code{\link{ps_interrupt}()},
\code{\link{ps_is_running}()},
\code{\link{ps_kill}()},
\code{\link{ps_name}()},
\code{\link{ps_num_fds}()},
\code{\link{ps_num_threads}()},
\code{\link{ps_open_files}()},
\code{\link{ps_pid}()},
\code{\link{ps_ppid}()},
\code{\link{ps_resume}()},
\code{\link{ps_send_signal}()},
\code{\link{ps_shared_libs}()},
\code{\link{ps_status}()},
\code{\link{ps_suspend}()},
\code{\link{ps_terminal}()},
\code{\link{ps_terminate}()},
\code{\link{ps_uids}()},
\code{\link{ps_username}()}
}
\concept{process handle functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/macos.R
\name{ps_apps}
\alias{ps_apps}
\title{List currently running applications}
\usage{
ps_apps()
}
\value{
A data frame with columns:
\itemize{
\item \code{pid}: integer process id.
\item \code{name}: process name.
\item \code{bundle_identifier}: bundle identifier, e.g. \code{com.apple.dock}.
\item \code{bundle_url}: bundle URL, a \verb{file://} URL to the app bundle.
\item \code{arch}: executable architecture, possible values are
arm64, i386, x86_64, ppc, ppc64.
\item \code{executable_url}: \verb{file://} URL to the executable file.
\item \code{launch_date}: launch time stamp, a \code{POSIXct} object, may be \code{NA}.
\item \code{finished_launching}: whether the app has finished launching.
\item \code{active}: whether the app is active.
\item \code{activation_policy}: one of the following values:
\itemize{
\item \code{regular}: the application is an ordinary app that appears in the
Dock and may have a user interface.
\item \code{accessory}: the application doesn’t appear in the Dock and
doesn’t have a menu bar, but it may be activated programmatically
or by clicking on one of its windows.
\item \code{prohibited}: the application doesn’t appear in the Dock and may
not create windows or be activated.
}
}
}
\description{
This function currently only works on macOS.
}
\examples{
\dontshow{if (ps_is_supported() && ps_os_type()[["MACOS"]] && !ps:::is_cran_check()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
ps_apps()
\dontshow{\}) # examplesIf}
}
